/**
   @file common.h

   Copyright (C) 2004-2010 Nokia Corporation. All rights reserved.

   @author Janne Yllehto <janne.ylalehto@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef _COMMON_H_
#define _COMMON_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/if_ether.h>
#include <mce/mode-names.h>
#include <stdio.h>
#include <net/if.h>
#include <connsettings/conn_settings.h>
#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>
#include <wlancond.h>

/* Scan results helper struct */
typedef struct scan_results_t
{
	char           ssid[WLANCOND_MAX_SSID_SIZE+1];
	guint32        ssid_len;
	unsigned char  bssid[ETH_ALEN];
	gint32         rssi;
	guint32        channel;
	guint32        cap_bits;
	guint32        extra_cap_bits;
	guint32        wpa_ie_len;
	guint8         *wpa_ie;
	guint8         *uuid_e;
} scan_results_t;

/* Generic WPA IE storage type */
typedef struct wpa_ie_list_t
{
	unsigned char  bssid[ETH_ALEN];
	guint32        ie_len;
	guint8         *wpa_ie;
} wpa_ie_list_t;

/* Own WPA IE storage type */
typedef struct wpa_ie_save_t
{
	guint32 ie_len;
	unsigned char *ie;
} wpa_ie_save_t;

/* This struct has all parameters needed for connection */
typedef struct connect_params_t
{
	dbus_int32_t mode;
	dbus_int32_t encryption;
	dbus_int32_t power_level;
	dbus_int32_t default_key;
	dbus_uint32_t adhoc_channel;
	dbus_uint32_t flags;
	guint32 channel;
	char ssid[WLANCOND_MAX_SSID_SIZE+1];
	unsigned int ssid_len;
	unsigned int authentication_type;
	int key_len[4];
	unsigned char key[4][WLANCOND_MAX_KEY_LEN];
	gboolean wep_try_open_mode_first;
	unsigned char bssid[ETH_ALEN];
} connect_params;

#define PMK_CACHE_SIZE 32
typedef struct pmksa_cache_t
{
	unsigned char mac[ETH_ALEN];
	unsigned char pmkid[WLANCOND_PMKID_LEN];
} pmksa_cache;

#define WLANCOND_IF_NAME	"wlan0"

/* WLAN status and state is kept in this struct.
   Scanning is separated from the state because scanning
   can be initiated separately from other WLAN activity
 */
typedef struct wlan_status_t
{
	/* Interface name cache */
	gchar ifname[IFNAMSIZ+1];
	/* Interface index */
	guint32 ifindex;
	/* Own MAC address */
	gchar own_mac[ETH_ALEN];
	/* Scan ssid */
	gchar scan_ssid[WLANCOND_MAX_SSID_SIZE+1];
	/* Scan ssid length */
	gint scan_ssid_len;
	/* Scanning state */
	guint scan;
	/* WLAN state */
	guint state;
	/* Power state */
	guint power;
	/* Cipher suites */
	guint pairwise_cipher;
	guint group_cipher;
	/* Our WPA IE */
	struct wpa_ie_save_t wpa_ie;
	/* connect params */
	struct connect_params_t conn;
	/* Connection settings */
	ConnSettings *setting;
	/* pmksa cache */
	GSList* pmk_cache;
	/* Roam cache */
	GSList* roam_cache;
	/* Association retry counter */
	guint retry_count;
	/* Roam scan */
	guint roam_scan;
	/* Roam scan timer ID */
	guint roam_scan_id;
	/* Last scan (in seconds) */
	gint last_scan;
	/* Normal scan timer ID */
	guint scan_id;
	/* Rescan timer ID */
	guint rescan_id;
	/* Desired powersave */
	guint powersave;
	/* Call state */
	guint call_type;
	/* Country code */
	char* country_code;
	/* Last country code update */
	gint last_code;
	/* Last country code calibrated*/
	gchar* last_calibrated_code;
	/* Network latency file */
	FILE* latency_file;
	/* IPTABLES enable file */
	FILE* iptables_file;
	/* Calibration needed */
	gboolean calibration_needed;
	/* Signal state */
	gboolean signal;
	/* IP configuration state */
	gboolean ip_ok;
	/* Allow all cipher_suites */
	gboolean allow_all_ciphers;
	/* Interface down by wlancond */
	gboolean if_down_self;
	/* Traffic count */
	gboolean high_traffic;
	/* enable full power save mode */
	gboolean full_psm_enabled;
	/* DBUS filters active */
	gboolean dbus_filters;
	/* IPtables loaded */
	gboolean iptables_loaded;
	/* Interface state */
	gboolean interface_up;
} wlan_status_t;

/* Roam cache entries are found by this key */
typedef struct roam_cache_key_t
{
	unsigned char* bssid;
	guint32 channel;
} roam_cache_key_t;

/* WLAN signal */
#define WLANCOND_LOW TRUE
#define WLANCOND_HIGH FALSE

/* Scan state */
enum scan_state {
	SCAN_NOT_ACTIVE = 0,
	SCAN_ACTIVE
};

/* Defines for call type checking */
#define	CALL_VOIP     (1<<0)
#define	CALL_CELLULAR (1<<1)
#define	CALL_GPRS     (1<<2)

/* Generic WLAN state */
enum wlan_state {
	WLAN_NOT_INITIALIZED = 0,
	WLAN_INITIALIZED,
	WLAN_INITIALIZED_FOR_SCAN,
	WLAN_INITIALIZED_FOR_CONNECTION,
	WLAN_NO_ADDRESS,
	WLAN_CONNECTED
};

/* WLAN power defines */
#define WLANCOND_TX_POWERLIMITED 1300 // 13dBm
#define WLANCOND_TX_POWER10DBM  1000 // 10mW = 10dBm = 1000 mBm
#define WLANCOND_TX_POWER100DBM 2000 // 100mW = 20dBm = 2000 mBm

/* WLAN power state */
enum wlan_power_state {
	WLANCOND_POWER_ON = 1,
	WLANCOND_CAM_TIMEOUT,
	WLANCOND_FULL_POWERSAVE
};

#define NULL_BSSID (unsigned char*)"\x00\x00\x00\x00\x00\x00"

#define WLANCOND_LATENCY_FILE "/dev/network_latency"
#define WLANCOND_DEFAULT_NETWORK_LATENCY 2000000000 // 100ms
#define WLANCOND_FULL_PSM_LATENCY 1900000001

#define MIC_FAILURE_TIMEOUT 60 // 60 seconds

/* Signal sending defines */
#define NO_SIGNAL           0
#define DISCONNECTED_SIGNAL 1

/* Maximum values */
#define WLANCOND_MAX_NETWORKS 30
#define WLANCOND_MAX_SCAN_TRIES 3

#define WLANCOND_INITIAL_ROAM_SCAN_DELAY 2 // 2 seconds
#define WLANCOND_ROAM_THRESHOLD 10 // dBm
#define WLANCOND_5GHZ_BENEFIT 5 // dBm
#define WLANCOND_MIN_ROAM_SCAN_INTERVAL 60 // 60 seconds
#define WLANCOND_MAX_ROAM_SCAN_INTERVAL 16*60 // 16 minutes
#define WLANCOND_COUNTRY_CACHE_TIME 3600 // 1 hour
#define WLANCOND_RSSI_PENALTY 25 // 25dBm
#define WLANCOND_MINIMUM_SIGNAL -99
#define WLANCOND_MINIMUM_AUTOCONNECT_RSSI -85

/* The default channel */
#define WLANCOND_INITIAL_CHANNEL 1

/* Split scan/normal scan */
#define SCAN_SYSFS_FILE "/sys/devices/platform/wl1271/"
#define SCAN_SIZE_FILE SCAN_SYSFS_FILE "scan_size"
#define SCAN_DELAY_FILE SCAN_SYSFS_FILE "scan_delay"
#define SCAN_SIZE_SPLIT   1
#define SCAN_SIZE_NORMAL  24
#define SCAN_DELAY_SPLIT  40 /* ms */
#define SCAN_DELAY_NORMAL 0  /* ms */

/* Settings paths */
#define WLANCOND_DEBUG_LEVEL "wlancond_debug_level"
#define WLANCOND_ALLOW_ALL_CIPHERS "allow_wep_ciphers_in_WPA"

/* Debug printing priority */
#define WLANCOND_PRIO_HIGH   2
#define WLANCOND_PRIO_MEDIUM 1
#define WLANCOND_PRIO_LOW    0

/* Deauthenticate command */
#define WLANCOND_MLME_DEAUTH	0

/* Disassociate command */
#define WLANCOND_MLME_DISASSOC	1		

/* Disconnect command */
#define WLANCOND_NL80211_MLME_DISCONNECT 0xD15C

/* Deauthenticate reasons */
#define WLANCOND_REASON_LEAVING 3
#define WLANCOND_REASON_MIC_FAILURE 14

/* Defines for the CSNET interface */
#define CSNET_INTERFACE "com.nokia.csd.CSNet"
#define CSNET_OBJECT "/com/nokia/csd/csnet"
#define CSNET_ACTION "Get"
#define CSNET_OPER_IFACE "com.nokia.csd.CSNet.NetworkOperator"
#define CSNET_OPER_MCC "OperatorMCC"
#define CSNET_OPERATOR_CHANGED_SIG "OperatorChanged"

/* Defines for the GPRS interface */
#define GPRS_DBUS_INTERFACE "com.nokia.csd.GPRS"
#define GPRS_DBUS_DATACOUNTERS_SIG "DataCounters"

/* Defines for the policy framework */
#define POLICY_SERVICE_NAME                "com.nokia.policy"
#define POLICY_ACTIONS_SIG                 "audio_actions"

/* Define for full power save mode flag */
#define FULL_PSM_SETTING "full_psm_enabled"

/* Extra capability bits*/
/* Internal cipher suite defines */
#define WLANCOND_WEP40        (1<<0)
#define WLANCOND_WEP104       (1<<1)
#define WLANCOND_WEP40_GROUP  (1<<2)
#define WLANCOND_WEP104_GROUP (1<<3)
/* U-APSD supported */
#define WLANCOND_UAPSD_SUPPORTED (1<<31)

/* Internal link mode options */
#define WLANCOND_IF_DORMANT 1
#define WLANCOND_IF_READY 6

/* Common functions */
void init_logging(void);
void wlancond_print(guint priority, const char *debug, ...);
gboolean monitor_wi(void);
void send_dbus_scan_results(GSList* scan_results_save, const char* sender,
		dbus_int32_t number_of_results);
void clean_scan_results(GSList **scan_results_save);
void disconnected_signal(void);
void mode_change(dbus_uint32_t radio_mode);
#ifdef ACTIVITY_CHECK
void activity_check(dbus_bool_t activity);
#endif

void clean_scan_results_item(gpointer data, gpointer user_data);
void print_mac(guint priority, const char *message, guchar* mac);
gboolean roam_scanning(void);
int netlink_send_linkmode(int lmode);
int set_iptables(void);
void remove_iptables(void);
gboolean monitor_iap_keys(void);
gboolean is_connected_state(void);

#endif /* _COMMON_H_ */
