/**
   @file dbus-handler.c

   Copyright (C) 2004-2010 Nokia Corporation. All rights reserved.

   @author Janne Yllehto <janne.ylalehto@nokia.com>

   Portions of this file are
   Copyright (c) 1997-2002 Jean Tourrilhes <jt@hpl.hp.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <glib.h>
#include <errno.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <linux/socket.h>
#include <sys/ioctl.h>
#include <osso-ic-dbus.h>
#include <syslog.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>
#include <eap-dbus.h>
#include <wlancond-dbus.h>
#include <mce/mode-names.h>
#ifdef USE_MCE_MODE
#include <mce/dbus-names.h>
#endif
#include "log.h"
#include "dbus.h"
#include "dbus-helper.h"
#include "dbus-handler.h"
#include "common.h"
#include "wpa.h"
#include "wps.h"
#include "wapi.h"

#define WLANCOND_SHUTDOWN_DELAY 12 //12s
#define WLANCOND_CONNECT_TIMEOUT 14 //14s
#define WLANCOND_SCAN_TIMEOUT 8 //8s
#define WLANCOND_RESCAN_DELAY 1
#define WLANCOND_GPRS_TIMEOUT 2 //2s

#define CALIBRATION_APPLICATION "/usr/bin/wl1271-cal"

/** Saved DBUS scan name */
static char *scan_name_cache = NULL;
/** Saved DBUS connect name */
static char *connect_name_cache = NULL;

/** Socket is global */
static int wlan_socket = -1;

/** WLAN status and state is here */
struct wlan_status_t wlan_status;

static gboolean _flight_mode = FALSE;
static gboolean power_down_after_scan = FALSE;
static dbus_bool_t saved_inactivity = FALSE;

/** Timer IDs */
static guint wlan_if_down_timer_id = 0;
static guint wlan_connect_timer_id = 0;
static guint wlan_gprs_timer_id = 0;

/** Debug level */
gint debug_level = 0;

#define WLAN_PREFIX_STR "wlan"
/**
   Wlancond debug printing function.
*/
void wlancond_print(guint priority, const char *debug, ...) {
	va_list args;
	char buffer[200];

	switch (debug_level) {
		/* In debug level 0 only high prio printing (release) */
	case 0:
		if (priority > WLANCOND_PRIO_MEDIUM) {
			va_start(args, debug);
			vsnprintf(buffer, sizeof(buffer), debug, args);
			va_end(args);

			syslog(LOG_INFO | LOG_DAEMON, "%s", buffer);
			return;
		}
		break;
	case 1:
		if (priority > WLANCOND_PRIO_LOW) {
			va_start(args, debug);
			vsnprintf(buffer, sizeof(buffer), debug, args);
			va_end(args);

			syslog(LOG_INFO | LOG_DAEMON, "%s", buffer);
			return;
		}
		break;
		/* Print everything */
	default:
		va_start(args, debug);
		vsnprintf(buffer, sizeof(buffer), debug, args);
		va_end(args);

		syslog(LOG_INFO | LOG_DAEMON, "%s", buffer);
		return;
	}
}
/**
 * Helper function for socket opening.
 */
int socket_open(void)
{
        if (wlan_socket > 0)
                return wlan_socket;

        wlan_socket = socket(AF_INET, SOCK_DGRAM, 0);

        if (wlan_socket < 0)
                die("socket() failed");

        return wlan_socket;
}

static int get_own_mac(void)
{
        struct ifreq req;

        memset(&req , 0, sizeof(req));
        memcpy(req.ifr_name, wlan_status.ifname, IFNAMSIZ);

        if (ioctl(socket_open(), SIOCGIFHWADDR, &req) < 0)
        {
                return -1;
        }

        memcpy(wlan_status.own_mac, req.ifr_hwaddr.sa_data, ETH_ALEN);

	print_mac(WLANCOND_PRIO_LOW, "MAC:", (guchar*) wlan_status.own_mac);

        return 0;
}

/**
    Helper function for getting integer value from settings.
    @param path Setting path to search for user specified value.
    @return settings integer.
*/
static gint get_setting_int(const gchar* path)
{
	gint value = -1;
	ConnSettingsValue *conn_setting_value;

	conn_setting_value = conn_settings_get(wlan_status.setting, path);
	if (conn_setting_value == NULL) {
		return -1;
	}

	if (conn_setting_value->type == CONN_SETTINGS_VALUE_INT) {
		value = conn_setting_value->value.int_val;
		DLOG_DEBUG("User selected value %d for %s", value, path);

	}

	conn_settings_value_destroy(conn_setting_value);

	return value;
}
/**
    Helper function for getting boolean value from the settings.
    @param path Setting path to search for user specified value.
    @param error Variable to hold possible error.
    @return value Boolean value.
*/
gboolean get_setting_bool(const gchar* path, gboolean *error)
{
	gboolean value = FALSE;
	ConnSettingsValue *conn_setting_value;

	*error = TRUE;

	conn_setting_value = conn_settings_get(wlan_status.setting, path);
	if (conn_setting_value == NULL) {
		return value;
	}

	if (conn_setting_value->type == CONN_SETTINGS_VALUE_BOOL) {
		value = conn_setting_value->value.bool_val;
		DLOG_DEBUG("User selected %s for %s", value==TRUE?
						"true":"false", path);
		*error = FALSE;
	}

	conn_settings_value_destroy(conn_setting_value);

	return value;
}
/**
   Initialize logging.
*/
void init_logging(void) {

	wlan_status.setting = conn_settings_open(CONN_SETTINGS_CONNECTION, 
						 NULL);

	if (wlan_status.setting == NULL ) { 
		DLOG_ERR("Opening connection settings failed");
		return;
	}
	debug_level = get_setting_int(WLANCOND_DEBUG_LEVEL);
	if (debug_level < 0)
		debug_level = 0;

	if (debug_level > 0) {
		DLOG_DEBUG("Debug level increased to %d", debug_level);
	}
}

/**
   Update our Information Element.
   @param wpa_ie WPA Information Element.
   @param wpa_ie_len WPA Information Element length.
*/
void update_own_ie(unsigned char* wpa_ie, guint wpa_ie_len) 
{
	g_free(wlan_status.wpa_ie.ie);
	wlan_status.wpa_ie.ie = wpa_ie;
	wlan_status.wpa_ie.ie_len = wpa_ie_len;
}

/**
   Get encryption info.
   @return status.
*/
guint get_encryption_info(void) 
{
	guint auth_status = 0;

	if (wlan_status.pairwise_cipher & CIPHER_SUITE_CCMP) {
		auth_status |= WLANCOND_WPA_AES;
	} else if (wlan_status.pairwise_cipher & CIPHER_SUITE_TKIP) {
		auth_status |= WLANCOND_WPA_TKIP;
	} else if (wlan_status.pairwise_cipher & CIPHER_SUITE_SMS4) {
		auth_status |= WLANCOND_WAPI_SMS4;
	} else if (wlan_status.pairwise_cipher & CIPHER_SUITE_WEP40 ||
		   wlan_status.pairwise_cipher & CIPHER_SUITE_WEP104) {
		auth_status |= WLANCOND_WEP;
	} else {
		auth_status |= WLANCOND_OPEN;
	}
	return auth_status;
}

/**
   Helper function for cleaning handler.
*/
int clean_dbus_handler(void)
{
	if (wlan_socket > 0) {
		close(wlan_socket);
		wlan_socket = -1;
	}
	return 0;
}
/**
   Helper function for mode change.
   @param radio_mode New mode.
*/
void mode_change(dbus_uint32_t radio_mode) {

        gboolean mode = !!(radio_mode & MCE_RADIO_STATE_WLAN);

	DLOG_INFO("WLAN flight mode changed to \"%s\"",	
			mode?"Normal":"Flight");

	if (!mode) {
		int signal = NO_SIGNAL;
		if (is_connected_state() == TRUE)
			signal = DISCONNECTED_SIGNAL;
		set_wlan_state(WLAN_NOT_INITIALIZED, signal, FALSE);
		_flight_mode = TRUE;	
	}
	else {
	 	_flight_mode = FALSE;		
	}
}
/**
   Helper function for mode change DBUS handling.
   @param message DBUS message.
*/
#ifdef USE_MCE_MODE
static DBusHandlerResult mode_change_dbus(DBusMessage *message) {

	dbus_uint32_t mode;

	if (!dbus_message_get_args(message, NULL,
				   DBUS_TYPE_UINT32, &mode,
				   DBUS_TYPE_INVALID)) {
		DLOG_ERR("Invalid arguments for device_mode_ind signal");
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	mode_change(mode);

	return DBUS_HANDLER_RESULT_HANDLED;
}
#ifdef ACTIVITY_CHECK
static DBusHandlerResult activity_check_dbus(DBusMessage *message) {

	if (!dbus_message_get_args(message, NULL,
				   DBUS_TYPE_BOOLEAN, &saved_inactivity,
				   DBUS_TYPE_INVALID)) {
		DLOG_ERR("Invalid arguments for device_activity signal");
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}
	activity_check(saved_inactivity);

	return DBUS_HANDLER_RESULT_HANDLED;
}
#endif
#endif

#ifdef ACTIVITY_CHECK
/**
 * Helper function for mode change.
 */
void activity_check(dbus_bool_t inactivity) {

	if (get_wlan_state() != WLAN_CONNECTED) {
                return;
	}

	if (inactivity == FALSE) {
		//DLOG_DEBUG("WLAN activity mode changed to active");
		if (nl80211_set_cqm(WLANCOND_CQM_THRESHOLD_ACTIVE, 
				    WLANCOND_CQM_HYSTERESIS)) 
			DLOG_ERR("nl80211 - CQM monitor setting failed");

		/* No need to check traffic */
		remove_iptables();
	} else {
		//DLOG_DEBUG("WLAN activity mode changed to inactive");
		if (nl80211_set_cqm(WLANCOND_CQM_THRESHOLD_IDLE, 
				    WLANCOND_CQM_HYSTERESIS)) 
			DLOG_ERR("nl80211 - CQM monitor setting failed");
		
		if (set_iptables() < 0)
			DLOG_ERR("Setting iptables failed");
	}
	set_power_state(wlan_status.powersave);
}

/**
 * Helper function for determining inactivity.
 */
gboolean get_inactivity_status(void)
{
	return saved_inactivity;
}
#endif
/**
   Check ICD DBUS signal.
   @param message DBUS message.
   @return DBusHandlerResult.
*/
static DBusHandlerResult icd_check_signal_dbus(DBusMessage *message) {

	char *icd_name;
	char *icd_type;
	char *icd_state;
	DBusError dbus_error;

	if ((get_wlan_state() != WLAN_CONNECTED &&
	     get_wlan_state() != WLAN_NO_ADDRESS) ||
	    get_mode() != WLANCOND_INFRA) {
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	dbus_error_init(&dbus_error);
	if (!dbus_message_get_args(message, &dbus_error,
				   DBUS_TYPE_STRING, &icd_name,
				   DBUS_TYPE_STRING, &icd_type,
				   DBUS_TYPE_STRING, &icd_state,
				   DBUS_TYPE_INVALID ) )
	{
		DLOG_ERR("Could not get args from signal, '%s'",
			 dbus_error.message);
		dbus_error_free(&dbus_error);
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	if (icd_state != NULL && strncmp(icd_state, "CONNECTED", 9) == 0) {

		set_wlan_state(WLAN_CONNECTED, NO_SIGNAL, TRUE);

		DLOG_DEBUG("Going to power save");

		if (set_power_state(wlan_status.powersave) == FALSE) {
			DLOG_ERR("Failed to set power save");
		}

	}

	//DLOG_DEBUG("Handled icd signal, icd_state:%s", icd_state);

	return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   Check EAP DBUS authentication signal.
   @param message DBUS message.
   @return DBusHandlerResult.
*/
static DBusHandlerResult eap_check_signal_dbus(DBusMessage *message) {

        DBusError dbus_error;
	dbus_uint32_t status;

        dbus_error_init(&dbus_error);
        if (!dbus_message_get_args(message, &dbus_error,
				   DBUS_TYPE_UINT32, &status,
                                   DBUS_TYPE_INVALID ) )
        {
                DLOG_ERR("Could not get args from signal, '%s'",
                         dbus_error.message);
                dbus_error_free(&dbus_error);
                return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
        }

	if (status == EAP_SUCCESS) 
	{
		/* Set powersave if IP layer is Ok */
		if (get_wlan_state() != WLAN_NO_ADDRESS)
			if (set_power_state(wlan_status.powersave) == FALSE)
				DLOG_ERR("Setting powersave failed (wpa)");

		if (netlink_send_linkmode(WLANCOND_IF_READY) != 0)
			DLOG_ERR("Setting oper state to ready failed");

		/* Successfull connection, set retry count to 0 */
		wlan_status.retry_count = 0;
	}

        return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   Runs calibration data to WLAN firmware.
*/
gint run_calibration(void) {
	gchar *args[2];
	gint count = 0;
	args[count++] = (gchar*) CALIBRATION_APPLICATION;
	args[count++] = NULL;

	DLOG_DEBUG("Running calibration");

	if (!g_spawn_sync (NULL, args, NULL, G_SPAWN_STDOUT_TO_DEV_NULL |
			   G_SPAWN_STDERR_TO_DEV_NULL,
			   NULL, NULL, NULL, NULL, NULL, NULL)) {
		return -1;
	}

	return 0;
}
/**
   Kill supplicant.
*/
#define KILL_SUPPLICANT "/usr/bin/killall"
#define SUPPLICANT_NAME "eapd.real"
void kill_supplicant(void) {
	
	DLOG_ERR("Supplicant error");
	
	system(KILL_SUPPLICANT " -9 " SUPPLICANT_NAME);
}
static void remove_wlan_gprs_timer(void)
{
	if (wlan_gprs_timer_id) {
		g_source_remove(wlan_gprs_timer_id);
		wlan_gprs_timer_id = 0;
	}
}
/**
   WLAN GPRS timer callback.
   @param data User data.
   @return status.
*/
static gboolean wlan_gprs_cb(void* data) {

	if (wlan_gprs_timer_id) {
		wlan_gprs_timer_id = 0;

		DLOG_DEBUG("GPRS became idle");	
		wlan_status.call_type &= ~CALL_GPRS;
		set_tx_power(wlan_status.conn.power_level);
	}
	return FALSE;
}
/**
   Handle GPRS DataCounters DBUS signal.
   @param message DBUS message.
   @return DBusHandlerResult.
*/
static DBusHandlerResult gprs_datacounters_signal(DBusMessage *message) {

	if (!(wlan_status.call_type & CALL_GPRS)) {
		DLOG_DEBUG("GPRS is active");
		wlan_status.call_type |= CALL_GPRS;
		set_tx_power(wlan_status.conn.power_level);
	}
	
	remove_wlan_gprs_timer();
	
	wlan_gprs_timer_id = g_timeout_add_seconds(
		WLANCOND_GPRS_TIMEOUT,
		wlan_gprs_cb,
		NULL);
	
	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
   Check CSD DBUS signal.
   @param message DBUS message.
   @return DBusHandlerResult.
*/
static DBusHandlerResult csd_check_signal_dbus(DBusMessage *message) {

	DBusError error;
	char* network_code;
	char* country_code;
	struct timeval tv;

	dbus_error_init(&error);

	if (!dbus_message_get_args(message, &error,
				   DBUS_TYPE_STRING, &network_code,
				   DBUS_TYPE_STRING, &country_code,
				   DBUS_TYPE_INVALID))
	{
		DLOG_ERR("Could not get args from signal, '%s'",
			 error.message);
		dbus_error_free(&error);
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	DLOG_DEBUG("Handled csd signal, country code: %s", country_code);

	if (country_code == NULL || strlen(country_code) == 0) {
		if (wlan_status.country_code != NULL && 
		    strlen(wlan_status.country_code) != 0) {
			/* We lost the country information.
			   Update last valid country code timestamp */
			if (wlan_status.calibration_needed == FALSE) {
				if (gettimeofday(&tv, NULL) <0)
					wlan_status.last_code = 0;
				else
					wlan_status.last_code = tv.tv_sec;	
			}
			g_free(wlan_status.country_code);
			wlan_status.country_code = NULL;
			wlan_status.calibration_needed = TRUE;
			DLOG_INFO("Country information lost");
		}
	} else if (wlan_status.country_code == NULL || 
		   strcmp(country_code, wlan_status.country_code)) {
		/* The new code is different from the old.
		   Check if the last calibrated matches the new one.
		   Then we don't need to run calibration again.
		*/
		if (wlan_status.last_calibrated_code && country_code && 
		    strcmp(wlan_status.last_calibrated_code, 
			   country_code) == 0) {
			DLOG_DEBUG("Calibration is not needed");
			wlan_status.calibration_needed = FALSE;
		} else {
			DLOG_DEBUG("Calibration is needed");
			wlan_status.calibration_needed = TRUE;
		}
		g_free(wlan_status.country_code);
		wlan_status.country_code = g_strdup(country_code);
		/* Stop caching. If we loose the network we might be
		 caching wrong value */
		wlan_status.last_code = 0;
		DLOG_INFO("Country code changed to: %s", country_code);
	}

	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
   Check if previous country code can be used.
   @return TRUE if previous code can be used.
*/
static gboolean check_country_timeout(void) {
	struct timeval tv;

	/* Initial state after boot */
	if (wlan_status.last_code == 0)
		return FALSE;
	
	if (gettimeofday(&tv, NULL) <0) 
		return FALSE;
	
	if (tv.tv_sec > wlan_status.last_code + WLANCOND_COUNTRY_CACHE_TIME)
		return FALSE;

	return TRUE;
}

/**
   Handles WLAN country.
   @return status.
*/
static gint handle_country(void) {
	if (wlan_status.calibration_needed ||
	    wlan_status.country_code == NULL) {
		
		if (wlan_status.country_code == NULL ||
		    strlen(wlan_status.country_code) == 0) {
			/* Check if we can use cached value */
			if (check_country_timeout() == TRUE) {
				DLOG_DEBUG("Cached code in use");
				return 0;
			} else {
				DLOG_DEBUG("Cannot use cached code");
			}
		}

		/* Read the code so we can check
		   if the code changes in CSD signal.
		*/
		char* code = check_country_code();
		
		/* Run calibration if we got new code */
		if (code != NULL)
			if (run_calibration() < 0)
				DLOG_ERR("Fatal: Could not calibrate");
		
		g_free(wlan_status.country_code);
		wlan_status.country_code = code;
		g_free(wlan_status.last_calibrated_code);
		wlan_status.last_calibrated_code = g_strdup(code);
		wlan_status.calibration_needed = FALSE;
	}
	return 0;
}
/**
 * Helper function for initializing handler structs
 */
int init_dbus_handler(void)
{
	gboolean error;

	handle_country();

	if (get_own_mac() < 0) {
		DLOG_ERR("Could not get own MAC address");
		return -1;
	}

	wlan_status.allow_all_ciphers = get_setting_bool(
		WLANCOND_ALLOW_ALL_CIPHERS, &error);
	
	return 0;
}

static gboolean in_flight_mode(void) {
	return _flight_mode;
}

void set_wlan_signal(gboolean high_or_low)
{
	if (high_or_low == WLANCOND_HIGH) {
		wlan_status.signal = WLANCOND_HIGH;
		remove_roam_scan_timer();
	} else {
		wlan_status.signal = WLANCOND_LOW;
	}
}

void remove_roam_scan_timer(void)
{
	if (wlan_status.roam_scan_id) {
		g_source_remove(wlan_status.roam_scan_id);
		wlan_status.roam_scan_id = 0;
	}
}

void remove_connect_timer(void)
{
	if (wlan_connect_timer_id) {
		g_source_remove(wlan_connect_timer_id);
		wlan_connect_timer_id = 0;
	}
}

static void remove_wlan_if_timer(void)
{
        // Remove shutdown timer if exists
	if (wlan_if_down_timer_id) {
		g_source_remove(wlan_if_down_timer_id);
		wlan_if_down_timer_id = 0;
	}
}

void remove_scan_timer(void)
{
	if (wlan_status.scan_id) {
		g_source_remove(wlan_status.scan_id);
		wlan_status.scan_id = 0;
	}
}
static void remove_rescan_timer(void)
{
	if (wlan_status.rescan_id) {
		g_source_remove(wlan_status.rescan_id);
		wlan_status.rescan_id = 0;
	}
}
/**
   Check how many scans are left before giving up.
*/
gboolean scan_retries_left(void) {
	guint scan_tries = WLANCOND_MAX_SCAN_TRIES;

	/* The following condition will retry every 
	   WLANCOND_CONNECT_TIMEOUT seconds in case of an unsuccessful 
	   connection but in case of WEP it will retry every 
	   WLANCOND_CONNECT_TIMEOUT/2 seconds but 
	   WLANCOND_MAX_SCAN_TRIES*2 times to give shared key and open 
	   authentication 3,3 tries each as this is now done in user 
	   space when using netlink library. */	
	if ((wlan_status.conn.encryption & WLANCOND_ENCRYPT_METHOD_MASK) == 
	    WLANCOND_WEP)
		scan_tries = WLANCOND_MAX_SCAN_TRIES*2;

	if (++wlan_status.retry_count <= scan_tries)
		return TRUE;
	else
		return FALSE;
}
/**
   Check how many scans are left before giving up.
*/
static gboolean no_connection_and_scan_retries_left(void) {
	guint scan_tries = WLANCOND_MAX_SCAN_TRIES*2;

	if (++wlan_status.retry_count <= scan_tries)
		return TRUE;
	else
		return FALSE;
}
/**
   WLAN connect timer callback.
   @param data User data.
   @return status.
*/
static gboolean wlan_connect_timer_cb(void* data)
{
	if (wlan_connect_timer_id && get_wlan_state() ==
	    WLAN_INITIALIZED_FOR_CONNECTION) {
		
		wlan_connect_timer_id = 0;

		DLOG_DEBUG("Association timeout, try: %d",
			   wlan_status.retry_count);

		DLOG_DEBUG("Disconnecting previous connection");
		mlme_command(wlan_status.conn.bssid, 
			     WLANCOND_NL80211_MLME_DISCONNECT, 
			     WLANCOND_REASON_LEAVING);

		/* Remove the failed AP from the list */
		roam_cache_key_t key = { wlan_status.conn.bssid,
					 wlan_status.conn.channel };
		remove_from_roam_cache(&key);

		/* Set BSSID to 0 */
		memset(wlan_status.conn.bssid, 0, ETH_ALEN);

		/* Clear channel */
		wlan_status.conn.channel = 0;

		if (find_connection_and_associate(wlan_status.roam_cache,
						  FALSE, FALSE, FALSE) == 0)
			return FALSE;

		/* Try to scan again if retries left */
		if (scan_retries_left()) {
			if (scan(wlan_status.conn.ssid,
				wlan_status.conn.ssid_len, TRUE) == 0) {
				return FALSE;
			}
		}

		set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL,
			       FALSE);
		return FALSE;
	}

	wlan_connect_timer_id = 0;

	//DLOG_DEBUG("Association OK");

	return FALSE;
}

/**
   WLAN scan callback.
   @param data User data.
   @return status.
*/
static gboolean wlan_scan_cb(void* data)
{

	wlan_status.scan_id = 0;

	DLOG_ERR("Scan failed, should not happen!");

	set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL, FALSE);

	return FALSE;
}
/**
   WLAN interface down callback.
   @param data User data.
   @return status.
*/
static gboolean wlan_if_down_cb(void* data)
{

	wlan_if_down_timer_id = 0;

	if (get_wlan_state() == WLAN_NOT_INITIALIZED) {
		DLOG_DEBUG("Delayed shutdown occurred");
		set_wlan_state(WLAN_NOT_INITIALIZED, NO_SIGNAL, FALSE);
		return FALSE;
	}

	//DLOG_DEBUG("Delayed shutdown did not happen");

	return FALSE;
}
/**
   MLME command.
   @param addr Access point MAC address.
   @param cmd Command.
   @param reason_code Reason for leaving.
   @return status.
*/
int mlme_command(guchar* addr, guint16 cmd, guint16 reason_code)
{
	gint ret;

	if (!wlan_status.ifindex)
		return -2;

	/* This following check for adhoc is because in netlink mlme_command
	   is not supported for adhoc networks, mlme command returns not 
	   supported error in case of adhoc. */

	if (wlan_status.conn.mode & WLANCOND_ADHOC) {
		ret = nl80211_leave_adhoc(wlan_status.ifindex);
		if (ret) {
			DLOG_ERR("nl80211: Failed to Leave IBSS");
			return -1;
		}
	} else {
		ret = nl80211_mlme_command(addr, cmd, reason_code, 
					   wlan_status.ifindex);
		if (ret) {
			DLOG_ERR("nl80211: Failed to run MLME command, ret=%d",
				ret);
			return -1;
		}
	}

	return 0;
}
/**
   Helper function for setting the operating mode.
   @param mode operating mode.
   @return status.
*/
static int set_mode(guint32 mode)
{
	int ret;
	guint32 threshold;

	ret = nl80211_set_op_mode(wlan_status.ifindex, mode);

	if (ret) {
		DLOG_ERR("nl80211 - Operating mode setting failed");
		return -1;
	}

	if (get_inactivity_status() == TRUE)
		threshold = WLANCOND_CQM_THRESHOLD_IDLE;
	else
		threshold = WLANCOND_CQM_THRESHOLD_ACTIVE;
	
	if (nl80211_set_cqm(threshold, WLANCOND_CQM_HYSTERESIS)) {
		DLOG_ERR("nl80211 - CQM monitor setting failed");
		return -1;
	}

	return 0;
}
/**
   Remove all timers.
*/
static void remove_timers(void) {
		/* Remove association timer */
		remove_connect_timer();

		/* Remove scan timer */
		remove_scan_timer();

		/* Remove rescan timer */
		remove_rescan_timer();

		// Remove shutdown timer if exists
		remove_wlan_if_timer();
		
		// Remove GPRS timer */
		remove_wlan_gprs_timer();
}

static void remove_connection_time_filters(void) {
	if (wlan_status.dbus_filters == FALSE)
		return;
		
	remove_icd_listener(get_dbus_connection());
#ifdef ACTIVITY_CHECK
	remove_activity_listener(get_dbus_connection());
#endif
	if (get_wpa_mode() == TRUE)
		remove_eap_listener(get_dbus_connection());

	remove_gprs_listener(get_dbus_connection());
	
	wlan_status.dbus_filters = FALSE;
}

/**
    Helper function for setting new wlan state
    @param new_state New state for WLAN.
    @param send_signal Should signal be sent.
    @param delay_shutdown Whether to delay shutdown or not.
*/
void set_wlan_state(int new_state, int send_signal, gboolean delay_shutdown)
{
	const char *status_table[] =
	{
		(char*)"WLAN_NOT_INITIALIZED",
		(char*)"WLAN_INITIALIZED",
		(char*)"WLAN_INITIALIZED_FOR_SCAN",
		(char*)"WLAN_INITIALIZED_FOR_CONNECTION",
		(char*)"WLAN_NO_ADDRESS",
		(char*)"WLAN_CONNECTED"
	};

	switch (new_state) {

	case WLAN_NOT_INITIALIZED:
		if (get_wlan_state() == WLAN_CONNECTED ||
		    get_wlan_state() == WLAN_NO_ADDRESS) {
			mlme_command(wlan_status.conn.bssid,
				     WLANCOND_NL80211_MLME_DISCONNECT,
				     WLANCOND_REASON_LEAVING);
		}

		remove_iptables();

		set_scan_state(SCAN_NOT_ACTIVE);

		if (get_wlan_state() != WLAN_NOT_INITIALIZED &&
		    get_wlan_state() != WLAN_INITIALIZED_FOR_SCAN)
			clear_wpa_mode();

		wlan_status.retry_count = 0;
		wlan_status.roam_scan = WLANCOND_MIN_ROAM_SCAN_INTERVAL;
		wlan_status.ip_ok = FALSE;
		wlan_status.last_scan = 0;

		remove_timers();

		set_wlan_signal(WLANCOND_HIGH);

		if (delay_shutdown == FALSE) {

			clean_roam_cache();

			if (set_interface_state(socket_open(), CLEAR,
						IFF_UP)<0) {
				DLOG_ERR("Could not set interface down");
			}

		} else {

			DLOG_DEBUG("Delaying interface shutdown");

			wlan_if_down_timer_id = g_timeout_add_seconds(
					WLANCOND_SHUTDOWN_DELAY,
					wlan_if_down_cb,
					NULL);

		}

		remove_connection_time_filters();

		if (send_signal == DISCONNECTED_SIGNAL)
			disconnected_signal();

		break;
	case WLAN_INITIALIZED_FOR_CONNECTION:
		/*
		   Set BSSID to 0, this state happens e.g when we drop from
		   the network
		*/
		memset(wlan_status.conn.bssid, 0, ETH_ALEN);

		/* Clear channel */
		wlan_status.conn.channel = 0;

		/* Remove association timer */
		remove_connect_timer();

		set_power_state(WLANCOND_POWER_ON);

		break;
	case WLAN_CONNECTED:

		/* With WPA we get signal when we can go to powersave */
		if (get_wpa_mode() == FALSE)
 			set_power_state(wlan_status.powersave);

		wlan_status.ip_ok = TRUE;
		break;
	default:
		break;
	}
	DLOG_DEBUG("Wlancond state change, old_state: %s, new_state: %s",
		status_table[wlan_status.state], status_table[new_state]);
	wlan_status.state = new_state;
}

/**
    Helper function for getting the wlan state.
    @return state.
*/
guint get_wlan_state(void)
{
	return wlan_status.state;
}

/**
    Helper function for setting new scan state.
    @param new_state New scan state.
*/
void set_scan_state(guint new_state)
{
	if (wlan_status.scan == new_state) {
		return;
	}
	if (new_state == SCAN_NOT_ACTIVE) {
		remove_scan_timer();
		if (wlan_status.scan == SCAN_ACTIVE &&
		    scan_name_cache != NULL) {
			DLOG_DEBUG("Sending empty results");
			send_dbus_scan_results(NULL, scan_name_cache, 0);
			g_free(scan_name_cache);
			scan_name_cache = NULL;
		}
	}

	DLOG_DEBUG("Wlancond scan change, old_state: %s, new_state: %s",
		   wlan_status.scan==SCAN_NOT_ACTIVE ? "SCAN_IDLE":"SCANNING",
		   new_state == SCAN_NOT_ACTIVE ? "SCAN_IDLE":"SCANNING");
	wlan_status.scan = new_state;
}

/**
    Helper function for getting the scan state.
    @return state.
*/
guint get_scan_state(void)
{
	return wlan_status.scan;
}
/**
    Helper function for getting the wlan mode.
    @return mode.
*/
guint get_mode(void)
{
	return wlan_status.conn.mode;
}

/**
   Set network latency.
   @param latency New latency.
*/
static void set_network_latency(int latency) {

	DLOG_DEBUG("Setting network latency to %i", latency);

	if (wlan_status.latency_file == NULL) {
		wlan_status.latency_file = fopen(WLANCOND_LATENCY_FILE, "w");
		if (wlan_status.latency_file == NULL) {
			DLOG_ERR("Cannot open: %s", WLANCOND_LATENCY_FILE);
			return;
		}
	}
	
	if (fwrite(&latency, sizeof(int), 1, wlan_status.latency_file) 
	    != 1) {
		DLOG_ERR("Could not write to: %s", WLANCOND_LATENCY_FILE);
	}
	
	fflush(wlan_status.latency_file);

	return;
}

gboolean can_adjust_psm(void) {
	/* Make sure we are properly connected, IP address OK and
	   WPA auth OK */
	if (get_wlan_state() == WLAN_CONNECTED &&
	    wlan_status.ip_ok == TRUE && 
	    wlan_status.retry_count == 0)
		return TRUE;
	
	return FALSE;
}
static gint compare_scan_entry(gconstpointer a, gconstpointer b)
{
	const scan_results_t *scan_entry = a;
	const roam_cache_key_t *key = b;
	int ret;
	if ((ret = memcmp(scan_entry->bssid, key->bssid, ETH_ALEN)))
		return ret;
	else
		return scan_entry->channel - key->channel;
}

/**
   Check if given AP supports U-APSD.
   @param mac MAC address.
   @return TRUE if U-APSD supported.
*/
static gboolean is_uapsd_supported(unsigned char* mac) {

	GSList *list;
	unsigned char *ap_mac = mac;

	/* If no MAC is supplied use the current AP */
	if (ap_mac == NULL)
		ap_mac = wlan_status.conn.bssid;

	roam_cache_key_t key = { ap_mac, wlan_status.conn.channel };
	list = g_slist_find_custom(wlan_status.roam_cache,
                                   &key,
                                   &compare_scan_entry);
	if (list == NULL)
		return FALSE;

	struct scan_results_t *entry = list->data;
	if (entry->extra_cap_bits & WLANCOND_UAPSD_SUPPORTED)
		return TRUE;

	return FALSE;
}

/**
   Request WLAN power state.
   @param state New power state.
   @return status.
*/
gboolean set_power_state(guint new_state)
{
	gboolean power_state;

	if (new_state == WLANCOND_CAM_TIMEOUT) {
		if ((wlan_status.call_type & CALL_VOIP) &&
		    is_uapsd_supported(NULL)) {
			DLOG_DEBUG("Full PSM because of VOIP and U-APSD");
			new_state = WLANCOND_FULL_POWERSAVE;
		} else if (wlan_status.full_psm_enabled) {
			/* If full psm is enabled in the settings */
			new_state = WLANCOND_FULL_POWERSAVE;
		} else if (wlan_status.high_traffic == FALSE &&
			   can_adjust_psm() == TRUE &&
			   get_inactivity_status() == TRUE) {
			/* If the traffic is low and the user is inactive,
			   let's go to full PSM
			*/
			new_state = WLANCOND_FULL_POWERSAVE;
		}
	}

	if (wlan_status.power == new_state)
		return TRUE;

	switch (new_state) {
	case WLANCOND_POWER_ON:
		power_state = FALSE;
		break;
	case WLANCOND_CAM_TIMEOUT:
		set_network_latency(WLANCOND_DEFAULT_NETWORK_LATENCY);
		power_state = TRUE;
		break;
	case WLANCOND_FULL_POWERSAVE:
		set_network_latency(WLANCOND_FULL_PSM_LATENCY);
		power_state = TRUE;
		break;
	default:
		power_state = TRUE;
		break;
	}

	if (nl80211_set_power_save(power_state, wlan_status.ifindex) != 0) {
		DLOG_ERR("set power failed, state: %s", 
			 power_state?"true":"false");
		return FALSE;
	}
	
	wlan_status.power = new_state;

	DLOG_DEBUG("New power state set: %i", new_state);

	return TRUE;
}

/**
   Initialize the WLAN interface.
   @param sock socket.
   @return previous state.
*/
static int init_if(int sock)
{
	if (!wlan_status.ifindex)
		return -1;

	int previous_state = get_wlan_state();

	if (previous_state == WLAN_NOT_INITIALIZED) {
		/* Check if interface is still up from delayed shutdown */
		if (wlan_status.interface_up == FALSE) {

			/* Start from the beginning */
			if (handle_country() < 0)
				return -1;

			if (set_interface_state(sock, SET, 
						IFF_UP | IFF_RUNNING) < 0) {
				return -1;
			}
		}

		set_power_state(WLANCOND_POWER_ON);
		set_wlan_state(WLAN_INITIALIZED, NO_SIGNAL, FALSE);
	}

	return previous_state;
}

/**
 * Save the name of the wireless device.
 */
void save_device_interface(void)
{
	strncpy(wlan_status.ifname, WLANCOND_IF_NAME, IFNAMSIZ);
	wlan_status.ifname[IFNAMSIZ] = '\0';
	wlan_status.ifindex = if_nametoindex(wlan_status.ifname);
	DLOG_DEBUG("WLAN interface Index %u", wlan_status.ifindex);
}

/**
   Set interface state.
   @param socket socket.
   @param dir direction.
   @param flags settings flags.
   @return status.
*/
int set_interface_state(int sock, int dir, short flags)
{
	struct ifreq ifr;

	memset(&ifr, 0, sizeof(ifr));

	strncpy(ifr.ifr_name, wlan_status.ifname, IFNAMSIZ);

	if (ioctl(sock, SIOCGIFFLAGS, &ifr) < 0) {
		DLOG_ERR("Could not get interface %s flags",
			 wlan_status.ifname);
		return -1;
	}
	if (dir == SET) {
		if (!add_context_listener(get_dbus_connection()))
			DLOG_ERR("Adding context listener failed");

		wlan_status.call_type = get_call_state();

		ifr.ifr_flags |= flags;
	} else {
		remove_context_listener(get_dbus_connection());

		wlan_status.if_down_self = TRUE;
		ifr.ifr_flags &= ~flags;
	}

	if (ioctl(sock, SIOCSIFFLAGS, &ifr) < 0) {
		DLOG_ERR("Could not set interface %s flags",
			 wlan_status.ifname);
		return -1;
	}
	
	wlan_status.interface_up = (dir == SET ? TRUE:FALSE);
	DLOG_DEBUG("%s is %s", wlan_status.ifname, 
		   wlan_status.interface_up == TRUE ? "UP":"DOWN");

	return 0;
}

/**
    Set tx power level.
    @param power Power level.
    @param sock socket.
    @return status.
*/
gboolean set_tx_power(guint power)
{
	int tx_power;

	if (wlan_status.call_type & (CALL_CELLULAR | CALL_GPRS)) {
		if (power == WLANCOND_TX_POWER10)
			tx_power = WLANCOND_TX_POWER10DBM;
		else
			tx_power = WLANCOND_TX_POWERLIMITED; 
	} else {
		if (power == WLANCOND_TX_POWER10) {
			tx_power = WLANCOND_TX_POWER10DBM;
		} else if (power == WLANCOND_TX_POWER100) {
			tx_power = WLANCOND_TX_POWER100DBM;
		} else {
			return FALSE;
		}
	}

	if (nl80211_set_tx_power(tx_power, wlan_status.ifindex) != 0) {
		DLOG_ERR("set tx power failed");
		return FALSE;
	}

	return TRUE;
}
/**
   Updates algorithms to internal status.
   @param encryption Encryption settings.
   @param scan_results Scan results.
   @return status.
*/
static int update_algorithms(guint32 encryption, 
			     struct scan_results_t *scan_results)
{
	wlan_status.group_cipher = 0;
	wlan_status.pairwise_cipher = 0;

	// Open
	if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_OPEN) {
		DLOG_DEBUG("Open mode");
		wlan_status.pairwise_cipher = CIPHER_SUITE_NONE;
		return 0;
	}

	// WEP
        if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WEP) {
		DLOG_DEBUG("WEP enabled");
		wlan_status.pairwise_cipher = CIPHER_SUITE_WEP40;
		return 0;
	}

	// WPA modes
	if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_PSK ||
	    (encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_EAP) {
		if ((encryption & WLANCOND_ENCRYPT_ALG_MASK)
		    == WLANCOND_WPA_TKIP) {
			DLOG_DEBUG("TKIP Selected for unicast");
			wlan_status.pairwise_cipher = CIPHER_SUITE_TKIP;
		} else if ((encryption & WLANCOND_ENCRYPT_ALG_MASK) ==
			   WLANCOND_WPA_AES) {
			DLOG_DEBUG("AES selected for unicast");
			wlan_status.pairwise_cipher = CIPHER_SUITE_CCMP;
		} else if (wlan_status.allow_all_ciphers == TRUE) {
			if (scan_results->extra_cap_bits & WLANCOND_WEP104) {
				DLOG_DEBUG("WEP104 selected for unicast");
				wlan_status.pairwise_cipher = CIPHER_SUITE_WEP104;
			} else if (scan_results->extra_cap_bits & WLANCOND_WEP40) {
				DLOG_DEBUG("WEP40 selected for unicast");
				wlan_status.pairwise_cipher = CIPHER_SUITE_WEP40;
			} else {
				DLOG_ERR("Not supported encryption %08x", encryption);
				return -1;
			}
		} else {
			DLOG_ERR("Not supported encryption %08x", encryption);
			return -1;
		}

		if ((encryption & WLANCOND_ENCRYPT_GROUP_ALG_MASK) ==
		    WLANCOND_WPA_TKIP_GROUP) {
			DLOG_DEBUG("TKIP Selected for group key");
			wlan_status.group_cipher = CIPHER_SUITE_TKIP;
		} else if ((encryption & WLANCOND_ENCRYPT_GROUP_ALG_MASK) ==
			   (unsigned int)WLANCOND_WPA_AES_GROUP) {
			DLOG_DEBUG("AES Selected for group key");
			wlan_status.group_cipher = CIPHER_SUITE_CCMP;

		} else if (wlan_status.allow_all_ciphers == TRUE) {

			if (scan_results->extra_cap_bits & WLANCOND_WEP104_GROUP) {
				DLOG_DEBUG("WEP104 selected for group key");
				wlan_status.group_cipher = CIPHER_SUITE_WEP104;
			} else if (scan_results->extra_cap_bits & WLANCOND_WEP40_GROUP) {
				DLOG_DEBUG("WEP40 selected for group key");
				wlan_status.group_cipher = CIPHER_SUITE_WEP40;
			} else {
				DLOG_ERR("Not supported encryption %08x", encryption);
				return -1;
			}
		} else {
			DLOG_ERR("Not supported encryption %08x", encryption);
			return -1;
		}
	} else if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == 
		   WLANCOND_WAPI_PSK ||
		   (encryption & WLANCOND_ENCRYPT_METHOD_MASK) == 
		   WLANCOND_WAPI_CERT) {
		if ((encryption & WLANCOND_ENCRYPT_ALG_MASK)
		    == WLANCOND_WAPI_SMS4) {
			DLOG_DEBUG("SMS4 Selected for unicast");
			wlan_status.pairwise_cipher = CIPHER_SUITE_SMS4;
		} else {
			DLOG_ERR("Not supported encryption %08x", encryption);
			return -1;
		}

		if ((encryption & WLANCOND_ENCRYPT_GROUP_ALG_MASK) ==
		    WLANCOND_WAPI_SMS4_GROUP) {
			DLOG_DEBUG("SMS4 Selected for multicast");
			wlan_status.group_cipher = CIPHER_SUITE_SMS4;
		} else {
			DLOG_ERR("Not supported encryption %08x", encryption);
			return -1;
		}
	}	

	return 0;
}
/**
   Clean roaming cache.
*/
void clean_roam_cache(void)
{
	clean_scan_results(&wlan_status.roam_cache);
}

/**
   Clear WPA mode related stuff.
*/
void clear_wpa_mode(void)
{
	update_own_ie(NULL, 0);

	wlan_status.pairwise_cipher = CIPHER_SUITE_NONE;
	wlan_status.group_cipher = CIPHER_SUITE_NONE;

	/* Clean PMK cache */
	g_slist_foreach(wlan_status.pmk_cache, (GFunc)g_free, NULL);
	g_slist_free(wlan_status.pmk_cache);
	wlan_status.pmk_cache = NULL;
}
/**
   Check if WPA or WAPI mode is in use.
   @return TRUE if mode is in use.
*/
gboolean get_wpa_mode(void)
{
	dbus_int32_t encryption = wlan_status.conn.encryption;
 
	if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_PSK ||
	    (encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_EAP ||
	    (encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_PSK ||
	    (encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_CERT) { 
		return TRUE;
	}
	return FALSE;
}

static gint compare_pmk_entry(gconstpointer a, gconstpointer b)
{
	const struct pmksa_cache_t *pmk_cache = a;

	return memcmp(pmk_cache->mac, b, ETH_ALEN);
}

/**
   Add PMKID to PMKSA cache.
   @param pmkid PMKID to add.
   @param mac MAC address associated to PMKID.
*/
static void add_to_pmksa_cache(unsigned char* pmkid, unsigned char* mac)
{
	guint i = 0;
	GSList *list;
	gboolean entry_found = FALSE;

	for (list = wlan_status.pmk_cache; list != NULL &&
		     entry_found == FALSE; list = list->next) {
		struct pmksa_cache_t *pmk_cache = list->data;
		/* First find if we have already the cache entry */
		if (memcmp(pmk_cache->mac, mac, ETH_ALEN) == 0) {
			DLOG_DEBUG("Found old entry: %i", i);
			/* Remove the old entry */
			wlan_status.pmk_cache = g_slist_remove(
				wlan_status.pmk_cache, pmk_cache);
			g_free(pmk_cache);

			entry_found = TRUE;
		} else {
			i++;
		}
	}

	if (i == PMK_CACHE_SIZE) {
		DLOG_DEBUG("Cache full, remove oldest");
		GSList *last_entry = g_slist_last(wlan_status.pmk_cache);
		wlan_status.pmk_cache = g_slist_remove(wlan_status.pmk_cache,
						       last_entry->data);
		g_free(last_entry->data);
	}
	print_mac(WLANCOND_PRIO_LOW, "Adding new entry:", mac);

	struct pmksa_cache_t *new_entry = g_malloc(sizeof(*new_entry));
	memcpy(new_entry->mac, mac, ETH_ALEN);
	memcpy(new_entry->pmkid, pmkid, WLANCOND_PMKID_LEN);

	wlan_status.pmk_cache = g_slist_prepend(wlan_status.pmk_cache,
						new_entry);
	return;
}

/**
 * Removes an entry with the given mac address from the PMKSA cache.
 *
 * @param mac The mac address to be removed.
 * @return status True if the entry was actually removed.
 */
gboolean remove_from_pmksa_cache(unsigned char* mac)
{
	GSList *list = g_slist_find_custom(wlan_status.pmk_cache, mac,
					   &compare_pmk_entry);
	if(!list)
		return FALSE;

	struct pmksa_cache_t *entry = list->data;

	print_mac(WLANCOND_PRIO_MEDIUM, "Removing PMKSA entry for:", mac);

	wlan_status.pmk_cache = g_slist_remove(wlan_status.pmk_cache,
					       entry);

	g_free(entry);

	return TRUE;
}

/**
   Find entry from PMKSA cache.
   @param mac MAC address to identify the entry.
   @param pmkid This pointer will be set to point the pmkid. This won't be
   set on errors.
   @return Zero on success, non-zero on errors.
*/
int find_pmkid_from_pmk_cache(unsigned char* mac,
			      unsigned char **pmkid)
{
	GSList *list;
	int pmksa_found;

	if (check_pmksa_cache((unsigned char *)wlan_status.own_mac, ETH_ALEN,
			      mac, ETH_ALEN,
			      wlan_status.conn.authentication_type,
			      wlan_status.pairwise_cipher,
			      wlan_status.group_cipher,
			      &pmksa_found))
	{
		DLOG_ERR("Error while querying the pmksa cache status "
			 "from eapd");
		return -1;
	}

	if(!pmksa_found) {
		DLOG_DEBUG("No cached pmksa found from eapd");

		remove_from_pmksa_cache(mac);
		*pmkid = NULL;

		return 0;
	}

	list = g_slist_find_custom(wlan_status.pmk_cache, mac, &compare_pmk_entry);
	if (list != NULL) {
		struct pmksa_cache_t *pmk_cache = list->data;
		print_mac(WLANCOND_PRIO_MEDIUM, "Found PMKSA entry for:", mac);
		*pmkid = pmk_cache->pmkid;
		return 0;
	}

	DLOG_DEBUG("No cached pmksa found from eapd");
	*pmkid = NULL;

	return 0;
}
static void rescan_destroy_cb(gpointer data)
{
	g_free(data);
}
/**
   Helper function for rescanning from a timer.
   @param data Optional scanning data.
   @return status.
*/
#define MAX_SCAN_COUNT 3
static gboolean rescan(void* data)
{
	int ret;

	if (data) {
		struct scan_cb_data *scan_data = data;
		ret = nl80211_scan(&wlan_status, scan_data->ssid_len, 
				   scan_data->ssid);
		if (ret == -EBUSY) { 
			if (scan_data->count++ >= MAX_SCAN_COUNT) {
				set_wlan_state(WLAN_NOT_INITIALIZED, 
					       DISCONNECTED_SIGNAL, FALSE);	
			} else {
				DLOG_ERR("Scan is busy, rescanning later");
				return TRUE;
			}
		} else if (ret != 0) {
			DLOG_ERR("Scan failed in rescan");
			set_wlan_state(WLAN_NOT_INITIALIZED, 
				       DISCONNECTED_SIGNAL, FALSE);
		}
	} else if (get_wlan_state() == WLAN_INITIALIZED_FOR_CONNECTION) {
		scan(wlan_status.scan_ssid, wlan_status.scan_ssid_len, TRUE);
	}
	wlan_status.rescan_id = 0;

	return FALSE;
}
/**
   Helper function for scanning.
   @param ssid SSID to scan.
   @param ssid_len SSID length.
   @return status.
*/
int scan(gchar *ssid, int ssid_len, gboolean add_timer)
{
	int ret;
 
	if (get_scan_state() == SCAN_ACTIVE)
		return 0;

	set_scan_state(SCAN_ACTIVE);

	ret = nl80211_scan(&wlan_status, ssid_len - 1, ssid);

	if (ret == -EBUSY) {
		DLOG_ERR("Scan is busy, rescanning later");
		struct scan_cb_data *scan_data = g_malloc(
			sizeof(struct scan_cb_data));
		memcpy(scan_data->ssid, ssid, ssid_len);
		scan_data->ssid_len = ssid_len -1;
		scan_data->count = 1;
		wlan_status.rescan_id = g_timeout_add_seconds_full(
			G_PRIORITY_DEFAULT,
			WLANCOND_RESCAN_DELAY,
			rescan,
			scan_data,
			rescan_destroy_cb);
	} else if (ret != 0) {
		DLOG_ERR("Scan command failed: %d", ret);
		set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL, 
			       FALSE);
		return -1;
	}

	if (ssid != wlan_status.scan_ssid) {
		memset(wlan_status.scan_ssid, 0, sizeof(wlan_status.scan_ssid));
		wlan_status.scan_ssid_len = ssid_len;
		
		if (ssid && ssid_len > 1)
			memcpy(wlan_status.scan_ssid, ssid, ssid_len);
	}
	if (add_timer == TRUE) {
		wlan_status.scan_id = g_timeout_add_seconds(
			WLANCOND_SCAN_TIMEOUT,
			wlan_scan_cb,
			NULL);
	}

	DLOG_INFO("Scan issued");

	return 0;
}

static void init_conn_params(struct connect_params_t *conn_params)
{
	memset(conn_params, 0, sizeof(*conn_params));
}

static void add_connection_time_filters(void) {

	if (wlan_status.dbus_filters == TRUE)
		return;

	if (!add_icd_listener(get_dbus_connection()))
		DLOG_ERR("Adding icd listener failed");

#ifdef ACTIVITY_CHECK
	if (!add_activity_listener(get_dbus_connection()))
		DLOG_ERR("Adding activity listener failed");

	saved_inactivity = get_activity();
#endif
	if (get_wpa_mode() == TRUE)
		if (!add_eap_listener(get_dbus_connection()))
			DLOG_ERR("Adding eap listener failed");

	if (!add_gprs_listener(get_dbus_connection()))
		DLOG_ERR("Adding GPRS listener failed");
	
	wlan_status.dbus_filters = TRUE;
}

/**
   Helper function for checking settings_and_connect DBUS parameters.
   @param conn Connection parameters.
   @param ssid SSID.
   @param key Encryption keys.
   @return status.
*/
static int check_connect_arguments(struct connect_params_t *conn, char* ssid,
                                   unsigned char** key)
{
	guint i;

	if (conn->flags & WLANCOND_DISABLE_POWERSAVE) {
		DLOG_DEBUG("Powersave disabled");
		wlan_status.powersave = WLANCOND_POWER_ON;
	} else {
		wlan_status.powersave = WLANCOND_CAM_TIMEOUT;
	}

	if (conn->power_level != WLANCOND_TX_POWER10 &&
	    conn->power_level != WLANCOND_TX_POWER100) {
		DLOG_ERR("Invalid power level");
		return -1;
	}

	switch (conn->mode) {
	case WLANCOND_ADHOC:
	case WLANCOND_INFRA:
		break;
	default:
		DLOG_ERR("Operating mode undefined");
		return -1;
	}
	/* Encryption settings */
	guint32 wpa2_mode = conn->encryption & WLANCOND_ENCRYPT_WPA2_MASK;

	DLOG_DEBUG("Encryption setting: %08x", conn->encryption);
	
	switch (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) {
	case WLANCOND_OPEN:
		break;
	case WLANCOND_WEP:
		wlan_status.conn.wep_try_open_mode_first = TRUE;
		break;
	case WLANCOND_WPA_PSK:
		DLOG_DEBUG("%s PSK selected",
			   wpa2_mode!=0?"WPA2":"WPA");
		if (wpa2_mode != 0)
			conn->authentication_type = EAP_AUTH_TYPE_WPA2_PSK;
		else
			conn->authentication_type = EAP_AUTH_TYPE_WPA_PSK;
		break;
	case WLANCOND_WPA_EAP:
		DLOG_DEBUG("%s EAP selected", wpa2_mode!=0?"WPA2":"WPA");
		if (wpa2_mode != 0)
			conn->authentication_type = EAP_AUTH_TYPE_WPA2_EAP;
		else
			conn->authentication_type = EAP_AUTH_TYPE_WPA_EAP;
		break;
	case WLANCOND_WAPI_PSK:
		DLOG_DEBUG("WAPI PSK selected");
		conn->authentication_type = EAP_AUTH_TYPE_WAPI_PSK;
		break;	
	case WLANCOND_WAPI_CERT:
		DLOG_DEBUG("WAPI CERT selected");
		conn->authentication_type = EAP_AUTH_TYPE_WAPI_CERT;
		break;
	default:
		DLOG_DEBUG("Unsupported encryption mode");
		return -1;
	}
	if ((conn->encryption & WLANCOND_WPS_MASK) != 0) {
		DLOG_DEBUG("WPS selected");
		conn->authentication_type = EAP_AUTH_TYPE_WFA_SC;
	}

	if (!ssid || conn->ssid_len <= 1 ||
	    conn->ssid_len > WLANCOND_MAX_SSID_SIZE + 1) {
		DLOG_DEBUG("Invalid SSID");
		return -1;
	}

	for (i = 0; i < 4; i++) {
		/* Only key lengths 5 and 13 are valid */
		if (conn->key_len[i] == WLANCOND_MIN_KEY_LEN || 
		    conn->key_len[i] == WLANCOND_MAX_KEY_LEN) {
			DLOG_DEBUG("Found key %d", i);
			memcpy(&conn->key[i][0], key[i], conn->key_len[i]);
		} else {
			conn->key_len[i] = 0;
		}
	}
	return 0;
}

/**
    Settings and connect D-BUS request.
    @param message DBUS message.
    @param connection DBUS connection.
    @return status.
*/
static DBusHandlerResult settings_and_connect_request(
	DBusMessage    *message,
	DBusConnection *connection) {

	DBusMessage *reply = NULL;
	DBusError derror;
	struct connect_params_t *conn;
	char *ssid;
	unsigned char* key[4];
	dbus_int32_t old_mode;
	int res;
	gboolean autoconnect = FALSE;
	GSList *scan_results = NULL;		

	dbus_error_init(&derror);

	if (in_flight_mode()) {
		reply = new_dbus_error(message, WLANCOND_ERROR_WLAN_DISABLED);
		send_and_unref(connection, reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	conn = &wlan_status.conn;
	/* Save previous mode */
	old_mode = conn->mode;
	init_conn_params(conn);

	if (dbus_message_get_args(
		    message, NULL,
		    DBUS_TYPE_INT32, &conn->power_level,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &ssid, &conn->ssid_len,
		    DBUS_TYPE_INT32, &conn->mode,
		    DBUS_TYPE_INT32, &conn->encryption,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &key[0], &conn->key_len[0],
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &key[1], &conn->key_len[1],
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &key[2], &conn->key_len[2],
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &key[3], &conn->key_len[3],
		    DBUS_TYPE_INT32, &conn->default_key,
		    DBUS_TYPE_UINT32, &conn->adhoc_channel,
		    DBUS_TYPE_UINT32, &conn->flags,
		    DBUS_TYPE_INVALID) == FALSE)
	{
		/* Try without flags */
		if (dbus_message_get_args(
			    message, &derror,
			    DBUS_TYPE_INT32, &conn->power_level,
			    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &ssid,
			    &conn->ssid_len,
			    DBUS_TYPE_INT32, &conn->mode,
			    DBUS_TYPE_INT32, &conn->encryption,
			    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
			    &key[0], &conn->key_len[0],
			    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
			    &key[1], &conn->key_len[1],
			    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
			    &key[2], &conn->key_len[2],
			    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
			    &key[3], &conn->key_len[3],
			    DBUS_TYPE_INT32, &conn->default_key,
			    DBUS_TYPE_UINT32, &conn->adhoc_channel,
			    DBUS_TYPE_INVALID) == FALSE) {
			
			DLOG_ERR("Failed to parse setting_and_connect: %s",
				 derror.message);
			dbus_error_free(&derror);
			goto param_err;
		}
	}

	if (check_connect_arguments(conn, ssid, key) < 0)
		goto param_err;

	set_power_state(WLANCOND_POWER_ON);

	/* If we change mode, do it when interface is down.
	   Also put interface down in WPS mode to clear old
	   scan results.
	*/
	if (old_mode != conn->mode ||
	    conn->encryption & WLANCOND_WPS_PUSH_BUTTON) {
		set_wlan_state(WLAN_NOT_INITIALIZED, NO_SIGNAL, FALSE);
	}
	if (nl80211_set_channel(wlan_status.ifindex, 
				WLANCOND_INITIAL_CHANNEL) < 0)
		DLOG_ERR("Changing channel failed");

	/* Mode */
	if (set_mode(conn->mode) < 0) {
		/* Restore old mode */
		conn->mode = old_mode;
                goto param_err;
	}

	if (init_if(socket_open()) < 0) {
		reply = new_dbus_error(message, WLANCOND_ERROR_INIT_FAILED);
		goto param_err;
	}

	remove_wlan_if_timer();

	if (set_tx_power(conn->power_level) != TRUE) {
		reply = new_dbus_error(message, WLANCOND_ERROR_IOCTL_FAILED);
		goto param_err;
	}
	
	if (conn->flags & WLANCOND_AUTOCONNECT) {
		DLOG_DEBUG("Autoconnect attempt");
		autoconnect = TRUE;
	}

	memcpy(conn->ssid, ssid, conn->ssid_len);

	if (get_wlan_state() == WLAN_CONNECTED ||
	    get_wlan_state() == WLAN_NO_ADDRESS) {
		DLOG_DEBUG("Disconnecting previous connection");
		mlme_command(wlan_status.conn.bssid, 
			     WLANCOND_NL80211_MLME_DISCONNECT, 
			     WLANCOND_REASON_LEAVING);
	}

	set_scan_state(SCAN_NOT_ACTIVE);
	set_wlan_state(WLAN_INITIALIZED_FOR_CONNECTION, NO_SIGNAL, TRUE);

	/* Make a broadcast scan to catch all WPS PBC registrars */
	if (conn->encryption & WLANCOND_WPS_PUSH_BUTTON) {
		DLOG_DEBUG("Broadcast scan for WPS");
		if (scan(NULL, 0, TRUE) < 0) {
			goto param_err;
		}
	} else {
		/* Try if our own cache has results */
		if ((res = find_connection_and_associate(
			     wlan_status.roam_cache,
			     FALSE, FALSE, autoconnect)) != 0) {
			
			/* If res == ETOOWEAKAP, all APs were too weak for
			   autoconnection, no need to continue */
			if (res == ETOOWEAKAP || res == ESUPPLICANT)
				goto param_err;
			
			/* Try if mac80211 has cached results */
			DLOG_DEBUG("Checking mac80211 cache...");

			wlan_status.scan_ssid_len = conn->ssid_len;
			memcpy(wlan_status.scan_ssid, ssid, conn->ssid_len);

			if (!nl80211_send_get_scan_results(&wlan_status, &scan_results))
				DLOG_DEBUG("No cached scan results");

			if ((res = find_connection_and_associate(
				     scan_results, 
				     TRUE, FALSE, autoconnect)) != 0) {
				
				/* If res == ETOOWEAKAP, all APs were too weak 
				   for autoconnection, no need to continue */
				if (res == ETOOWEAKAP || res == ESUPPLICANT)
					goto param_err;

				clean_scan_results(&scan_results);
				if (scan(conn->ssid, conn->ssid_len, TRUE) < 0)
					goto param_err;
			}
		}
	}

	clean_scan_results(&scan_results);
	g_free(connect_name_cache);
	connect_name_cache = g_strdup(dbus_message_get_sender(message));

	add_connection_time_filters();

	reply = new_dbus_method_return(message);

	gchar* ifname = wlan_status.ifname;

	append_dbus_args(reply,
			DBUS_TYPE_STRING, &ifname,
			DBUS_TYPE_INVALID);

	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;

param_err:
	clean_scan_results(&scan_results);
	if (reply == NULL) {
		DLOG_DEBUG("Parameter error in settings_and_connect");
		reply = new_dbus_error(message, DBUS_ERROR_INVALID_ARGS);
	}
	send_and_unref(connection, reply);
	set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL, TRUE);

	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
    Associate function to associate to selected access point.
    @param scan_results Scan results.
    @return status.
*/
int associate(struct scan_results_t *scan_results)
{
	struct connect_params_t *conn = &wlan_status.conn;
	gint ret;
	
	DLOG_INFO("Starting to associate");

	guint connection_timeout = WLANCOND_CONNECT_TIMEOUT;

	if (get_wlan_state() == WLAN_CONNECTED) {
		DLOG_DEBUG("Disconnecting previous connection");
		mlme_command(conn->bssid, WLANCOND_NL80211_MLME_DISCONNECT, 
			     WLANCOND_REASON_LEAVING);
		if (get_wpa_mode() == TRUE)
			disassociate_eap();
		set_wlan_state(WLAN_INITIALIZED_FOR_CONNECTION, NO_SIGNAL, 
			       TRUE);
	}

	if (update_algorithms(conn->encryption, scan_results) < 0) {
		return -1;
	}

	if (netlink_send_linkmode(WLANCOND_IF_DORMANT) != 0) {
		DLOG_ERR("Setting oper state to dormant failed");
		// Not fatal we can continue association
	}

	memcpy(conn->bssid, scan_results->bssid, ETH_ALEN);
	conn->channel = scan_results->channel;

	if (get_wpa_mode() == TRUE) {
		if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_PSK ||
		   (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_CERT) {
			if ((ret = generate_wapi_ie(conn->encryption, &wlan_status)) < 0) {
				DLOG_ERR("nl80211 - nl80211_encryption_method failed");
				return ret;
			}
		}
		else if ((ret = nl80211_encryption_method(conn->encryption, 
						     &wlan_status)) < 0) {
			DLOG_ERR("nl80211 - nl80211_encryption_method failed");
			return ret;
		}	
	}

	if (conn->mode & WLANCOND_ADHOC) {
		if (conn->adhoc_channel != 0 && (conn->mode & WLANCOND_ADHOC)) {
			if (conn->adhoc_channel < WLANCOND_MIN_WLAN_CHANNEL ||
			    conn->adhoc_channel > WLANCOND_MAX_WLAN_CHANNEL) {
				DLOG_ERR("Invalid ad-hoc channel: %d", 
					 conn->adhoc_channel);
				return -1;
			}
			scan_results->channel = conn->adhoc_channel;
		}
	} else {
		if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == 
		    WLANCOND_WEP)
			connection_timeout = WLANCOND_CONNECT_TIMEOUT/2;
		else
			connection_timeout = WLANCOND_CONNECT_TIMEOUT;
	}

	ret = nl80211_connect(scan_results, &wlan_status);	
	if (ret) {
		DLOG_ERR("nl80211 - connect returned error");
		return -1;
	}

	/* Set association timeout timer */
	wlan_connect_timer_id = g_timeout_add_seconds(
		connection_timeout,
		wlan_connect_timer_cb, NULL);

	return 0;
}

/**
     Scan D-BUS request.
     @param message DBUS message.
     @param connection DBUS connection.
     @return status.
*/
static DBusHandlerResult scan_request(DBusMessage    *message,
				      DBusConnection *connection) {

	DBusMessage *reply = NULL;
	DBusMessageIter iter, array_iter;
	char *ssid;
	gint ssid_len;
	const char* sender;
	dbus_int32_t power_level;
	dbus_int32_t flags;
	gint previous_state = 0;

	if (in_flight_mode()) {
		reply = new_dbus_error(message, WLANCOND_ERROR_WLAN_DISABLED);
		send_and_unref(connection, reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	sender = dbus_message_get_sender(message);
	if (sender == NULL) {
		goto param_err;
	}

	DLOG_DEBUG("Got scan request from %s", sender);

	/* Do not scan if we are scanning already or if we are associating */
	if (get_scan_state() == SCAN_ACTIVE || wlan_connect_timer_id != 0) {
		reply = new_dbus_error(message, WLANCOND_ERROR_ALREADY_ACTIVE);
		send_and_unref(connection, reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	if ((previous_state = init_if(socket_open())) < 0) {
		reply = new_dbus_error(message, WLANCOND_ERROR_INIT_FAILED);
		goto param_err;
	}

	if (previous_state == WLAN_NOT_INITIALIZED) {
		set_wlan_state(WLAN_INITIALIZED_FOR_SCAN, NO_SIGNAL, TRUE);
	}

	dbus_message_iter_init(message, &iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_INT32)
		goto param_err;
	dbus_message_iter_get_basic(&iter, &power_level);

	dbus_message_iter_next(&iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY ||
		dbus_message_iter_get_element_type(&iter) != DBUS_TYPE_BYTE)
		goto param_err;
	dbus_message_iter_recurse(&iter, &array_iter);
	dbus_message_iter_get_fixed_array(&array_iter, &ssid, &ssid_len);

	if (ssid_len > WLANCOND_MAX_SSID_SIZE+1)
		goto param_err;

	dbus_message_iter_next(&iter);

	power_down_after_scan = FALSE;

	if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_UINT32) {
		dbus_message_iter_get_basic(&iter, &flags);
		DLOG_DEBUG("Found flags: %08x", flags);

		if (flags & WLANCOND_NO_DELAYED_SHUTDOWN)
			power_down_after_scan = TRUE;
	}

	if (power_level != WLANCOND_TX_POWER10 &&
			power_level != WLANCOND_TX_POWER100) {
		DLOG_ERR("Invalid power level");
		goto param_err;
	}

	if (set_tx_power(power_level) != TRUE) {
		reply = new_dbus_error(message, WLANCOND_ERROR_IOCTL_FAILED);
		goto param_err;
	}

	if (scan(ssid, ssid_len, TRUE) < 0) {
		reply = new_dbus_error(message, WLANCOND_ERROR_IOCTL_FAILED);
		goto param_err;
	}

	g_free(scan_name_cache);
	scan_name_cache = g_strdup(sender);

	reply = new_dbus_method_return(message);
	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;

param_err:
	if (get_wlan_state() == WLAN_INITIALIZED_FOR_SCAN) {
		set_wlan_state(WLAN_NOT_INITIALIZED, NO_SIGNAL, FALSE);
	}
	if (reply == NULL) {
		DLOG_DEBUG("Parameter error in scan request");
		reply = new_dbus_error(message, DBUS_ERROR_INVALID_ARGS);
	}
	send_and_unref(connection, reply);
	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
     Network compare.
     @param a Pointer to scan_result 1.
     @param b Pointer to scan result 2.
     @return 0 if a=b, -1 if a is better than b and 1 if b is better than a.
*/
static int network_compare(gconstpointer a, gconstpointer b)
{
	const struct scan_results_t *results_a = a;
	const struct scan_results_t *results_b = b;

	if (wlan_status.scan_ssid_len > 1) {

		//DLOG_DEBUG("Scan ssid = %s", scan_ssid);

		gint a_eq = strncmp(wlan_status.scan_ssid,
				    results_a->ssid, WLANCOND_MAX_SSID_SIZE);
                gint b_eq = strncmp(wlan_status.scan_ssid,
				    results_b->ssid, WLANCOND_MAX_SSID_SIZE);
		// Check if either network match the scan SSID
		if (!a_eq && !b_eq) {
			//DLOG_DEBUG("Both (%s, %s) match scan SSID",
			//results_a->ssid, results_b->ssid);
			return 0;
		}

		if (!a_eq && b_eq) {
			//DLOG_DEBUG("%s is better than %s",
			//results_a->ssid, results_b->ssid);
			return -1;
		}

		if (a_eq && !b_eq) {

			//DLOG_DEBUG("%s is better than %s",
			//results_b->ssid, results_a->ssid);
			return 1;
		}

	}
	//DLOG_DEBUG("No scan ssid, returning just RSSI values");

	return (results_a->rssi > results_b->rssi) ?
		-1 : (results_a->rssi < results_b->rssi) ? 1 : 0;
}

/**
   Add scan results to roam cache.
   @param scan_results Scan results to add.
*/
static void add_to_roam_cache(struct scan_results_t *scan_results)
{
	GSList *list;
	roam_cache_key_t key = { scan_results->bssid, scan_results->channel };
	list = g_slist_find_custom(wlan_status.roam_cache, &key,
				   &compare_scan_entry);

	if (list != NULL) {
		struct scan_results_t *roam_cache_entry = list->data;
		print_mac(WLANCOND_PRIO_LOW, "Found old entry for:",
			  scan_results->bssid);
		
		/* Remove the old entry */
		wlan_status.roam_cache = g_slist_remove(
				wlan_status.roam_cache, roam_cache_entry);

		clean_scan_results_item(roam_cache_entry, NULL);
	}

	print_mac(WLANCOND_PRIO_LOW, "New AP to roam cache:",
		  scan_results->bssid);

	struct scan_results_t *results_to_list = copy_scan_results(scan_results);
	wlan_status.roam_cache = g_slist_prepend(wlan_status.roam_cache,
						 results_to_list);

	return;
}

/**
   Make a copy of scan results.
   @param results Scan results to copy.
   @return Copied results.
*/
struct scan_results_t* copy_scan_results(struct scan_results_t *results) {
	
	struct scan_results_t *copy = g_slice_dup(struct scan_results_t, 
						  results);
	copy->wpa_ie = g_memdup(results->wpa_ie, results->wpa_ie_len);
	copy->uuid_e = g_memdup(results->uuid_e, MAX_UUID_E_LEN);
	
	return copy;
}

/**
   Remove from roam cache.
   @param bssid BSSID to remove.
   @return status.
*/
gboolean remove_from_roam_cache(roam_cache_key_t* const key)
{
	GSList *list;

	list = g_slist_find_custom(wlan_status.roam_cache, key,
				   &compare_scan_entry);

	if (list != NULL) {
		struct scan_results_t *roam_cache_entry = list->data;
		print_mac(WLANCOND_PRIO_LOW, "Found entry to be removed:",
			  key->bssid);

                wlan_status.roam_cache = g_slist_remove(
			wlan_status.roam_cache, roam_cache_entry);

		clean_scan_results_item(roam_cache_entry, NULL);

		return TRUE;
	}

	return FALSE;
}

/**
   Give penalty to failed AP.
   @param bssid BSSID.
   @return status.
*/
gboolean decrease_signal_in_roam_cache(roam_cache_key_t* const key)
{
	GSList *list;

	list = g_slist_find_custom(wlan_status.roam_cache, key,
				   &compare_scan_entry);

	if (list != NULL) {
		struct scan_results_t *roam_cache_entry = list->data;
		print_mac(WLANCOND_PRIO_LOW, "Found entry to be decreased:",
			  key->bssid);

		roam_cache_entry->rssi -= WLANCOND_RSSI_PENALTY;

		return TRUE;
	}

	return FALSE;
}
/**
   Compare group ciphers.
   @param c1 Cipher 1.
   @param c2 Cipher 2.
   @return 1 if matches.
*/
static int check_group_cipher(guint32 c1, guint32 c2)
{
	guint32 m1 = (c1 & WLANCOND_ENCRYPT_GROUP_ALG_MASK);
	guint32 m2 = (c2 & WLANCOND_ENCRYPT_GROUP_ALG_MASK);

	if (m1 == m2)
                return 1;
	if (m2 == WLANCOND_WPA_TKIP_GROUP && (m1 & WLANCOND_WPA_TKIP_GROUP))
		return 1;
	if (m2 == (unsigned int)WLANCOND_WPA_AES_GROUP &&
	    (m1 & WLANCOND_WPA_AES_GROUP))
		return 1;
	if (m2 == WLANCOND_WAPI_SMS4_GROUP && (m1 & WLANCOND_WAPI_SMS4_GROUP))
		return 1;

	DLOG_DEBUG("Group ciphers don't match");

	return -1;
}
/**
   Compare ciphers.
   @param c1 Cipher 1.
   @param c2 Cipher 2.
   @return 1 if matches.
*/
static int check_ciphers(guint32 c1, guint32 c2)
{
	guint32 u1 = (c1 & WLANCOND_ENCRYPT_ALG_MASK);
	guint32 u2 = (c2 & WLANCOND_ENCRYPT_ALG_MASK);

	if (check_group_cipher(c1, c2) < 0) {
		DLOG_DEBUG("Group cipher's don't match");
                return -1;
	}

	if (u1 == u2)
                return 1;
	if (u2 == WLANCOND_WPA_TKIP && (u1 & WLANCOND_WPA_TKIP))
		return 1;
	if (u2 == WLANCOND_WPA_AES && (u1 & WLANCOND_WPA_AES))
		return 1;
	if (u2 == WLANCOND_WAPI_SMS4 && (u1 & WLANCOND_WAPI_SMS4))
		return 1;

	DLOG_DEBUG("Unicast ciphers don't match");

	return -1;
}
static gboolean wlan_roam_scan_cb(void* data)
{
	wlan_status.roam_scan_id = 0;
	struct timeval tv;

	if (wlan_status.signal == WLANCOND_LOW &&
	    get_wlan_state() == WLAN_CONNECTED) {

		DLOG_DEBUG("Roam scan timeout, initiating new scan");

		if (scan(wlan_status.conn.ssid, wlan_status.conn.ssid_len, 
			 TRUE) < 0) {
			return FALSE;
		}
		if (gettimeofday(&tv, NULL) >= 0)
			wlan_status.last_scan = tv.tv_sec;
	}

        return FALSE;
}

/**
   Schedule scan.
   @param seconds Delay scan for this many seconds.
*/
void schedule_scan(guint seconds) {

	/* Remove old timer */
	remove_roam_scan_timer();
	wlan_status.roam_scan_id = g_timeout_add_seconds(
		seconds,
		wlan_roam_scan_cb,
		NULL);
}

/**
   Reschedule scan.
*/
static void reschedule_scan(void)
{
        /*
           If we are active use shortest scan interval, otherwise
           use exponential backoff
        */
	if (get_inactivity_status() == FALSE) {
		wlan_status.roam_scan = WLANCOND_MIN_ROAM_SCAN_INTERVAL;
	} else {
                if (wlan_status.roam_scan <= WLANCOND_MIN_ROAM_SCAN_INTERVAL) {
			wlan_status.roam_scan = WLANCOND_MIN_ROAM_SCAN_INTERVAL;
		} else if (wlan_status.roam_scan >=
				WLANCOND_MAX_ROAM_SCAN_INTERVAL) {
			wlan_status.roam_scan = WLANCOND_MAX_ROAM_SCAN_INTERVAL;
		} else {
			wlan_status.roam_scan = wlan_status.roam_scan * 2;
		}
	}

	schedule_scan(wlan_status.roam_scan);
}

/**
   Check capabilities from scan results.
   @param scan_results Scan results.
   @param conn Connection paramters.
   @return status.
*/
static gboolean check_capabilities(struct scan_results_t *scan_results,
                                   struct connect_params_t *conn)
{
	// Check mode
	if ((scan_results->cap_bits & WLANCOND_MODE_MASK) != (guint32)conn->mode)
		return FALSE;
	if ((scan_results->cap_bits & WLANCOND_ENCRYPT_METHOD_MASK) !=
	    (guint32)(conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK))
		return FALSE;
	
	if ((scan_results->cap_bits & WLANCOND_ENCRYPT_WPA2_MASK) !=
	    (guint32)(conn->encryption & WLANCOND_ENCRYPT_WPA2_MASK))
		return FALSE;

	if (check_ciphers(scan_results->cap_bits, conn->encryption) < 0)
		return FALSE;
	
	return TRUE;
}
/**
     Compare current best AP to the candidate AP.
     @param current_best Scan results for the current best AP.
     @param candidate Scan results for the candidate AP.
     @return TRUE if better than the current best AP.
*/
static gboolean better_than_current_best(struct scan_results_t *current_best,
					 struct scan_results_t *candidate) {
	gint current_best_rssi;
	gint candidate_rssi;

	/* 5 GHz APs get WLANCOND_5GHZ_BENEFIT */
	if (current_best->channel > 14)
		current_best_rssi = current_best->rssi + WLANCOND_5GHZ_BENEFIT;
	else
		current_best_rssi = current_best->rssi;
	
	if (candidate->channel > 14)
		candidate_rssi = candidate->rssi + WLANCOND_5GHZ_BENEFIT;
	else
		candidate_rssi = candidate->rssi;
	
	if (current_best_rssi > candidate_rssi)
		return FALSE;
	
	return TRUE;
}

/**
     Find connection.
     @param ap_list List of access points.
     @param conn Connection parameters.
     @param update_roam_cache Update roaming cache or not.
     @return scan_results_t Returns best connection if found.
*/
struct scan_results_t* find_connection(
	GSList* ap_list, struct connect_params_t *conn,
	gboolean update_roam_cache)
{
	GSList *list;
	struct scan_results_t *best_connection = NULL;
	gint current_rssi = 0;

	/* If update roam cache, clean it first */
	if (update_roam_cache == TRUE)
		clean_roam_cache();

	for (list = ap_list; list != NULL; list = list->next) {
		struct scan_results_t *scan_results = list->data;
		if (scan_results->ssid_len == conn->ssid_len &&
		    memcmp(scan_results->ssid, conn->ssid,
			   scan_results->ssid_len) == 0) {
			print_mac(WLANCOND_PRIO_LOW, "Found AP:",
				  scan_results->bssid);
			
			/* Find the current AP so that we know it's RSSI */
			if (memcmp(scan_results->bssid, wlan_status.conn.bssid,
			    ETH_ALEN) == 0 &&
			    scan_results->channel == wlan_status.conn.channel) {
				current_rssi = scan_results->rssi;
				DLOG_DEBUG("Current AP: %d", current_rssi);
			}

			if (check_capabilities(scan_results, conn) ==
			    FALSE)
				continue;
			
			if (is_ap_in_black_list(scan_results->bssid) == TRUE) {
				DLOG_INFO("AP is in black list, discarded");
				continue;
			}

			/* At this point we know the connection is good,
			   add to the roam cache
			 */
			if (update_roam_cache == TRUE) {
				add_to_roam_cache(scan_results);
			}

			if (best_connection == NULL ||
			    better_than_current_best(best_connection,
						     scan_results)) {
				DLOG_DEBUG("Best connection: %d (old %d)",
					   scan_results->rssi,
					   best_connection == NULL ?
					   WLANCOND_MINIMUM_SIGNAL:
					   best_connection->rssi);
				    best_connection = scan_results;
			}
		}
	}

	/* Did not find any connection */
	if (best_connection == NULL)
		return NULL;

	/* Check if we are already connected but the best connection is not
	   worth changing
	 */
	if (current_rssi != 0) {
		if (best_connection->rssi < current_rssi +
		    WLANCOND_ROAM_THRESHOLD) {
			DLOG_DEBUG("Best connection not good enough");
			return NULL;
		}
	}

	/* Check if the signal level is good enough for connection */
	if (best_connection->rssi < WLANCOND_MINIMUM_SIGNAL)
		return NULL;

	return best_connection;
}
/**
   Select Adhoc channel.
   @param ap_list List of scanned APs.
   @return selected channel.
*/
static dbus_uint32_t find_adhoc_channel(GSList *ap_list) {

	dbus_uint32_t used_channel_list = 0;
	dbus_uint32_t selected_channel = 0;
	GSList* list;
	guint32 i;

	/* Create a map of free channels */
	for (list = ap_list; list != NULL; list = list->next) {
		struct scan_results_t *scan_results = list->data;
		used_channel_list |= 1 << scan_results->channel;
	}

	for (i = 1; i <= 11; i++) {
		if (!(used_channel_list & (1 << i))) {
			selected_channel = i;
			break;
		}
	}

	if (selected_channel == 0) {
		/* No free channel found, choose one randomly from
		 * channels 1 - 11. */
		selected_channel = g_random_int_range(1, 12);
	}

	//DLOG_DEBUG("Selected adhoc channel: %d", selected_channel);

	return selected_channel;
}
/**
   Check if the selected scan result is from the same AP as one of the previous
   results.
   @param scan_results List of scanned APs.
   @param entry AP scan result list entry.
   @return TRUE if duplicate.
*/
static gboolean check_if_duplicate_entry(GSList *scan_results,
					 GSList *entry) {
	
	GSList *list;
	struct scan_results_t *entry_result = entry->data;
	
	if (entry_result->uuid_e == NULL)
		return FALSE;

	for (list = scan_results; list && list != entry; list = list->next) {
		struct scan_results_t *result = list->data;	
		if (result->cap_bits & WLANCOND_WPS_PUSH_BUTTON &&
		    result->cap_bits & WLANCOND_WPS_CONFIGURED && 
		    result->uuid_e && memcmp(result->uuid_e, 
					     entry_result->uuid_e, 
					     MAX_UUID_E_LEN) == 0) {
			print_mac(WLANCOND_PRIO_MEDIUM, 
				  "Found matching previous entry:", 
				  result->bssid);
			return TRUE;
		}	
	}
	return FALSE;
}	
/**
   Find connection and associate.
   @param scan_results Scan results.
   @return status.
*/
int find_connection_and_associate(GSList *scan_results,
				  gboolean update_roam_cache,
				  gboolean create_new_adhoc,
				  gboolean autoconnect)
{
	struct scan_results_t adhoc;
	struct connect_params_t *conn = &wlan_status.conn;
	guint wps_pbc_registrars = 0;
	GSList* list;

	/* Check for too many registrars in the WPS PBC
	   method */
	if (conn->encryption & WLANCOND_WPS_PUSH_BUTTON) {
		for (list = scan_results; list != NULL; list = list->next) {
			struct scan_results_t *scan_result = list->data;
			
			if (scan_result->cap_bits & WLANCOND_WPS_PUSH_BUTTON
			    &&
			    scan_result->cap_bits & WLANCOND_WPS_CONFIGURED) 
			{
				print_mac(WLANCOND_PRIO_MEDIUM, 
					  "Found active PBC session from:", 
					  scan_result->bssid);
				/* 
				   Check if this is a duplicate of an earlier
				   list entry.
				*/
				if (check_if_duplicate_entry(scan_results, 
							     list) == FALSE) {
					if (++wps_pbc_registrars > 1) {
						DLOG_ERR("Too many WPS PBC registrars");
						return ETOOMANYREGISTRARS;
					}
				}
			}
		}
		/* In case of WPS push button we double check for 
		   extra registrars */
		if (wlan_status.retry_count < 1)
			return ENEEDEXTRASCAN;
	}

	struct scan_results_t *connection = find_connection(
		scan_results, &wlan_status.conn, update_roam_cache);

        if (connection == NULL && conn->mode == WLANCOND_ADHOC &&
	    create_new_adhoc == TRUE) {
		DLOG_DEBUG("No existing adhoc connection");
		memset(&adhoc, 0, sizeof(adhoc));
		connection = &adhoc;
		memcpy(connection->ssid, conn->ssid, conn->ssid_len);
		connection->channel = find_adhoc_channel(scan_results);
	}

	if (connection) {
		if (autoconnect == TRUE && 
		    connection->rssi < WLANCOND_MINIMUM_AUTOCONNECT_RSSI) {
			DLOG_WARN("RSSI too low for autoconnect");
			return ETOOWEAKAP;
		}
		int ret = associate(connection);
		if (ret < 0) {
			memset(wlan_status.conn.bssid, 0, ETH_ALEN);
			wlan_status.conn.channel = 0;
		}
		if (ret == ESUPPLICANT)
			kill_supplicant();
		return ret;
	}
	return -1;
}
/**
   Send WPS too many registrars signal.
*/
static void registrar_error_signal(void)
{
	DBusMessage *registrar_error;

	registrar_error = new_dbus_signal(
		WLANCOND_SIG_PATH,
		WLANCOND_SIG_INTERFACE,
		WLANCOND_REGISTRAR_ERROR_SIG,
		NULL);
	
	send_and_unref(get_dbus_connection(), registrar_error);
}

/**
    Connects based on scan results.
    @param scan_results List of scan results.
*/
static void connect_from_scan_results(GSList *scan_results)
{
	gboolean autoconnect = !!(wlan_status.conn.flags & WLANCOND_AUTOCONNECT);

	int status = find_connection_and_associate(
		scan_results, TRUE, TRUE, autoconnect);

	clean_scan_results(&scan_results);

	if (status == 0)
		return;

	DLOG_DEBUG("Could not find suitable network");

	if (get_wlan_state() == WLAN_INITIALIZED_FOR_CONNECTION) {	
		/* In WPS too many registrars case we send
		   different signal
		*/
		if (status == ETOOMANYREGISTRARS) {
			set_wlan_state(WLAN_NOT_INITIALIZED,
				       NO_SIGNAL, FALSE);
			registrar_error_signal();
		} else if (status != ETOOWEAKAP && status != ESUPPLICANT &&
			   /* Try rescanning if retries left */
			   no_connection_and_scan_retries_left()) {
			DLOG_DEBUG("Rescanning");
			wlan_status.rescan_id = g_timeout_add_seconds(
				WLANCOND_RESCAN_DELAY,
				rescan,
				NULL);
		} else {
			set_wlan_state(WLAN_NOT_INITIALIZED,
				       DISCONNECTED_SIGNAL, FALSE);
		}
		return;
	}

	/* We are connected but would prefer better connection */
	reschedule_scan();
}

/**
     Scan results request.
     @param ifindex Interface index.
     @return status.
*/
gboolean ask_scan_results(void)
{
	GSList *scan_results_save = NULL;
	dbus_int32_t number_of_results;

	if (nl80211_send_get_scan_results(&wlan_status, &scan_results_save)) {
		clean_scan_results(&scan_results_save);
		return FALSE;
	}
	/* First send scan results if someone was expecting them */
	if (scan_name_cache != NULL) {	
		
		number_of_results = g_slist_length(scan_results_save);

		/* Sort the list only if the amount of networks is very high and
		   we need to restrict the results */
		if (number_of_results > WLANCOND_MAX_NETWORKS)
			scan_results_save = g_slist_sort(scan_results_save,
							 network_compare);

		send_dbus_scan_results(scan_results_save, scan_name_cache,
				       number_of_results);
		g_free(scan_name_cache);
		scan_name_cache = NULL;
	}
	
	/* Try to associate if state is initialized_for_connection or
	   signal level is low */
	if ((get_wlan_state() == WLAN_INITIALIZED_FOR_CONNECTION ||
	     wlan_status.signal == WLANCOND_LOW) &&
	    get_scan_state() == SCAN_ACTIVE) {

		DLOG_DEBUG("Connect from scan");

		set_scan_state(SCAN_NOT_ACTIVE);

		connect_from_scan_results(scan_results_save);

                return TRUE;
	}

	set_scan_state(SCAN_NOT_ACTIVE);

	if (get_wlan_state() == WLAN_INITIALIZED_FOR_SCAN &&
	    power_down_after_scan == TRUE) {
		set_interface_state(socket_open(), CLEAR, IFF_UP);
	}

	if (get_wlan_state() == WLAN_INITIALIZED_FOR_SCAN) {
		/* Save scan results temporarily */
		if (wlan_status.roam_cache) {
			clean_roam_cache();
		}
		wlan_status.roam_cache = scan_results_save;
		set_wlan_state(WLAN_NOT_INITIALIZED, NO_SIGNAL, TRUE);
	} else {
		clean_scan_results(&scan_results_save);
	}

	return TRUE;
}

/**
    Handles scan results ready netlink event
*/
void nl80211_scan_ready(void) {
	remove_scan_timer();

	if (get_scan_state() == SCAN_ACTIVE) {
		DLOG_INFO("Scan results ready -- scan active");
		if (ask_scan_results() == FALSE) {
			DLOG_ERR("Getting scan results failed");

			set_wlan_state(WLAN_NOT_INITIALIZED,
					DISCONNECTED_SIGNAL,
					FALSE);
		}
	} else {
		DLOG_DEBUG("Scan results ready -- "
			   "not requested");
	}
}

/**
    Disconnect WLAN and remove kernel module D-BUS request.
    @param message DBUS message.
    @param connection DBUS connection.
    @return status.
*/
static DBusHandlerResult disconnect_request(DBusMessage    *message,
		DBusConnection *connection) {
	DBusMessage *reply;

	/* Set_wlan_state puts IF down */
	set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL, FALSE);

	reply = new_dbus_method_return(message);
	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
    Disassociate WLAN D-BUS request.
    @param message DBUS message.
    @param connection DBUS connection.
    @return status.
*/
static DBusHandlerResult disassociate_request(DBusMessage    *message,
		DBusConnection *connection) {
	DBusMessage *reply;

	if (get_wlan_state() != WLAN_CONNECTED &&
	    get_wlan_state() != WLAN_NO_ADDRESS) {
		DLOG_DEBUG("Not in correct state for disassociation");

		reply = new_dbus_method_return(message);
		send_and_unref(connection, reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	mlme_command(wlan_status.conn.bssid, WLANCOND_NL80211_MLME_DISCONNECT,
		     WLANCOND_REASON_LEAVING);

	set_wlan_state(WLAN_INITIALIZED_FOR_CONNECTION, NO_SIGNAL, TRUE);

	DLOG_DEBUG("Disassociated, trying to find a new connection");

	scan(wlan_status.conn.ssid, wlan_status.conn.ssid_len, TRUE);

	reply = new_dbus_method_return(message);
        send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
   Status D_BUS request.
   @param message DBUS message.
   @param connection DBUS connection.
   @return status.
*/
static DBusHandlerResult status_request(DBusMessage    *message,
					DBusConnection *connection) {
	DBusMessage *reply = NULL;
	dbus_uint32_t sens = 0;
	dbus_uint32_t security = 0;
	dbus_uint32_t capability = 0;
	dbus_uint32_t channel = 0;
	GSList *list;
	struct scan_results_t *current_conn_ap = NULL;
	struct connect_params_t *conn = &wlan_status.conn;
	char *essid = conn->ssid;
	unsigned char *bssid = conn->bssid;
	guint8 power;

	if (get_wlan_state() != WLAN_CONNECTED &&
	    get_wlan_state() != WLAN_NO_ADDRESS &&
	    get_mode() != WLANCOND_ADHOC) {
		reply = new_dbus_error(message, WLANCOND_ERROR_IOCTL_FAILED);
		send_and_unref(connection, reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	roam_cache_key_t key = { conn->bssid, conn->channel };
	list = g_slist_find_custom(wlan_status.roam_cache, &key, 
				   &compare_scan_entry);
	if (list == NULL) {	
		if (get_mode() != WLANCOND_ADHOC) {
			DLOG_ERR("nl80211: MAC entry not found");
			reply = new_dbus_error(message, WLANCOND_ERROR_FATAL);
			goto param_err;
		}
	} else {
		current_conn_ap = list->data;
	}

	/* Link quality i.e. stats */
	if (nl80211_get_signal_info(&power, &wlan_status) !=0) {	
		if (get_mode() != WLANCOND_ADHOC) {
			DLOG_ERR("Could not get statistics");
			reply = new_dbus_error(message, 
					       WLANCOND_ERROR_IOCTL_FAILED);
			goto param_err;
		}
	} else {
		sens = power - 0x100;
	}
	
	/* Channel */
	channel = conn->channel;

	if (channel < WLANCOND_MIN_WLAN_CHANNEL ||
	    channel > WLANCOND_MAX_WLAN_CHANNEL) {
		channel = 0;
		DLOG_DEBUG("Got invalid channel");
	}

	/* Mode (Adhoc/Infra) */
	if (conn->mode == WLANCOND_ADHOC) {
		capability |= WLANCOND_ADHOC;
	} else if (conn->mode == WLANCOND_INFRA) {
		capability |= WLANCOND_INFRA;
	}

	/* encryption status */
	security = wlan_status.conn.encryption;

	/* Speed / Rate */
	if (current_conn_ap != NULL)
		capability |= current_conn_ap->cap_bits & WLANCOND_RATE_MASK;

	reply = new_dbus_method_return(message);

	gchar* ifname = wlan_status.ifname;

	append_dbus_args(reply,
			 DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
			 &essid, conn->ssid_len-1,
			 DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &bssid, ETH_ALEN,
			 DBUS_TYPE_UINT32, &sens,
			 DBUS_TYPE_UINT32, &channel,
			 DBUS_TYPE_UINT32, &capability,
			 DBUS_TYPE_UINT32, &security,
			 DBUS_TYPE_STRING, &ifname,
			 DBUS_TYPE_INVALID);

	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;

param_err:	
	set_wlan_state(WLAN_NOT_INITIALIZED, DISCONNECTED_SIGNAL, FALSE);

	if (reply == NULL) {
		DLOG_DEBUG("Parameter error in status request");
		reply = new_dbus_error(message, DBUS_ERROR_INVALID_ARGS);
	}

	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   Interface D_BUS request.
   @param message DBUS message.
   @param connection DBUS connection.
   @return status.
*/
static DBusHandlerResult interface_request(DBusMessage    *message,
		DBusConnection *connection) {
	DBusMessage *reply;
	gchar* ifname = wlan_status.ifname;

	reply = new_dbus_method_return(message);

	append_dbus_args(reply,
			 DBUS_TYPE_STRING, &ifname,
			 DBUS_TYPE_INVALID);

	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   Connection status D_BUS request.
   @param message DBUS message.
   @param connection DBUS connection.
   @return status.
*/
static DBusHandlerResult connection_status_request(
		DBusMessage    *message,
		DBusConnection *connection) {

	DBusMessage *reply;
	dbus_bool_t state = FALSE;

	guint state_v = get_wlan_state();

	if (state_v == WLAN_INITIALIZED ||
	    state_v == WLAN_INITIALIZED_FOR_CONNECTION ||
	    state_v == WLAN_NO_ADDRESS ||
	    state_v == WLAN_CONNECTED)
		state = TRUE;

	reply = new_dbus_method_return(message);

	append_dbus_args(reply,
			DBUS_TYPE_BOOLEAN, &state,
			DBUS_TYPE_INVALID);

	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}
/**
   Set pmksa D-BUS request.
   @param message DBUS message.
   @param connection DBUS connection.
   @return status.
*/
static DBusHandlerResult set_pmksa_request(DBusMessage    *message,
		DBusConnection *connection) {

	DBusMessage *reply = NULL;
	unsigned int pmkid_len, mac_len;
	unsigned char *pmkid;
	unsigned char *mac;
	dbus_uint32_t action;
	DBusError derror;

	dbus_error_init(&derror);

	if (dbus_message_get_args(
		    message, &derror,
		    DBUS_TYPE_UINT32, &action,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &pmkid, &pmkid_len,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &mac, &mac_len,
		    DBUS_TYPE_INVALID) == FALSE)
	{
		DLOG_ERR("Failed to parse set_pmksa request: %s",
			 derror.message);
		dbus_error_free(&derror);
		goto param_err;
	}

	if (action != WLANCOND_PMKSA_ADD) {
		DLOG_ERR("Invalid action");
		goto param_err;
	}

	if (pmkid == NULL || pmkid_len != WLANCOND_PMKID_LEN || mac == NULL
	    || mac_len != ETH_ALEN) {
		DLOG_ERR("Invalid arguments");
		goto param_err;
	}

	add_to_pmksa_cache(pmkid, mac);

	print_mac(WLANCOND_PRIO_LOW, "PMKSA added successfully for address:",
		  mac);

	reply = new_dbus_method_return(message);
	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;

param_err:
	if (reply == NULL) {
		DLOG_DEBUG("Parameter error in set_pmksa");
		reply = new_dbus_error(message, DBUS_ERROR_INVALID_ARGS);
	}
	send_and_unref(connection, reply);
	return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   Set powersave D-BUS request.
   @param message DBUS message.
   @param connection DBUS connection.
   @return status.
*/
static DBusHandlerResult set_powersave_request(DBusMessage    *message,
		DBusConnection *connection) {

	DBusMessage *reply = NULL;
	DBusError error;
	dbus_bool_t onoff;

	dbus_error_init(&error);

	if (dbus_message_get_args(message, &error,
				  DBUS_TYPE_BOOLEAN, &onoff,
				  DBUS_TYPE_INVALID) == FALSE) {
		DLOG_ERR("Failed to parse message: %s",
			 error.message);
		dbus_error_free(&error);
		
		send_invalid_args(connection, message);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	/* Powersave can be allowed only when we are properly connected
	   or when the entity asking for connection wants powersave despite of
	   the state */
	if (onoff == TRUE) {
		if (get_wlan_state() == WLAN_NOT_INITIALIZED) {
			set_wlan_state(WLAN_NOT_INITIALIZED,
				       NO_SIGNAL, TRUE);
		} else if (get_wlan_state() != WLAN_NO_ADDRESS ||
			   (connect_name_cache != NULL &&
			    strcmp(dbus_message_get_sender(message),
				   connect_name_cache) == 0)) {
			if (set_power_state(wlan_status.powersave) == FALSE) {
				DLOG_ERR("Setting powersave failed");
				// Not fatal
			}
		}
	} else {
		// Go to full power
		if (set_power_state(WLANCOND_POWER_ON) == FALSE) {
			DLOG_ERR("Setting powersave failed");
			// Not fatal
		}
	}

	DLOG_DEBUG("WLAN powersave %s", onoff==TRUE?"on":"off");

	reply = new_dbus_method_return(message);
	send_and_unref(connection, reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

/**
   WPA IE D-BUS push.
   @param ap_mac_addr Access point MAC address.
   @return status.
*/
int wpa_ie_push(char* ap_mac_addr) {

	DBusMessage *msg;
	DBusMessage *reply;
	DBusError derr;
	struct connect_params_t *conn = &wlan_status.conn;
	dbus_bool_t autoconnect = !!(conn->flags & WLANCOND_AUTOCONNECT);
	char *ssid = conn->ssid;
	unsigned char *ap_wpa_ie;
	guint32 ap_wpa_ie_len;
	GSList *list;
	
	if (conn->authentication_type != EAP_AUTH_TYPE_WFA_SC) {
		if (wlan_status.wpa_ie.ie_len == 0 ||
		    conn->ssid_len == 0) {
			DLOG_ERR("WPA IE / SSID not valid");
			return -1;
		}
	}	

	roam_cache_key_t key = { (unsigned char*) ap_mac_addr, conn->channel };
	list = g_slist_find_custom(wlan_status.roam_cache, &key,
				   &compare_scan_entry);

	if (list == NULL)
		return -1;

	struct scan_results_t *current_ap_entry = list->data;
	ap_wpa_ie = current_ap_entry->wpa_ie;
	ap_wpa_ie_len = current_ap_entry->wpa_ie_len;

	dbus_error_init(&derr);

	msg = dbus_message_new_method_call(
		EAP_SERVICE,
		EAP_REQ_PATH,
		EAP_REQ_INTERFACE,
		EAP_WPA_IE_PUSH_REQ);
	
	if (msg == NULL) {
		return -1;
	}

	dbus_message_set_auto_start(msg, FALSE);

	append_dbus_args(
		msg,
		DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
		&wlan_status.wpa_ie.ie, wlan_status.wpa_ie.ie_len,
		DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &ap_wpa_ie,
		ap_wpa_ie_len,
		DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
		&ssid, conn->ssid_len -1,
		DBUS_TYPE_UINT32, &wlan_status.pairwise_cipher,
		DBUS_TYPE_UINT32, &wlan_status.group_cipher,
		DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
		&ap_mac_addr, ETH_ALEN,
		DBUS_TYPE_UINT32, &conn->authentication_type,
		DBUS_TYPE_BOOLEAN, &autoconnect,
		DBUS_TYPE_INVALID);
	
	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, -1, &derr);

	dbus_message_unref(msg);
	
	if (dbus_error_is_set(&derr)) {
		DLOG_ERR("EAP returned error: %s", derr.name);

		dbus_error_free(&derr);
		if (reply)
			dbus_message_unref(reply);
		return -1;
	}

	dbus_message_unref(reply);

	return 0;
}
/**
   WPA MIC failure event D-BUS request.
   @param key_type Key type, Unicast/Broadcast key.
   @param is_fatal Error is fatal if true.
   @return status.
*/
int wpa_mic_failure_event(dbus_bool_t key_type, dbus_bool_t is_fatal) {
	DBusMessage *msg;
	DBusMessage *reply;
	DBusError derr;

	msg = dbus_message_new_method_call(
		EAP_SERVICE,
		EAP_REQ_PATH,
		EAP_REQ_INTERFACE,
		EAP_WPA_MIC_FAILURE_REQ);
	
	if (msg == NULL) {
		return -1;
	}
	
	append_dbus_args(msg,
			 DBUS_TYPE_BOOLEAN, &key_type,
			 DBUS_TYPE_BOOLEAN, &is_fatal,
			 DBUS_TYPE_INVALID);
	
	dbus_error_init(&derr);

	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, -1, &derr);
	
	dbus_message_unref(msg);

	if (dbus_error_is_set(&derr)) {
		DLOG_ERR("EAP returned error: %s", derr.name);

		dbus_error_free(&derr);
		if (reply)
			dbus_message_unref(reply);
		return -1;
	}

	dbus_message_unref(reply);

	return 0;
}
/**
    Disassociate callback.
    @param pending Pending DBUS message.
    @param user_data Callback data.
    @return status
*/
static void disassociate_cb(DBusPendingCall *pending,
			    void *user_data)
{
	DBusMessage *reply;
	DBusError error;

	dbus_error_init (&error);

	reply = dbus_pending_call_steal_reply(pending);

	if (dbus_set_error_from_message(&error, reply)) {

		DLOG_DEBUG("EAP disassociate call result:%s", error.name);

		dbus_error_free(&error);
	}

	if (reply)
		dbus_message_unref(reply);
	dbus_pending_call_unref(pending);
}

/**
   Disassociate EAP D-BUS request.
   @return status.
*/
int disassociate_eap(void) {
	DBusMessage *msg;
	DBusPendingCall *pending;

	msg = dbus_message_new_method_call(
		EAP_SERVICE,
		EAP_REQ_PATH,
		EAP_REQ_INTERFACE,
		EAP_DISASSOCIATE_REQ);
	
	if (msg == NULL) {
		return -1;
	}

	dbus_message_set_auto_start(msg, FALSE);

	if (!dbus_connection_send_with_reply(get_dbus_connection(),
					     msg, &pending, -1))
		die("Out of memory");

	if (!dbus_pending_call_set_notify (pending, disassociate_cb, NULL,
					   NULL))
		die("Out of memory");

	dbus_message_unref(msg);

	return 0;
}

/**
 * Checks the PMKSA cache status from the EAP daemon.
 *
 * @param own_mac                   Our own mac address.
 * @param own_mac_len               The length of the address(usually ETH_ALEN)
 * @param bssid                     The mac address of the access point.
 * @param bssid_len                 The length of the address(usually ETH_ALEN)
 * @param authentication_type       The authentication type used.
 * @param pairwise_key_cipher_suite The cipher suite to be used
 * @param group_key_cipher_suite    The cipher suite to be used
 * @param status                    TRUE if the key exists in the cache. On
 *                                  errors this value won't be set.
 * @return                          0 on success, non-zero on errors.
 */
int check_pmksa_cache(unsigned char* own_mac, int own_mac_len,
		      unsigned char* bssid, int bssid_len,
		      uint32_t authentication_type,
		      uint32_t pairwise_key_cipher_suite,
		      uint32_t group_key_cipher_suite,
		      int *status)
{
	DBusMessage *msg = NULL;
	DBusMessage *reply = NULL;
	DBusError error;
	dbus_bool_t found;

	dbus_error_init (&error);

	msg = dbus_message_new_method_call(
		EAP_SERVICE,
		EAP_REQ_PATH,
		EAP_REQ_INTERFACE,
		EAP_CHECK_PMKSA_CACHE_REQ);
	
	if (msg == NULL)
		return -1;

	dbus_message_set_auto_start(msg, FALSE);
	
	if (dbus_message_append_args(
		    msg,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &own_mac, own_mac_len,
		    DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &bssid, bssid_len,
		    DBUS_TYPE_UINT32, &authentication_type,
		    DBUS_TYPE_UINT32, &pairwise_key_cipher_suite,
		    DBUS_TYPE_UINT32, &group_key_cipher_suite,
		    DBUS_TYPE_INVALID) == FALSE)
	{
		DLOG_ERR("Unable to add args to dbus method call.");
		dbus_message_unref(msg);
		return -1;
	}
	
	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, -1, &error);

	dbus_message_unref(msg);

	if (dbus_error_is_set(&error)) {
		DLOG_ERR("EAP returned error: %s", error.message);
		dbus_error_free(&error);

		goto error;
	}

	if (!dbus_message_get_args(reply, &error,
				DBUS_TYPE_BOOLEAN, &found,
				DBUS_TYPE_INVALID))
	{
		DLOG_ERR("Error parsing the return value: %s", error.message);
		if (dbus_error_is_set(&error))
			dbus_error_free(&error);

		goto error;
	}

	dbus_message_unref(reply);

	*status = found;

	return 0;

error:
	if (reply)
		dbus_message_unref(reply);

	return -1;
}
/**
 * Write scan sysfs file.
 * @param file File to be written.
 * @param value Value to be written.
 */
static void write_file(const char* file, guint value) {
	char buf[4];
	FILE *sysfsfile;

	sysfsfile = fopen(file, "w");
	if (sysfsfile == NULL)
		return;
	snprintf(buf, 3, "%d", value);
	if (fwrite(buf, 3, 1, sysfsfile) != 1) {
		DLOG_ERR("Could not write to: %s", file);
	}
	fclose(sysfsfile);
	return;
}
/**
 * Sets the scan mode.
 * @param scan_mode The scan mode.
 */
static void set_scan_params(guint scan_mode) {
	static guint current_scan_mode = 0;

	if (current_scan_mode == scan_mode)
		return;
	
	current_scan_mode = scan_mode;
	
	if (scan_mode == SCAN_SPLIT) {
		/* Scan size */
		write_file(SCAN_SIZE_FILE, SCAN_SIZE_SPLIT);
		/* Scan delay */
		write_file(SCAN_DELAY_FILE, SCAN_DELAY_SPLIT);
	} else {
		/* Scan size */
		write_file(SCAN_SIZE_FILE, SCAN_SIZE_NORMAL);
		/* Scan delay */
		write_file(SCAN_DELAY_FILE, SCAN_DELAY_NORMAL);
	}
	return;
}
/**
 * Sets the call type. This might also change the power save mode.
 *
 * @param type The call type as a string.
 */
void set_call_type(const char *type)
{
	guint previous_call_type;

	if (!type)
		return;
	
	previous_call_type = wlan_status.call_type;

	wlan_status.call_type &= ~CALL_CELLULAR;
	wlan_status.call_type &= ~CALL_VOIP;
	
	if (!strcmp(type, "none")) {
		set_scan_params(SCAN_NORMAL);		
		DLOG_DEBUG("Call cleared");
	} else if (!strcmp(type, "cellular")) {
		wlan_status.call_type |= CALL_CELLULAR;
		set_tx_power(wlan_status.conn.power_level);
		set_scan_params(SCAN_NORMAL);
		DLOG_DEBUG("Reduced power level");
	} else {
		wlan_status.call_type |= CALL_VOIP;
		set_scan_params(SCAN_SPLIT);
		DLOG_DEBUG("VOIP call");
	}
	
	/* We get multiple events for the call type so set the power
	   only once when a call is released.
	*/
	if ((previous_call_type & CALL_CELLULAR) && 
	    (!strcmp(type, "none") ||
	     (wlan_status.call_type & CALL_VOIP))) {
		if (wlan_status.conn.power_level) {
			set_tx_power(wlan_status.conn.power_level);
			DLOG_DEBUG("Restored power level");
		}
	}
	if (can_adjust_psm())
		set_power_state(wlan_status.powersave);
}
/**
 * Parses the context from the policy actions method params. This sets the
 * call type according to the call_audio_type variable.
 *
 * @param actit Dbus message iterator to the context.
 * @return TRUE on success, FALSE otherwise.
 */
int context_parser(DBusMessageIter *actit)
{
	DBusMessageIter  cmdit;
	DBusMessageIter  argit;
	DBusMessageIter  valit;
	char            *argname;
	char            *argval;
	char            *variable;
	char            *value;

	do {
		variable = value = NULL;
		
		dbus_message_iter_recurse(actit, &cmdit);

		do {
			if (dbus_message_iter_get_arg_type(&cmdit) !=
			    DBUS_TYPE_STRUCT)
				return FALSE;

			dbus_message_iter_recurse(&cmdit, &argit);

			if (dbus_message_iter_get_arg_type(&argit) !=
			    DBUS_TYPE_STRING)
				return FALSE;

			dbus_message_iter_get_basic(&argit, (void *)&argname);

			if (!dbus_message_iter_next(&argit))
				return FALSE;

			if (dbus_message_iter_get_arg_type(&argit) !=
			    DBUS_TYPE_VARIANT)
				return FALSE;

			dbus_message_iter_recurse(&argit, &valit);

			if (dbus_message_iter_get_arg_type(&valit) !=
			    DBUS_TYPE_STRING)
				return FALSE;

			dbus_message_iter_get_basic(&valit, (void *)&argval);

			if (!strcmp(argname, "variable")) {
				variable = argval;
			}
			else if (!strcmp(argname, "value")) {
				value = argval;
			}

			if (variable && !strcmp(variable, "call_audio_type"))
				set_call_type(value);

		} while (dbus_message_iter_next(&cmdit));

	} while (dbus_message_iter_next(actit));

	return TRUE;
}

/**
   Handles policy actions method call.
   @param msg DBUS message.
   @return result.
 */
static DBusHandlerResult handle_policy_actions(DBusMessage *msg)
{
	dbus_uint32_t    txid;
	char            *actname;
	DBusMessageIter  msgit;
	DBusMessageIter  arrit;
	DBusMessageIter  entit;
	DBusMessageIter  actit;
	int              success = TRUE;

	dbus_message_iter_init(msg, &msgit);

	if (dbus_message_iter_get_arg_type(&msgit) != DBUS_TYPE_UINT32)
		return DBUS_HANDLER_RESULT_HANDLED;

	dbus_message_iter_get_basic(&msgit, (void *)&txid);

	if (!dbus_message_iter_next(&msgit) ||
	    dbus_message_iter_get_arg_type(&msgit) !=
	    DBUS_TYPE_ARRAY) {
		success = FALSE;
		goto out;
	}

	dbus_message_iter_recurse(&msgit, &arrit);

	do {
		if (dbus_message_iter_get_arg_type(&arrit) !=
		    DBUS_TYPE_DICT_ENTRY) {
			success = FALSE;
			continue;
		}

		dbus_message_iter_recurse(&arrit, &entit);

		do {
			if (dbus_message_iter_get_arg_type(&entit) !=
			    DBUS_TYPE_STRING) {
				success = FALSE;
				continue;
			}

			dbus_message_iter_get_basic(&entit, (void *)&actname);

			if (!dbus_message_iter_next(&entit) ||
			    dbus_message_iter_get_arg_type(&entit)
			    != DBUS_TYPE_ARRAY) {
				success = FALSE;
				continue;
			}

			dbus_message_iter_recurse(&entit, &actit);
			
			if (dbus_message_iter_get_arg_type(&actit) !=
			    DBUS_TYPE_ARRAY) {
				success = FALSE;
				continue;
			}
			
			if (!strcmp(actname, "com.nokia.policy.context"))
				success &= context_parser(&actit);
			
		} while (dbus_message_iter_next(&entit));
		
	} while (dbus_message_iter_next(&arrit));

out:
	if (!success)
		DLOG_DEBUG("Failed to parse the policy actions message.");

	return DBUS_HANDLER_RESULT_HANDLED;
}

typedef DBusHandlerResult (*handler_func)(DBusMessage *message,
		DBusConnection *connection);

typedef struct {
	const char *interface;
	const char *name;
	handler_func func;
} method_handler_t;

static method_handler_t handlers[] = {
	{ WLANCOND_REQ_INTERFACE, WLANCOND_SETTINGS_AND_CONNECT_REQ,
		settings_and_connect_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_SCAN_REQ, scan_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_STATUS_REQ, status_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_INTERFACE_REQ, interface_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_CONNECTION_STATUS_REQ,
		connection_status_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_SET_PMKSA_REQ, set_pmksa_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_SET_POWERSAVE_REQ,
		set_powersave_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_DISCONNECT_REQ, disconnect_request},
	{ WLANCOND_REQ_INTERFACE, WLANCOND_DISASSOCIATE_REQ,
		disassociate_request},
	{ NULL }
};

/**
    Generic handler for D-Bus requests.
    @param message DBUS message.
    @param connection DBUS connection.
    @return status.
*/
DBusHandlerResult wlancond_req_handler(DBusConnection     *connection,
		DBusMessage        *message,
		void               *user_data)
{
	method_handler_t *handler;

	DLOG_DEBUG("Received %s.%s",
		   dbus_message_get_interface(message),
		   dbus_message_get_member(message));

	if (dbus_message_is_signal(message,
				   ICD_DBUS_INTERFACE,
				   ICD_STATUS_CHANGED_SIG))
		return icd_check_signal_dbus(message);
	
	else if (dbus_message_is_signal(message,
				   CSNET_OPER_IFACE,
				   CSNET_OPERATOR_CHANGED_SIG))
		return csd_check_signal_dbus(message);

	else if (dbus_message_is_signal(message,
	                           EAP_SIG_INTERFACE,
	                           EAP_AUTH_STATUS_SIG))
		return eap_check_signal_dbus(message);	
	else if (dbus_message_is_signal(message, 
					POLICY_SERVICE_NAME,
					POLICY_ACTIONS_SIG))
		return handle_policy_actions(message);	
	else if (dbus_message_is_signal(message,
					GPRS_DBUS_INTERFACE,
					GPRS_DBUS_DATACOUNTERS_SIG)) {
		return gprs_datacounters_signal(message);
	}
#ifdef USE_MCE_MODE
	else if (dbus_message_is_signal(message,
				   MCE_SIGNAL_IF,
				   MCE_RADIO_STATES_SIG)) {
		return mode_change_dbus(message);
	}
#ifdef ACTIVITY_CHECK
	else if (dbus_message_is_signal(message,
				   MCE_SIGNAL_IF,
				   MCE_INACTIVITY_SIG)) {
		return activity_check_dbus(message);
	}
#endif
#endif
	/* The rest should be just method calls */
	else if (dbus_message_get_type(message) != DBUS_MESSAGE_TYPE_METHOD_CALL) {
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}
	
	for (handler = handlers; handler->interface != NULL; handler++) {
		if (dbus_message_is_method_call(message,
						handler->interface,
						handler->name)) {
			DLOG_DEBUG("Received %s", handler->name);
			return handler->func(message, connection);
		}
	}
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static DBusObjectPathVTable wlancond_req_vtable = {
	.message_function    = wlancond_req_handler,
	.unregister_function = NULL
};


/**
    Create bindings for D-BUS handlers.
    @param connection DBUS connection.
*/
void init_dbus_handlers(DBusConnection *connection) {
	dbus_bool_t ret;
	ret = dbus_connection_register_object_path(connection,
						   WLANCOND_REQ_PATH,
						   &wlancond_req_vtable,
						   NULL);
	if (ret == FALSE)
		DLOG_ERR("dbus_connection_register_object_path failed");

#ifdef USE_MCE_MODE
	if (!add_mode_listener(connection))
		DLOG_ERR("Adding mode listener failed");
#endif
	if (!add_csd_listener(connection))
		DLOG_ERR("Adding csd listener failed");
}

/**
   Destroy D-BUS handlers.
   @param connection DBUS connection.
*/
void destroy_dbus_handlers(DBusConnection *connection) {
	dbus_connection_unregister_object_path(connection, WLANCOND_REQ_PATH);
}
