/**
   @file dbus-signal.c

   Copyright (C) 2004-2010 Nokia Corporation. All rights reserved.

   @author Janne Yllehto <janne.ylalehto@nokia.com>

   Portions of this file are
   Copyright (c) 1997-2002 Jean Tourrilhes <jt@hpl.hp.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <wlancond-dbus.h>
#include <eap-dbus.h>
#include <libnetfilter_log/libnetfilter_log.h>

#include "common.h"
#include "daemon.h"
#include "dbus.h"
#include "dbus-handler.h"
#include "log.h"
#include "dbus-helper.h"
#include "wpa.h"
#include "wapi.h"

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>

extern struct wlan_status_t wlan_status;

/* This structure has required fields for
   making a message to send the link mode 
   signal to the kernel.
*/
typedef struct rt_link_mode
{
	struct nlmsghdr nlh;
	struct ifinfomsg info;
	struct rtattr rta;
	char rta_data[4];
} rt_link_mode;

/** Netlink socket is global */
static int netlink_socket = -1;
/** NFLOG handle */
static struct nflog_handle *nf_handle;

#define NFLOG_PACKET_SIZE 512

/* Local functions */

void print_mac(guint priority, const char *message, guchar* mac)
{
	if (priority > WLANCOND_PRIO_MEDIUM)
		DLOG_INFO("%s %02x:%02x:%02x:%02x:%02x:%02x", message, mac[0],
			  mac[1], mac[2], mac[3], mac[4], mac[5]);
	else {
		DLOG_DEBUG("%s %02x:%02x:%02x:%02x:%02x:%02x", message, mac[0],
			   mac[1], mac[2], mac[3], mac[4], mac[5]);
	}
}
void clean_scan_results_item(gpointer data, gpointer user_data)
{
	struct scan_results_t *scan_results = data;
	g_free(scan_results->wpa_ie);
	g_free(scan_results->uuid_e);
	g_slice_free(struct scan_results_t, scan_results);
}

/**
  Remove saved scan results.
  @param scan_results_save structure where scan results are saved.
 */
void clean_scan_results(GSList **scan_results_save)
{
	//DLOG_DEBUG("Cleaning scan results");

	g_slist_foreach(*scan_results_save, clean_scan_results_item, NULL);
	g_slist_free(*scan_results_save);

	*scan_results_save = NULL;

}

/**
  Send scan results to DBUS.
  @param scan_results_save structure where scan results are saved.
  @param sender The message is sent to this entity.
  @param number_of_results number of scan results.
 */
void send_dbus_scan_results(GSList *scan_results_save, const char* sender,
			    dbus_int32_t number_of_results)
{
	DBusMessage *results;
	DBusMessageIter iter, sub;
	GSList *list;
	int list_count = 0;
	unsigned char* v;
	char* p;

	if (sender == NULL || strnlen(sender, 5) == 0)
		return;

	DLOG_INFO("Scan results (%d APs) to %s", number_of_results, sender);

	results = new_dbus_signal(WLANCOND_SIG_PATH,
				  WLANCOND_SIG_INTERFACE,
				  WLANCOND_SCAN_RESULTS_SIG,
				  sender);
	
	dbus_message_iter_init_append(results, &iter);
	
	if (number_of_results > WLANCOND_MAX_NETWORKS) {
		DLOG_DEBUG("Limiting result %d to %d", number_of_results,
			   WLANCOND_MAX_NETWORKS);
		number_of_results = WLANCOND_MAX_NETWORKS;
	}
	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_INT32,
					    &number_of_results))
		die("Out of memory");
	
	for (list = scan_results_save; list != NULL && list_count++ <=
		     number_of_results; list = list->next) {
		struct scan_results_t *scan_results = list->data;
		DLOG_DEBUG("AP (%d) is %s, rssi:%d channel:%d cap:%08x",
			   list_count,
			   scan_results->ssid,
			   scan_results->rssi, scan_results->channel,
			   scan_results->cap_bits);
		
		p = scan_results->ssid;

		if (!dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
						      "y", &sub))
			die("Out of memory");
		if (!dbus_message_iter_append_fixed_array(
			    &sub, DBUS_TYPE_BYTE, &p,
			    scan_results->ssid_len))
			die("Out of memory");
		if (!dbus_message_iter_close_container(&iter, &sub))
			die("Out of memory");

		v = scan_results->bssid;
		if (!dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
						      "y", &sub))
			die("Out of memory");
		if (!dbus_message_iter_append_fixed_array(&sub, DBUS_TYPE_BYTE,
							  &v, ETH_ALEN))
			die("Out of memory");
		if (!dbus_message_iter_close_container(&iter, &sub))
			die("Out of memory");
		if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_INT32,
						    &scan_results->rssi))
			die("Out of memory");
		if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32,
						    &scan_results->channel))
			die("Out of memory");
		if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32,
						    &scan_results->cap_bits))
			die("Out of memory");
	}

	send_and_unref(get_dbus_connection(), results);
}

/**
  Send disconnected signal.
 */
void disconnected_signal(void)
{
	DBusMessage *disconnected;

	disconnected = new_dbus_signal(
		WLANCOND_SIG_PATH,
		WLANCOND_SIG_INTERFACE,
		WLANCOND_DISCONNECTED_SIG,
		NULL);
	
	gchar* ifname = wlan_status.ifname;
	
	append_dbus_args(disconnected,
			 DBUS_TYPE_STRING, &ifname,
			 DBUS_TYPE_INVALID);
	
	send_and_unref(get_dbus_connection(), disconnected);
}

/**
  Send connected signal.
  @param scan_results Scan results to be sent.
 */
void connected_signal(char* bssid)
{
	DBusMessage *connected;
	dbus_int32_t auth_status = get_encryption_info();
	dbus_uint32_t wlan_mode = wlan_status.conn.mode;

	if (!(wlan_mode & (WLANCOND_ADHOC | WLANCOND_INFRA))) {
		DLOG_ERR("WLAN mode unknown");
		return;
	}

	connected = new_dbus_signal(
		WLANCOND_SIG_PATH,
		WLANCOND_SIG_INTERFACE,
		WLANCOND_CONNECTED_SIG,
		NULL);
	
	gchar* ifname = wlan_status.ifname;

	append_dbus_args(connected,
			 DBUS_TYPE_STRING, &ifname,
			 DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &bssid, ETH_ALEN,
			 DBUS_TYPE_INT32, &auth_status,
			 DBUS_TYPE_UINT32, &wlan_mode,
			 DBUS_TYPE_INVALID);
	
	send_and_unref(get_dbus_connection(), connected);
}

/**
  Check if we have scanned in last WLANCOND_MIN_ROAM_SCAN_INTERVAL.
  @return TRUE if we have scanned, FALSE otherwise.
 */
gboolean roam_scanning(void) {
	struct timeval tv;

	if (gettimeofday(&tv, NULL) < 0)
		return FALSE;

	if (tv.tv_sec > wlan_status.last_scan + WLANCOND_MIN_ROAM_SCAN_INTERVAL)
		return FALSE;

	DLOG_DEBUG("Already roam scanning");
	return TRUE;
}
/**
   Get connected state.
   @return TRUE if connected, FALSE otherwise.
*/
gboolean is_connected_state(void) {
	guint state_v = get_wlan_state();
	
	if (state_v == WLAN_INITIALIZED_FOR_CONNECTION ||
	    state_v == WLAN_NO_ADDRESS ||
	    state_v == WLAN_CONNECTED)
		return TRUE;
	return FALSE;
}
/**
  Netlink event handling continues, now we know that we have a message.
  @param hdr Pointer to message header.
 */
static void handle_message(struct nlmsghdr *hdr)
{
	struct ifinfomsg *infomsg;
	unsigned ifindex;	

	if (hdr->nlmsg_type != RTM_DELLINK && hdr->nlmsg_type != RTM_NEWLINK)
		return;  
	if (hdr->nlmsg_len - sizeof(*hdr) < sizeof(*infomsg))
		return;

	infomsg = NLMSG_DATA(hdr);
	ifindex = infomsg->ifi_index;

	if (wlan_status.ifindex &&
	    hdr->nlmsg_type == RTM_DELLINK &&
	    ifindex == wlan_status.ifindex) {
		
		DLOG_DEBUG("Current interface in use, disconnecting...");
		
		wlan_status.ifindex = 0;

		set_wlan_state(WLAN_NOT_INITIALIZED,
			       DISCONNECTED_SIGNAL,
			       FALSE);

		clean_dbus_handler();
	}
	else if (hdr->nlmsg_type == RTM_NEWLINK) {
		if (ifindex == wlan_status.ifindex) {
			/* Check if the interface state has changed and if it
			   was not up */
			if ((infomsg->ifi_change & IFF_UP) && 
			    !(infomsg->ifi_flags & IFF_UP)) {
				/* ifconfig wlan0 down */
				if (wlan_status.if_down_self == TRUE) {
					wlan_status.if_down_self = FALSE;
				} else {
					DLOG_DEBUG("If down by someone else?");
					int signal = NO_SIGNAL;

					if (is_connected_state() == TRUE)
						signal = DISCONNECTED_SIGNAL;
					set_wlan_state(WLAN_NOT_INITIALIZED,
						       signal,
						       FALSE);
				}
			}
		} else {
			char ifname[IFNAMSIZ+1];
			
			if (if_indextoname(ifindex, ifname) == NULL)
				return;
			if (strncmp(wlan_status.ifname, ifname, sizeof(ifname)))
				return;
			
			DLOG_DEBUG("Interface added");
			
			clean_dbus_handler();
			run_calibration();
			wlan_status.ifindex = ifindex;
		}
	}
}

/**
  Start netlink event handling
  @param fd File descriptor.
 */
static void handle_netlink_event(int fd)
{
	char buf[1024];
	struct sockaddr_nl nl;
	socklen_t nl_len = sizeof(struct sockaddr_nl);
	int res;

	while (1) {
		res = recvfrom (fd, buf, sizeof(buf), MSG_DONTWAIT,
				(struct sockaddr*)&nl, &nl_len);

		/* Error */
		if (res < 0) {
			if (errno != EINTR && errno != EAGAIN) {
				DLOG_ERR("Error reading netlink socket");
			}
			/* Don't do anything */
			return;
		}

		/* EOF */
		if (res == 0) {
			return;
		}
		int len;
		struct nlmsghdr *hdr = (struct nlmsghdr*)buf;
		/* real handling in this loop */
		while (res >= (int)sizeof(*hdr))
		{
			len = hdr->nlmsg_len;

			if ((len - (int)sizeof(*hdr) < 0) || len > res) {
				DLOG_ERR("Error in netlink message length");
				break;
			}
			/* Ok, we have good message */
			handle_message(hdr);

			/* Get ready for next message */
			len = NLMSG_ALIGN(len);
			res -= len;
			hdr = (struct nlmsghdr*)((char*)hdr+len);
		}
	}
}

/**
  Initialize wireless interface
  @param rth private struct.
  @return status.
 */
static int init_wi (struct sockaddr_nl *rth)
{
	unsigned int addr_len;

	netlink_socket = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (netlink_socket < 0) {
		DLOG_ERR("socket()");
		return -1;
	}
	memset(rth, 0, sizeof(struct sockaddr_nl));

	rth->nl_family = AF_NETLINK;
	rth->nl_groups = RTMGRP_LINK;

	if (bind(netlink_socket, (struct sockaddr*)rth, 
		 sizeof(struct sockaddr_nl)) < 0) {
		DLOG_ERR("bind()");
		return -1;
	}
	addr_len = sizeof(struct sockaddr_nl);
	if (getsockname(netlink_socket, (struct sockaddr*)rth, &addr_len) < 0) {
		DLOG_ERR("Cannot getsockname");
		return -1;
	}
	if (addr_len != sizeof(struct sockaddr_nl)) {
		DLOG_ERR("Wrong address length %d", addr_len);
		return -1;
	}
	if (rth->nl_family != AF_NETLINK) {
		DLOG_ERR("Wrong address family %d", rth->nl_family);
		return -1;
	}

	return 0;
}

/**
  Callback function for wireless events.
  @param chan GLIB IO Channel.
  @param cond GLIB IO condition.
  @param data private pointer.
  @return status.
 */
static gboolean _monitor_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
{
	int fd;

	if (cond != G_IO_IN) {
		guint watch_id = *((guint *)data);
		DLOG_ERR("Error message from wireless interface");
		g_source_remove(watch_id);
		g_io_channel_unref(chan);
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(chan);
	if (fd >= 0) {
		handle_netlink_event(fd);
	}

	return TRUE;
}

/**
  Callback function for NFLOG events.
  @param chan GLIB IO Channel
  @param cond GLIB IO condition
  @param data privat pointer.
  @return status.
 */
static gboolean _monitor_nflog_cb(GIOChannel *chan, GIOCondition cond, 
				  gpointer data)
{
	int fd;
	int len;
	char buf[NFLOG_PACKET_SIZE];

	if (cond != G_IO_IN) {
		guint watch_id = *((guint *)data);
		DLOG_ERR("Error message from NFLOG interface");
		g_source_remove(watch_id);
		g_io_channel_unref(chan);
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(chan);
	if (fd >= 0) {
		len = recv(fd, &buf, sizeof(buf), 0);
		if (len >= 0) {
			nflog_handle_packet(nf_handle, buf, len);
		}
	}
	return TRUE;
}

static int nflog_cb(struct nflog_g_handle *gh, struct nfgenmsg *nfmsg,
		    struct nflog_data *nfa, void *data) {
	
	if (!nfa)
		return 0;
	
	const char* p = nflog_get_prefix(nfa);
	if (p) {
		DLOG_DEBUG("NFLOG CB: %s", p);
		if (strncmp(p, "ABOVE", 5) == 0) {
			wlan_status.high_traffic = TRUE;
		} else if (strncmp(p, "BELOW", 5) == 0) {
			wlan_status.high_traffic = FALSE;	
		} else {
			return 0;
		}
		
		if (get_wlan_state() == WLAN_CONNECTED)
			set_power_state(wlan_status.powersave);
	}
	return 0;
}
/**
   Starts monitoring of NFLOG events.
   @return status.
*/
static int init_nflog(void) {
	static guint nflog_watch_id = 0;
	struct nflog_g_handle *group; 
	GIOChannel *gio;
	int fd;

	nf_handle = nflog_open();
	if (!nf_handle)
		return -1;

	group = nflog_bind_group(nf_handle, 0);
	if (!group) {
		nflog_close(nf_handle);
		return -1;
	}

	if (nflog_set_mode(group, NFULNL_COPY_NONE, NFLOG_PACKET_SIZE) < 0) {
		nflog_unbind_group(group);
		nflog_close(nf_handle);
		return -1;
	}
	
	nflog_callback_register(group, &nflog_cb, NULL);
	
	fd = nflog_fd(nf_handle);
	gio = g_io_channel_unix_new(fd);
	g_io_channel_set_close_on_unref(gio, TRUE);
	nflog_watch_id = g_io_add_watch(gio, G_IO_IN | G_IO_PRI | G_IO_ERR |
					G_IO_HUP, _monitor_nflog_cb, 
					&nflog_watch_id);

	return 0;
}

#define WLAN_IPTABLES "/etc/network/wlan_iptables.sh"
#define ENABLE_IPTABLES "/proc/net/nf_condition/enable_throughput"
/**
   Set IPTABLES state.
   @param state New state.
*/
static void set_iptables_state(unsigned int state) {	
	char buf[4];

	memset(&buf, 0, sizeof(buf));

	DLOG_DEBUG("Setting iptables to %s", state==1?"ON":"OFF");

	if (wlan_status.iptables_file == NULL) {
		wlan_status.iptables_file = fopen(ENABLE_IPTABLES, "w");
		if (wlan_status.iptables_file == NULL) {
			DLOG_DEBUG("Cannot open: %s", ENABLE_IPTABLES);
			return;
		}
	}

	snprintf(buf, 2, "%d", state);
	
	if (fwrite(buf, 2, 1, wlan_status.iptables_file) != 1) {
		DLOG_ERR("Could not write to: %s", ENABLE_IPTABLES);
	}

	fflush(wlan_status.iptables_file);
}
/**
   Set iptables inactive.
*/
void remove_iptables(void) {
	
	/* Currently we just set the enabler to zero */
	set_iptables_state(0);
}
/**
   Load iptables and set them active.
*/
int set_iptables(void) {
	if (wlan_status.iptables_loaded == FALSE) {
		/* Also check that the file does not exist.
		   If wlancond crashed the rules are already
		   loaded.
		*/
		FILE *file = fopen(ENABLE_IPTABLES, "r");
		if (file == NULL) {
			
			DLOG_DEBUG("Loading new rules");
			
			if (system("sh " WLAN_IPTABLES) < 0) {
				DLOG_ERR("Running iptables failed");
				return -1;
			}
			wlan_status.iptables_loaded = TRUE;
		} else {
			fclose(file);
		}
	}

	set_iptables_state(1);

	return 0;
}

/**
  Starts monitoring of wireless events.
  @return status.
 */
gboolean monitor_wi(void) {
	static guint watch_id = 0;
	GIOChannel *gio;
	struct sockaddr_nl rth;

	if (init_wi(&rth) < 0)
		return FALSE;

	if (init_nflog() < 0)
		DLOG_ERR("NFLOG init failed");

	gio = g_io_channel_unix_new(netlink_socket);
	g_io_channel_set_close_on_unref(gio, TRUE);
	watch_id = g_io_add_watch(gio, G_IO_IN | G_IO_PRI | G_IO_ERR |
			G_IO_HUP, _monitor_cb, &watch_id);
	return TRUE;
}

/**
  Sends the link mode signal to the kernel.
  @param lmode mode status.
  @return status.
 */
int netlink_send_linkmode(int lmode)
{
	struct rt_link_mode msg;
	static int seq;
	int ret;

	memset(&msg, 0, sizeof(msg));

	msg.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
	msg.nlh.nlmsg_type = RTM_SETLINK;
	msg.nlh.nlmsg_flags = NLM_F_REQUEST;
	msg.nlh.nlmsg_seq = ++seq;
	msg.nlh.nlmsg_pid = 0;

	msg.info.ifi_family = AF_UNSPEC;
	msg.info.ifi_type = 0;
	msg.info.ifi_index = wlan_status.ifindex;
	msg.info.ifi_flags = 0;
	msg.info.ifi_change = 0;

	/* let's happily mix operstate and linkmode...*/
	msg.rta.rta_len = RTA_LENGTH(sizeof(char));
	*((char *)RTA_DATA(&msg.rta)) = lmode;
	msg.nlh.nlmsg_len = NLMSG_ALIGN(msg.nlh.nlmsg_len) +
		RTA_LENGTH(sizeof(char));

	if (lmode == WLANCOND_IF_DORMANT) {
		msg.rta.rta_type = IFLA_LINKMODE;
	} else {
		msg.rta.rta_type = IFLA_OPERSTATE;
	}

	DLOG_DEBUG("Linkmode set to: \"%s\"", 
		   lmode==WLANCOND_IF_DORMANT?"Dormant":"Up");

	ret = send(netlink_socket, &msg, msg.nlh.nlmsg_len, 0);
	if (ret < 0) {
		DLOG_ERR("Sending link mode failed:"
			   " %d (%s)", ret, strerror(-ret));
		return -1;
	}

        return 0;
}

/**
   Callback function for power save flag change.
   @param type notifying GConfClient. 
   @param id connection id.
   @param key GConfEntry.
   @param value ConnSettingsValue.
   @param userData user data for call back.
*/
static void monitor_full_ps_cb(ConnSettingsType type, 
			       const char *id, 
			       const char *key, 
			       ConnSettingsValue *value, 
			       void *user_data) {
	
	if (value != NULL && key != NULL) {
		if (strcmp(key, FULL_PSM_SETTING) == 0) {
			if (value->type == CONN_SETTINGS_VALUE_BOOL) { 
				wlan_status.full_psm_enabled = 
					value->value.bool_val;
				DLOG_DEBUG("Full power save flag is %s", 
					   value->value.bool_val==TRUE?
					   "enabled":"disabled");

				if (can_adjust_psm())
					set_power_state(wlan_status.powersave);
			}
		}
	}
}

/**
  Starts monitoring the power save flag change.
  @return status.
 */
gboolean monitor_iap_keys() {

	gboolean error;

	wlan_status.full_psm_enabled = get_setting_bool(
		FULL_PSM_SETTING, &error);

	if (conn_settings_add_notify(
		    wlan_status.setting, 
		    (ConnSettingsNotifyFunc*) monitor_full_ps_cb, 
		    NULL) != TRUE) {
		DLOG_ERR("Adding notifier for full ps failed");
		return FALSE;
  	}
	
	return TRUE;
}
