/**
  @file libnl-handler.c

  Copyright (C) 2010 Nokia Corporation. All rights reserved.

  @author Usman Rahim <usman.rahim@nokia.com>

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/if_ether.h>
#include <dbus/dbus.h>
#include <wlancond.h>
#include <eap-dbus.h>

#include <glib.h>
#include <glib-object.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "wpa.h"
#include "wapi.h"
#include "daemon.h"
#include "dbus-handler.h"
#include "log.h"
#include "libnl-handler.h"

static int nl80211_scan_result_handler(struct nl_msg *msg, void* arg);
static int nl80211_parse_ies(unsigned char* const ie, size_t ie_len, scan_results_t* const scan_result);
static void save_caps(struct scan_results_t *, struct ap_info_t *, unsigned char *, unsigned int);

/** LIBNL */
static struct nl_handle *nl_handle = NULL;
static struct nl_handle *nl_event_handle = NULL;
static struct nl_cache *nl_cache = NULL;
static struct nl_cache *nl_event_cache = NULL;
static struct genl_family *nl80211 = NULL;
static struct nl_cb *nl_cb = NULL;
static unsigned int nl_family_id;
static wlan_status_t * scan_wlan_status = NULL;

/**
   Read group name and ID returned by the get family command.
   @param msg nl_msg structure.
   @param arg Updated with received data.
   @return status.
*/
static int get_family_cb(struct nl_msg *msg, void *arg)
{
    struct get_family_t *data = arg;
    struct nlattr *tb[CTRL_ATTR_MAX + 1];
    struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
    struct nlattr *grp;
    int index;
     
    if (nla_parse(tb, CTRL_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		  genlmsg_attrlen(gnlh, 0), NULL)) {
	    DLOG_ERR("nl80211: Failed to parse get family attributes!");
	    return NL_SKIP;
    }	
    
    if (!tb[CTRL_ATTR_MCAST_GROUPS])
            return NL_SKIP;
         
    nla_for_each_nested(grp, tb[CTRL_ATTR_MCAST_GROUPS], index) {
	    struct nlattr *tbg[CTRL_ATTR_MCAST_GRP_MAX + 1];
	    if (nla_parse(tbg, CTRL_ATTR_MCAST_GRP_MAX, nla_data(grp),
			  nla_len(grp), NULL)) {
		    DLOG_ERR("nl80211: Failed to parse nested get "
			     "family attributes!");
		    return NL_SKIP;
		}
	    
	    if (!tbg[CTRL_ATTR_MCAST_GRP_NAME] ||
		!tbg[CTRL_ATTR_MCAST_GRP_ID] ||
		strncmp(nla_data(tbg[CTRL_ATTR_MCAST_GRP_NAME]),
			data->group_name, 
			nla_len(tbg[CTRL_ATTR_MCAST_GRP_NAME])) != 0)
		    continue;
	    
	    data->id = nla_get_u32(tbg[CTRL_ATTR_MCAST_GRP_ID]);
	    break;
    };
    
    return NL_SKIP;
}

/**
  Ask for group name and id using get family command.
  @param family_name Family to which group belongs.
  @param group_name Group to be extracted.
  @return status.
 */
static int nl80211_get_family(const char *family_name, const char *group_name)
{
    struct get_family_t data = { group_name, -ENOENT };
    struct nl_msg *msg;
    int ret = -1;
    void *hdr;
    
    int id = genl_ctrl_resolve(nl_event_handle, "nlctrl");
    if (id < 0) {
	    DLOG_ERR("nl80211: Failed to put nl80211 message");
	    return ret;
    }
    
    msg = nlmsg_alloc();
    if (!msg)
            return ret;
    
    hdr = genlmsg_put(msg, getpid(), 0, id, 0, 0, CTRL_CMD_GETFAMILY, 0);
    if (hdr == NULL) {
	    DLOG_ERR("nl80211: Failed to put nl80211 message");
	    goto nla_put_failure;
    }
    
    NLA_PUT_STRING(msg, CTRL_ATTR_FAMILY_NAME, family_name);
    
    ret = nl_send_recv_msgs(msg, get_family_cb, &data);
    if (ret) {
	    DLOG_ERR("nl80211: failed to send get family ret=%d (%s)", ret,
		     strerror(-ret));
	    goto nla_put_failure;
    }
    
    if (ret == 0)  
	    ret = data.id;
    
nla_put_failure:
    nlmsg_free(msg); 
    return ret;
} 

/**
  Handle netlink wireless event.
  @param sa_data MAC address of the connected AP.
  @return status.
 */
static void handle_netlink_wap_event(unsigned char *sa_data)
{
	print_mac(WLANCOND_PRIO_HIGH, "nl80211: BSSID:", sa_data);

	// Check if the address is valid
	if (memcmp(sa_data, NULL_BSSID, ETH_ALEN)) {

		if (get_wlan_state() == WLAN_INITIALIZED_FOR_SCAN ||
		    get_wlan_state() == WLAN_NOT_INITIALIZED) {
			DLOG_DEBUG("Connection opened while not initialized?"); 
			nl80211_get_phy_info(wlan_status.ifindex);
			connected_signal((char *)sa_data);
			return;
		}

		remove_connect_timer();

		set_wlan_signal(WLANCOND_HIGH);

		connected_signal((char *)sa_data);

		if (get_wpa_mode() == TRUE ||
		    wlan_status.conn.authentication_type ==
		    EAP_AUTH_TYPE_WFA_SC) {
			if (wpa_ie_push((char *) sa_data) < 0) {
				set_wlan_state(WLAN_NOT_INITIALIZED,
					       DISCONNECTED_SIGNAL,
					       FALSE);
				return;
			}
		} else {
			/* In WPA mode retry count is zeroed only after
			   succesful authentication and we get information
			   about that when the supplicant sends EAP_AUTH_STATUS
			   signal which we catch.
			*/
			wlan_status.retry_count = 0;
			
			if (netlink_send_linkmode(WLANCOND_IF_READY) != 0) {
				DLOG_ERR("Setting oper state to "
					 "ready failed");
			}
		}

		// If not roaming, set NO_ADDRESS state
		if (wlan_status.ip_ok == FALSE) {
			/* Set iptables only if we are not active */
			if (get_inactivity_status() == TRUE) {
				if (set_iptables() < 0)
					DLOG_ERR("Setting iptables failed");
			}
			set_wlan_state(WLAN_NO_ADDRESS, NO_SIGNAL, TRUE);
		} else {
			set_wlan_state(WLAN_CONNECTED, NO_SIGNAL, TRUE);
		}

		/* Clear AUTOCONNECT flag */
		wlan_status.conn.flags &= ~WLANCOND_AUTOCONNECT;
	} else {
		if (get_wlan_state() == WLAN_INITIALIZED_FOR_CONNECTION &&
		    get_scan_state() == SCAN_ACTIVE) {
			/* We are alredy searching for better connection */
			return;
		}
		if (get_wlan_state() != WLAN_CONNECTED &&
		    get_wlan_state() != WLAN_NO_ADDRESS &&
		    get_wlan_state() != WLAN_INITIALIZED_FOR_CONNECTION) {
			DLOG_ERR("nl80211: Not even connected?!?");
			disconnected_signal();
			return;
		}
		/* No roaming in WPS state */
		if (wlan_status.conn.authentication_type ==
		    EAP_AUTH_TYPE_WFA_SC) {
			return;
		}

		// Disconnect supplicant
		if (get_wpa_mode() == TRUE &&
		    (get_wlan_state() == WLAN_CONNECTED ||
		     get_wlan_state() == WLAN_NO_ADDRESS)) {
			disassociate_eap();
		}

		DLOG_DEBUG("nl80211: Trying to find a new connection");

		if (is_connected_state()) {
			DLOG_DEBUG("Disconnecting previous connection");
			mlme_command(wlan_status.conn.bssid, 
				     WLANCOND_NL80211_MLME_DISCONNECT, 
				     WLANCOND_REASON_LEAVING);
		}

		/* Decrese failed AP signal */

		roam_cache_key_t key = { wlan_status.conn.bssid, wlan_status.conn.channel };
		decrease_signal_in_roam_cache(&key);

		set_wlan_state(WLAN_INITIALIZED_FOR_CONNECTION, NO_SIGNAL,
			       TRUE);

		int ret = find_connection_and_associate(wlan_status.roam_cache,
							FALSE, FALSE, FALSE);
		if (ret == 0)
			return;
		
		if (ret == ESUPPLICANT) {
			set_wlan_state(WLAN_NOT_INITIALIZED,
				       DISCONNECTED_SIGNAL, FALSE);
			return;
		}
		/* Break eternal loop if all APs are failing */
		if (scan_retries_left() == FALSE) {
			DLOG_ERR("nl80211: Too many failures: %d",
				 wlan_status.retry_count);
			set_wlan_state(WLAN_NOT_INITIALIZED,
				       DISCONNECTED_SIGNAL,
				       FALSE);
			return;
		}

		/* No luck, start scanning */
		scan(wlan_status.conn.ssid, wlan_status.conn.ssid_len, TRUE);
	}
}

static int ignore_seq_check(struct nl_msg *msg, void *arg)
{
        return NL_OK;
}

/**
  Process connection quality management events.
  @param tb Received event attributes.
*/
static void process_cqm_event(struct nlattr *tb[])
{
	struct nlattr *monitor[NL80211_ATTR_CQM_MAX + 1];
	struct nla_policy cqm_policy[NL80211_ATTR_CQM_MAX + 1] = {
                [NL80211_ATTR_CQM_RSSI_THOLD] = { .type = NLA_U32 },
                [NL80211_ATTR_CQM_RSSI_HYST] = { .type = NLA_U8 },
                [NL80211_ATTR_CQM_RSSI_THRESHOLD_EVENT] = { .type = NLA_U32 },
        };

	if (nla_parse_nested(monitor, NL80211_ATTR_CQM_MAX, 
			     tb[NL80211_ATTR_CQM], cqm_policy)) {
		DLOG_ERR("nl80211: Failed to parse nested cqm attributes!");
		return;
	}

	if (monitor[NL80211_ATTR_CQM_RSSI_THRESHOLD_EVENT]) {
		if (nla_get_u32(monitor[NL80211_ATTR_CQM_RSSI_THRESHOLD_EVENT]) == 
				NL80211_CQM_RSSI_THRESHOLD_EVENT_HIGH) {
			DLOG_INFO("High signal");
			set_wlan_signal(WLANCOND_HIGH);
		} else {
			DLOG_INFO("Low signal");
			set_wlan_signal(WLANCOND_LOW);
			if (get_wlan_state() == WLAN_CONNECTED) {
				if (roam_scanning())
					// Already roam scanned 
					schedule_scan(WLANCOND_MIN_ROAM_SCAN_INTERVAL);
				else
					schedule_scan(WLANCOND_INITIAL_ROAM_SCAN_DELAY);
			}
		}
	}
}

/**
  Print or process netlink wireless events.
  @param msg nl_msg structure.
  @param arg arguments for the function.
  @return status.
*/
static int print_netlink_event_token(struct nl_msg *msg, void *arg)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];

	if (nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		      genlmsg_attrlen(gnlh, 0), NULL)) {
		DLOG_ERR("nl80211: Failed to parse netlink event attributes!");
		return NL_SKIP;
	}

	if (tb[NL80211_ATTR_IFINDEX]) {
		if (nla_get_u32(tb[NL80211_ATTR_IFINDEX]) != wlan_status.ifindex) {
			DLOG_DEBUG("nl80211: Ignoring, not a wireless event.");
			return NL_SKIP;
		}
	}

	switch (gnlh->cmd) {
	case NL80211_CMD_TRIGGER_SCAN:
		DLOG_DEBUG("nl80211: scan triggered");
		break;
	case NL80211_CMD_NEW_SCAN_RESULTS:
	case NL80211_CMD_SCAN_ABORTED:
		nl80211_scan_ready();
		break;
	case NL80211_CMD_AUTHENTICATE:
	case NL80211_CMD_ASSOCIATE:
		if (tb[NL80211_ATTR_TIMED_OUT]) {
			DLOG_ERR("authentication/association timed out");
			handle_netlink_wap_event(NULL_BSSID);
		}
		break;
	case NL80211_CMD_DEAUTHENTICATE:
	case NL80211_CMD_DISASSOCIATE:
		break;
	case NL80211_CMD_CONNECT:
	case NL80211_CMD_ROAM:
		if (nla_get_u16(tb[NL80211_ATTR_STATUS_CODE]) == 0) {
			unsigned char *bssid = nla_data(tb[NL80211_ATTR_MAC]);
			handle_netlink_wap_event(bssid);
		} else {
			DLOG_DEBUG("nl80211: Failed to connect. reason: %d", 
				   nla_get_u16(tb[NL80211_ATTR_STATUS_CODE]));
			handle_netlink_wap_event(NULL_BSSID);
		}
		break;
	case NL80211_CMD_DISCONNECT:
		if (tb[NL80211_ATTR_DISCONNECTED_BY_AP]) {
			DLOG_DEBUG("nl80211: Disconnected by the access point.");
			handle_netlink_wap_event(NULL_BSSID);
		} else {
			if (tb[NL80211_ATTR_REASON_CODE]) {
				DLOG_DEBUG("nl80211: Disconnected by local choice due to reason: %d",
					   nla_get_u16(tb[NL80211_ATTR_REASON_CODE]));
			} else {
				DLOG_DEBUG("nl80211: Disconnected");
				if (get_wlan_state() == WLAN_NO_ADDRESS ||
				    get_wlan_state() == WLAN_CONNECTED)
					handle_netlink_wap_event(NULL_BSSID);
			}
		}
        break;
	case NL80211_CMD_MICHAEL_MIC_FAILURE:
		DLOG_DEBUG("nl80211: MIC failure event received.");
		dbus_bool_t key_type = TRUE;
		
		if (tb[NL80211_ATTR_KEY_TYPE]) {
                	enum nl80211_key_type type = nla_get_u32(tb[NL80211_ATTR_KEY_TYPE]);
	
	                if (type == NL80211_KEYTYPE_PAIRWISE)
				key_type = FALSE;
	        }
		DLOG_DEBUG("MIC failure event for %s key", 
			   key_type==FALSE?"unicast":"group");
		if ((wlan_status.conn.encryption & WLANCOND_ENCRYPT_METHOD_MASK)
		    != WLANCOND_WAPI_PSK &&
		    (wlan_status.conn.encryption & WLANCOND_ENCRYPT_METHOD_MASK)
		    != WLANCOND_WAPI_CERT) {
		        handle_mic_failure(key_type, wlan_status.conn.bssid);
		}
		break;
	case NL80211_CMD_JOIN_IBSS:
		if (tb[NL80211_ATTR_MAC]) {
			unsigned char *bssid = nla_data(tb[NL80211_ATTR_MAC]);
			handle_netlink_wap_event(bssid);
		} else {
			DLOG_ERR("nl80211: Joined adhoc with no address.");
		}
		break;
	case NL80211_CMD_GET_STATION:
	case NL80211_CMD_SET_STATION:
	case NL80211_CMD_NEW_STATION:
	case NL80211_CMD_DEL_STATION:
		break;
	case NL80211_CMD_NOTIFY_CQM:
		process_cqm_event(tb);		
		break;
	default:
		DLOG_DEBUG("nl80211: unknown event cmd (%d)", gnlh->cmd);
		break;
	}
	return NL_SKIP;
}

/**
  Callback function for netlink wireless events.
  @param chan GLIB IO Channel.
  @param cond GLIB IO condition.
  @param data private pointer.
  @return status.
*/
static gboolean event_receive(GIOChannel *chan, GIOCondition cond, gpointer data)
{
	struct nl_cb *cb;
	int ret;

	cb = nl_cb_clone(nl_cb);

	if (!cb) {
		DLOG_ERR("nl80211: event receive cb clone fail");
		return TRUE;
	}

	if (cond != G_IO_IN) {
		DLOG_INFO("nl80211: received event is not g_io_in");
		guint watch_id = *((guint *)data);
		g_source_remove(watch_id);
		g_io_channel_unref(chan);
		goto nla_put_failure;
	}

	ret = nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, ignore_seq_check, NULL);
	if (ret) {
		DLOG_ERR("nl80211: ignore seq cb set returned error: %d", ret);		
		goto nla_put_failure;
	}	
	ret = nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, print_netlink_event_token, NULL);
	if (ret) {
		DLOG_ERR("nl80211: print netlink event cb set returned error: %d", ret);		
		goto nla_put_failure;
	}
	if (fcntl(nl_socket_get_fd(nl_event_handle), F_SETFL, O_NONBLOCK) == -1) {
		DLOG_ERR("nl80211: failed to set receive netlink message");
		goto nla_put_failure;
	}
	if (nl_recvmsgs(nl_event_handle, cb)) {
		DLOG_ERR("nl80211: failed to receive netlink message");
		goto nla_put_failure;
	}
	nl_cb_put(cb);
	return TRUE;

nla_put_failure:
	nl_cb_put(cb);
	return FALSE;
}

/**
   initialize the netlink library sockets
   @return status.
*/
int init_nl80211(void)
{
	int ret = -1;
	static guint watch_id = 0;
	GIOChannel *gio;

	/* Initialize generic netlink and nl80211 */

	nl_cb = nl_cb_alloc(NL_CB_DEFAULT);
	if (!nl_cb) {
		DLOG_ERR("nl80211: failed to allocate netlink callbacks");
		return ret;
	}

	nl_handle = nl_handle_alloc();
	if (!nl_handle) {
		DLOG_ERR("nl80211: failed to allocate netlink handle");
		goto out_cb_clear;
	}

	nl_event_handle = nl_handle_alloc_cb(nl_cb);
	if (!nl_event_handle) {
		DLOG_ERR("nl80211: failed to allocate netlink event handle");
		goto out_handle_destroy_b;
	}

	if (genl_connect(nl_handle)) {
		DLOG_ERR("nl80211: failed to connect netlink");
		goto out_handle_destroy_a;
	}

	nl_disable_sequence_check(nl_event_handle);

	if (genl_connect(nl_event_handle)) {
		DLOG_ERR("nl80211: failed to connect netlink event");
		goto out_handle_destroy_a;
	}

	nl_cache = genl_ctrl_alloc_cache(nl_handle);
	if (!nl_cache) {
		DLOG_ERR("nl80211: failed to allocate netlink control cache");
		goto out_handle_destroy_a;
	}
	
	nl_event_cache = genl_ctrl_alloc_cache(nl_event_handle);
	if (!nl_event_cache) {
		DLOG_ERR("nl80211: failed to allocate netlink event control cache");
		goto out_clean_cache_b;
	}

	nl80211 = genl_ctrl_search_by_name(nl_cache, "nl80211");
	if (!nl80211) {
		DLOG_ERR("nl80211: didn't find nl80211 netlink control");
		goto out_clean_cache_a;
	}

	ret = nl80211_get_family("nl80211", "scan");
	if (ret >= 0) {
		ret = nl_socket_add_membership(nl_event_handle, ret);
	} else {
		DLOG_ERR("nl80211: couldn't add multicast membership"
			 " for scan events: %d (%s)", ret, strerror(-ret));
		goto out_clean_cache_a;
	}

	ret = nl80211_get_family("nl80211", "mlme");
	if (ret >= 0) {
		ret = nl_socket_add_membership(nl_event_handle, ret);
	} else {
		DLOG_ERR("nl80211: couldn't add multicast membership"
			 " for mlme events: %d (%s)", ret, strerror(-ret));
		goto out_clean_cache_a;
	}

	nl_family_id = genl_family_get_id(nl80211);
	DLOG_DEBUG("nl80211: netlink family id %d", nl_family_id);

	gio = g_io_channel_unix_new(nl_socket_get_fd(nl_event_handle));
	g_io_channel_set_close_on_unref(gio, TRUE);
	watch_id = g_io_add_watch(gio, G_IO_IN | G_IO_PRI | G_IO_ERR |
			G_IO_HUP, event_receive, &watch_id);

	return 0;

out_clean_cache_a:
	nl_cache_free(nl_event_cache);

out_clean_cache_b:
	nl_cache_free(nl_cache);

out_handle_destroy_a:
	nl_handle_destroy(nl_event_handle);

out_handle_destroy_b:
	nl_handle_destroy(nl_handle);

out_cb_clear:
	nl_cb_put(nl_cb);

	return ret;
}

static int error_handler(struct sockaddr_nl *nla, struct nlmsgerr *err,
			 void *arg)
{
	int *ret = arg;
	DLOG_DEBUG("nl80211: %s() %d\n", __func__, err->error);
	*ret = err->error;
	return NL_SKIP;
}

static int finish_handler(struct nl_msg *msg, void *arg)
{
	int *ret = arg;
	DLOG_DEBUG("nl80211: %s()\n", __func__);
	*ret = 0;
	return NL_SKIP;
}

static int ack_handler(struct nl_msg *msg, void *arg)
{
	int *ret = arg;
	//DLOG_DEBUG("nl80211: %s()\n", __func__);
	*ret = 0;
	return NL_STOP;
}

/**
   Send the netlink command to the kernel.
   @param msg structure Containing the command.
   @param valid_handler Pointer to call back function.
   @param valid_data Argument for call back function.
   @return status.
*/
int nl_send_recv_msgs(struct nl_msg *msg, int (*valid_handler)(struct nl_msg *, void *), void *valid_data)
{
	struct nl_cb *cb;
	int ret;

	cb = nl_cb_clone(nl_cb);

	if (!cb) {
		DLOG_ERR("nl80211: failed to alloc netlink callback");
		ret = -nl_get_errno();
		return ret;
	}
	ret = nl_send_auto_complete(nl_handle, msg);

	if (ret < 0) {
		DLOG_ERR("nl80211: failed to send netlink message");
		goto end_func;
	}

	ret = 1;

	if (nl_cb_err(cb, NL_CB_CUSTOM, error_handler, &ret)) {
		DLOG_DEBUG("nl80211: send and recv nl_cb_err returned error");		
		ret = -1;
		goto end_func;
	}
	if (nl_cb_set(cb, NL_CB_FINISH, NL_CB_CUSTOM, finish_handler, &ret)) {
		DLOG_DEBUG("nl80211: send and recv set finnish handler returned error");		
		ret = -1;
		goto end_func;
	}
	if (nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &ret)) {
		DLOG_DEBUG("nl80211: send and recv set ack handler returned error");		
		ret = -1;
		goto end_func;
	}

	if (valid_handler) {
		if (nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, valid_handler, 
			      valid_data)) {
			DLOG_DEBUG("nl80211: send and recv set valid handler error");		
			ret = -1;
			goto end_func;
		}
	}

	while (ret > 0) {
		if (nl_recvmsgs(nl_handle, cb)) {
			DLOG_ERR("nl80211: failed to receive netlink message"
				 " after command");
			goto end_func;
		}
	}

end_func:
	nl_cb_put(cb);
	return ret;
}

/**
   Cleaning the pointers on close down.
*/
void nl80211_cleanup(void)
{
	DLOG_DEBUG("nl80211: cleanup destroying all");

	if (nl80211)
		genl_family_put(nl80211);

	if (nl_cache)
		nl_cache_free(nl_cache);

	if (nl_handle)
		nl_handle_destroy(nl_handle);
}

/**
    Helper function for setting the WEP keys in netlink way.
    @param conn Connection parameters.
    @param msg Structure containing the command.
    @return status.
*/
static int nl80211_set_wep_keys(struct connect_params_t *conn, struct nl_msg *msg)
{
	int i, privacy = 0; 
	struct nlattr *nl_keys, *nl_key;
	
	for (i = 0; i < 4; i++) {
		if (conn->key_len[i] == 0)
			continue;

		privacy = 1;
		break;
	}
	if (!privacy)
		return 0;
	
	DLOG_DEBUG("nl80211: setting wep keys"); 
	NLA_PUT_FLAG(msg, NL80211_ATTR_PRIVACY);
	nl_keys = nla_nest_start(msg, NL80211_ATTR_KEYS);

	if (!nl_keys)
		goto nla_put_failure;

	for (i = 0; i < 4; i++) {
		if (conn->key_len[i] == 0)
			continue;

		char *key_phrase;
		nl_key = nla_nest_start(msg, i);
		if (!nl_key)
			goto nla_put_failure;

		NLA_PUT_U8(msg, NL80211_KEY_IDX, i);

		if (conn->key_len[i] == 5)
			NLA_PUT_U32(msg, NL80211_KEY_CIPHER, 
				    WLANCOND_CIPHER_SUITE_WEP40);
		else
			NLA_PUT_U32(msg, NL80211_KEY_CIPHER, 
				    WLANCOND_CIPHER_SUITE_WEP104);
		
		key_phrase = (caddr_t) &conn->key[i][0];

		NLA_PUT(msg, NL80211_KEY_DATA, conn->key_len[i], key_phrase);	

		/* If the default key is 1 and only one key is specified 
		   then it never enters the following if condition, adding 
		   +1 to i as the for loop starts from 0. */ 	
		
		if (conn->default_key == i+1) {
			NLA_PUT_FLAG(msg, NL80211_KEY_DEFAULT);
			DLOG_DEBUG("nl80211: default key: %d", 
				   conn->default_key);
		}
		nla_nest_end(msg, nl_key);
	}
	nla_nest_end(msg, nl_keys);
	return 0;

nla_put_failure:
	return -ENOBUFS;
}
/**
   Convert channel to frequency.
   @param channel Channel.
   @return frequency.
*/
static guint32 channel_to_freq(guint32 channel) {	
	guint32 freq;

	if (channel < 14) {
		freq = 2407 + (5 * /* 5 is the bandwidth of one channel */ channel);
	} else if (channel == 14) {
		freq = 2484;
	} else {
		freq = 5 * (channel + 1000);
	}
	return freq;
}
/**
   Convert frequency to channel.
   @param freq Frequency.
   @return channel.
*/
static guint32 freq_to_channel(guint32 freq)
{
	if (freq < 2484)
		return (freq - 2407) / 5;

	if (freq == 2484)
		return 14;

	return freq / 5 - 1000;
}
/**
    Connect function to connect to selected access point.
    @param scan_results Scan results.
    @return status.
*/
int nl80211_connect(struct scan_results_t *scan_results, struct wlan_status_t *wlan_status)
{
	struct connect_params_t *conn = &wlan_status->conn;
	struct nl_msg *msg;
	void *hdr;
	gint ret = -1;
	guint mode = NL80211_CMD_CONNECT;
	/* wep_shared_and_open_try is used to try the alternate 
	   authentication type in wep */
	enum nl80211_auth_type type = NL80211_AUTHTYPE_OPEN_SYSTEM;

	DLOG_DEBUG("nl80211: Starting the nl80211_connect");

	msg = nlmsg_alloc();

	if (msg == NULL) {
		return -ENOMEM;
	}
	if (get_mode() == WLANCOND_ADHOC)
		mode = NL80211_CMD_JOIN_IBSS;
	
	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, mode, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put nl80211 connect message");
		goto nla_put_failure;
	}

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, wlan_status->ifindex);

	NLA_PUT(msg, NL80211_ATTR_SSID, conn->ssid_len-1, conn->ssid);
	DLOG_DEBUG("nl80211: ESSID: %s", conn->ssid);

	if (memcmp(conn->bssid, "\0\0\0\0\0\0", ETH_ALEN)) {
		NLA_PUT(msg, NL80211_ATTR_MAC, ETH_ALEN, conn->bssid);
		print_mac(WLANCOND_PRIO_HIGH, "nl80211: BSSID", conn->bssid);
	}

	if (scan_results->channel) {
		guint32 freq = channel_to_freq(scan_results->channel);

		NLA_PUT_U32(msg, NL80211_ATTR_WIPHY_FREQ, freq);
		DLOG_DEBUG("nl80211: Channel: %d", scan_results->channel);
	}

	if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WEP) {
		DLOG_DEBUG("nl80211: WEP selected");

		if (get_mode() == WLANCOND_INFRA) {
			if (wlan_status->conn.wep_try_open_mode_first == TRUE) {
				type = NL80211_AUTHTYPE_OPEN_SYSTEM;		
				DLOG_DEBUG("nl80211: open system");
				wlan_status->conn.wep_try_open_mode_first = FALSE;			
			} else {
				type = NL80211_AUTHTYPE_SHARED_KEY;
				DLOG_DEBUG("nl80211: shared key");
				wlan_status->conn.wep_try_open_mode_first = TRUE;
			}
		}

		ret = nl80211_set_wep_keys(conn, msg);
		if (ret) {
			DLOG_DEBUG("nl80211: set key failure");
			goto nla_put_failure;
		}
	}

	NLA_PUT_U32(msg, NL80211_ATTR_AUTH_TYPE, type);

	if (get_wpa_mode() == TRUE ||
	    conn->authentication_type == EAP_AUTH_TYPE_WFA_SC) {
		struct wpa_ie_save_t *wpa_ie = &wlan_status->wpa_ie;

		NLA_PUT(msg, NL80211_ATTR_IE, wpa_ie->ie_len, wpa_ie->ie);
		
		if (get_wpa_mode() == TRUE || 
		   (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WEP) {
			NLA_PUT_FLAG(msg, NL80211_ATTR_PRIVACY);
		}
		DLOG_DEBUG("nl80211: Setting IE, len %d", wpa_ie->ie_len);
   	}	

	if (scan_results->wpa_ie_len && scan_results->wpa_ie &&
	   (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) != WLANCOND_WAPI_PSK &&
	   (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) != WLANCOND_WAPI_CERT) {

		enum nl80211_wpa_versions ver;

		if (conn->encryption & WLANCOND_ENCRYPT_WPA2_MASK)
			ver = NL80211_WPA_VERSION_2;
		else
			ver = NL80211_WPA_VERSION_1;

		DLOG_DEBUG("nl80211: %s", ver==NL80211_WPA_VERSION_2?
			   "WPA2":"WPA1");

		NLA_PUT_U32(msg, NL80211_ATTR_WPA_VERSIONS, ver);
    	}

	if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_PSK || 
            (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_EAP) {

		switch (wlan_status->pairwise_cipher) {
		case CIPHER_SUITE_TKIP:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITES_PAIRWISE, WLANCOND_CIPHER_SUITE_TKIP);
			DLOG_DEBUG("nl80211: TKIP pairwise");
			break;
		case CIPHER_SUITE_CCMP:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITES_PAIRWISE, WLANCOND_CIPHER_SUITE_CCMP);
			DLOG_DEBUG("nl80211: AES pairwise");
			break;
		case CIPHER_SUITE_WEP104:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITES_PAIRWISE, WLANCOND_CIPHER_SUITE_WEP104);
			DLOG_DEBUG("nl80211: WEP104 pairwise");
			break;
		case CIPHER_SUITE_WEP40:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITES_PAIRWISE, WLANCOND_CIPHER_SUITE_WEP40);
			DLOG_DEBUG("nl80211: WEP40 pairwise");
			break;
		default:
			DLOG_ERR("NL80211: Not supported unicast encryption");
			nlmsg_free(msg);
			return -1;
		} /* switch */

		switch (wlan_status->group_cipher) {
		case CIPHER_SUITE_TKIP:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITE_GROUP, WLANCOND_CIPHER_SUITE_TKIP);
			DLOG_DEBUG("nl80211: TKIP group");
			break;
		case CIPHER_SUITE_CCMP:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITE_GROUP, WLANCOND_CIPHER_SUITE_CCMP);
			DLOG_DEBUG("nl80211: AES group");
			break;
		case CIPHER_SUITE_WEP104:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITE_GROUP, WLANCOND_CIPHER_SUITE_WEP104);
			DLOG_DEBUG("nl80211: WEP104 group");
			break;
		case CIPHER_SUITE_WEP40:
			NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITE_GROUP, WLANCOND_CIPHER_SUITE_WEP40);
			DLOG_DEBUG("nl80211: WEP40 group");
			break;
		default:
			DLOG_ERR("nl80211: Not supported multicast encryption");
			nlmsg_free(msg);
			return -1;
		} /* switch */

		if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WPA_PSK) {
			NLA_PUT_U32(msg, NL80211_ATTR_AKM_SUITES, WLANCOND_KEY_MGMT_PSK);
			DLOG_DEBUG("nl80211: PSK");
		}
		else if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) ==  WLANCOND_WPA_EAP) { 
			NLA_PUT_U32(msg, NL80211_ATTR_AKM_SUITES, WLANCOND_KEY_MGMT_802_1X);
			DLOG_DEBUG("nl80211: EAP");
		}
		else {
			DLOG_ERR("nl80211: unsupported AKM ");
			goto nla_put_failure;
		}
	} /* WPA/WPA2 */
	else if ((conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_PSK ||
		 (conn->encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_CERT) {
			NLA_PUT_U16(msg, NL80211_ATTR_CONTROL_PORT_ETHERTYPE, 0x88B4);
			DLOG_DEBUG("nl80211: WAPI");
			NLA_PUT_FLAG(msg, NL80211_ATTR_CONTROL_PORT_NO_ENCRYPT);
	}

	/* Set IBSS basic data rates */
	if (get_mode() == WLANCOND_ADHOC) {
		if (scan_results->channel <= 14) {
			/* 2.4 GHz */
			const unsigned char rates[] = {
				2,  /*   1 Mbps */
				4,  /*   2 Mbps */
				11, /* 5.5 Mbps */
				22  /*  11 Mbps */
			};
			if (sizeof(rates) / sizeof(rates[0]) <= 
			    NL80211_MAX_SUPP_RATES)
				NLA_PUT(msg, NL80211_ATTR_BSS_BASIC_RATES, 
					sizeof(rates) / sizeof(rates[0]), 
					rates);
		} else {
			/* 5 GHz */
			const unsigned char rates[] = {
				12,  /*  6 Mbps */
				24,  /* 12 Mbps */
				48   /* 24 Mbps */
			};
			if (sizeof(rates) / sizeof(rates[0]) <= 
			    NL80211_MAX_SUPP_RATES)
				NLA_PUT(msg, NL80211_ATTR_BSS_BASIC_RATES, 
					sizeof(rates) / sizeof(rates[0]), 
					rates);
		}
	}
		
	ret = nl_send_recv_msgs(msg, NULL, NULL);
	if (ret) {
		DLOG_ERR("nl80211: MLME connect failed: ret=%d (%s)", 
			 ret, strerror(-ret));
		goto nla_put_failure;
	}

	ret = 0;
	DLOG_DEBUG("nl80211: Connect request send successfully");

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

/**
   Helper function for setting the operating mode using netlink.
   @param ifindex Interface index.   
   @param mode Operating mode.
   @return status.
*/
int nl80211_set_op_mode(guint32 ifindex, guint32 mode)
{
	struct nl_msg *msg;
	gint ret = -1;
	void *hdr;
	enum nl80211_iftype nlmode;
	
	msg = nlmsg_alloc();
	if (!msg)
		return -ENOMEM;

	switch (mode) {

		case WLANCOND_ADHOC:
			nlmode = NL80211_IFTYPE_ADHOC;
			DLOG_DEBUG("nl80211: selecting mode: ADHOC");
			break;
		case WLANCOND_INFRA:
			nlmode = NL80211_IFTYPE_STATION;
			DLOG_DEBUG("nl80211: selecting mode: INFRA");
			break;
		default:
			DLOG_ERR("Operating mode undefined");
			goto nla_put_failure;
	}
	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_SET_INTERFACE, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put set interface message");
		goto nla_put_failure;
	}
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
	NLA_PUT_U32(msg, NL80211_ATTR_IFTYPE, nlmode);

	ret = nl_send_recv_msgs(msg, NULL, NULL);

	if (ret) {
		DLOG_ERR("nl80211: failed to send set interface ret=%d (%s)", 
			 ret, strerror(-ret));
                goto nla_put_failure;
        }
	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}
/**
  Process phy info callback.
  @param msg nl_msg structure.
  @param arg arguments for the function.
  @return status.
*/
static int phy_info_handler(struct nl_msg *msg, void *arg)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb_msg[NL80211_ATTR_MAX + 1];
	
	if (nla_parse(tb_msg, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		      genlmsg_attrlen(gnlh, 0), NULL))
		return NL_SKIP;
	
	if (tb_msg[NL80211_ATTR_IFTYPE]) {
		switch (nla_get_u32(tb_msg[NL80211_ATTR_IFTYPE])) {
		case NL80211_IFTYPE_ADHOC:
			wlan_status.conn.mode = WLANCOND_ADHOC;
			break;
		case NL80211_IFTYPE_STATION:
			wlan_status.conn.mode = WLANCOND_INFRA;
			break;
		default:
			break;
		}
	}
	return NL_SKIP;
}
/**
   Helper function for getting the phy information.
   @param ifindex Interface index.   
   @return status.
*/
int nl80211_get_phy_info(guint32 ifindex)
{
	struct nl_msg *msg;
	gint ret = -1;
	void *hdr;
	
	msg = nlmsg_alloc();
	if (!msg)
		return -ENOMEM;

	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_GET_INTERFACE, 0);
	if (hdr == NULL) {
		goto nla_put_failure;
	}
	
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);

	ret = nl_send_recv_msgs(msg, phy_info_handler, NULL);

	if (ret)
		DLOG_ERR("nl80211: failed to send get phy info ret=%d (%s)", 
			 ret, strerror(-ret));

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}
/**
   Helper function for setting the channel.
   @param ifindex Interface index.   
   @param channel Channel.   
   @return status.
*/	
int nl80211_set_channel(guint32 ifindex, guint32 channel)
{
	struct nl_msg *msg;
	gint ret = -1;
	void *hdr;
	
	msg = nlmsg_alloc();
	if (!msg)
		return -ENOMEM;

	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_SET_WIPHY, 0);
	if (hdr == NULL) {
		goto nla_put_failure;
	}
	guint32 freq = channel_to_freq(channel);
	
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
	NLA_PUT_U32(msg, NL80211_ATTR_WIPHY_FREQ, freq);
	
	ret = nl_send_recv_msgs(msg, NULL, NULL);
	
	if (ret)
		DLOG_ERR("nl80211: failed to send set channel ret=%d (%s)", 
			 ret, strerror(-ret));
	
nla_put_failure:
	nlmsg_free(msg);
	return ret;
}	
/**
   Leave the adhoc network.
   @param ifindex Interface index.   
   @return status.
*/
int nl80211_leave_adhoc(guint32 ifindex)
{
        struct nl_msg *msg;
        gint ret = -1;
	void *hdr;

        msg = nlmsg_alloc();
        if (!msg)
		return -ENOMEM;

        hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, NL80211_CMD_LEAVE_IBSS, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put leave ibss message");
		goto nla_put_failure;
	}
        NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
        ret = nl_send_recv_msgs(msg, NULL, NULL);
        if (ret) {
                DLOG_ERR("nl80211: Leave IBSS failed: ret=%d (%s)", ret, 
			 strerror(-ret));
                goto nla_put_failure;
        }
        ret = 0;
        DLOG_DEBUG("nl80211: Leave IBSS request sent successfully");

nla_put_failure:
        nlmsg_free(msg);
        return ret;
}
/**
   MLME nl80211 commands.
   @param addr Access point MAC address.
   @param cmd Command.
   @param reason_code Reason for leaving.
   @param ifindex Interface index.
   @return status.
*/
int nl80211_mlme_command(guchar* addr, guint16 cmd, guint16 reason_code, guint32 ifindex)
{
	gint ret = -1;
	struct nl_msg *msg;
	void *hdr;

	msg = nlmsg_alloc();
	if (!msg)
		return -ENOMEM;

	switch (cmd)
	{	
		case WLANCOND_MLME_DEAUTH:
			DLOG_INFO("nl80211: Deauthenticating");
			hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0,
					  NL80211_CMD_DEAUTHENTICATE, 0);
			if (hdr == NULL) {
				DLOG_ERR("nl80211: Failed to deauthenticate");
				goto nla_put_failure;
			}		
			break;
		case WLANCOND_MLME_DISASSOC:		
			DLOG_INFO("nl80211: Disassociating");
			hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0,
					  NL80211_CMD_DISASSOCIATE, 0);
			if (hdr == NULL) {
				DLOG_ERR("nl80211: Failed to disassociate");
				goto nla_put_failure;
			}
			break;
		case WLANCOND_NL80211_MLME_DISCONNECT:
			DLOG_INFO("nl80211: Disconnecting");
			hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0,
					  NL80211_CMD_DISCONNECT, 0);
			if (hdr == NULL) {
				DLOG_ERR("nl80211: Failed to disconnect");
				goto nla_put_failure;
			}
			break;
		default:
			DLOG_ERR("nl80211: Invalid mlme command");
			goto nla_put_failure;
	}
		
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
	NLA_PUT_U16(msg, NL80211_ATTR_REASON_CODE, reason_code);
	NLA_PUT(msg, NL80211_ATTR_MAC, ETH_ALEN, addr);
	
	ret = nl_send_recv_msgs(msg, NULL, NULL);
        if (ret) {
		DLOG_DEBUG("nl80211: MLME command failed in %s: ret=%d (%s)", 
			   __FUNCTION__, ret, strerror(-ret));
                goto nla_put_failure;
        }
	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

/**
   Read data from get station command.
   @param msg nl_msg structure.
   @param arg will be updated with signal power.
   @return status.
*/
static int nl80211_signal_power_handler(struct nl_msg *msg, void *arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *sinfo[NL80211_STA_INFO_MAX + 1];
	guint8 *power = arg;
	*power = 0;

	static struct nla_policy stats_policy[NL80211_STA_INFO_MAX + 1] = {
		[NL80211_STA_INFO_INACTIVE_TIME] = { .type = NLA_U32 },
		[NL80211_STA_INFO_RX_BYTES] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_BYTES] = { .type = NLA_U32 },
		[NL80211_STA_INFO_RX_PACKETS] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_PACKETS] = { .type = NLA_U32 },
		[NL80211_STA_INFO_SIGNAL] = { .type = NLA_U8 },
		[NL80211_STA_INFO_TX_BITRATE] = { .type = NLA_NESTED },
		[NL80211_STA_INFO_LLID] = { .type = NLA_U16 },
		[NL80211_STA_INFO_PLID] = { .type = NLA_U16 },
		[NL80211_STA_INFO_PLINK_STATE] = { .type = NLA_U8 },
	};

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		  genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_STA_INFO]) {
		DLOG_ERR("nl80211: sta stats missing!");
		return NL_SKIP;
	}
	if (nla_parse_nested(sinfo, NL80211_STA_INFO_MAX,
			     tb[NL80211_ATTR_STA_INFO],
			     stats_policy)) {
		DLOG_ERR("nl80211: Failed to parse nested info attributes!");
		return NL_SKIP;
	}

	if (sinfo[NL80211_STA_INFO_SIGNAL])
		*power = (guint8)nla_get_u8(sinfo[NL80211_STA_INFO_SIGNAL]);
	
	return NL_SKIP;
}

/**
   Ask for station stats.
   @param power Signal power.
   @param wlan_status Wlan status.
   @return status.
*/
int nl80211_get_signal_info(guint8 *power, struct wlan_status_t *wlan_status)
{
        struct nl_msg *msg;
	gint ret = -1;
	void *hdr;
	struct connect_params_t *conn = &wlan_status->conn;
	
        msg = nlmsg_alloc();
        if (!msg)
                return ret;

        hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_GET_STATION, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put get station message");
		goto nla_put_failure;
	}

        NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, wlan_status->ifindex);
	NLA_PUT(msg, NL80211_ATTR_MAC, ETH_ALEN, conn->bssid);

	ret = nl_send_recv_msgs(msg, nl80211_signal_power_handler, power);
        if (ret) {
		DLOG_ERR("nl80211: MLME command failed in %s: ret=%d (%s)", 
			   __FUNCTION__, ret, strerror(-ret));
                goto nla_put_failure;
        }
	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

/**
   Set the power save state.
   @param power_save Enable/disable.
   @param ifindex Interface index.
   @return status.
*/
int nl80211_set_power_save(gboolean new_state, guint32 ifindex)
{
        struct nl_msg *msg;
	gint ret = -1;
	void *hdr;
	enum nl80211_ps_state ps_state;

        msg = nlmsg_alloc();
        if (!msg)
                return ret;

	if (new_state)
		ps_state = NL80211_PS_ENABLED;		
	else
		ps_state = NL80211_PS_DISABLED;
		
        hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_SET_POWER_SAVE, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put set power save message");
		goto nla_put_failure;
	}

        NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
	NLA_PUT_U32(msg, NL80211_ATTR_PS_STATE, ps_state);

	ret = nl_send_recv_msgs(msg, NULL, NULL);
        if (ret) {
		DLOG_ERR("nl80211: MLME command failed in %s: ret=%d (%s)", 
			   __FUNCTION__, ret, strerror(-ret));
                goto nla_put_failure;
        }
	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

/**
    Function to send scan trigger command.
    @param wlan_status Pointer to WLAN status structure.
    @param ssid_len Length of SSID if given.
    @param ssid SSID if given.
    @return status.
*/
int nl80211_scan(wlan_status_t* const wlan_status, int ssid_len, char* const ssid)
{
	struct nl_msg *msg;
	struct nl_msg *ssid_attr = NULL;

	void *hdr;
	gint ret = -ENOBUFS;

	msg = nlmsg_alloc();
	if (!msg) {
		nl_perror(__func__);
		return -ENOMEM;
	}

	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_TRIGGER_SCAN, 0);
	if (!hdr) {
		DLOG_ERR("nl80211: Failed to put nl80211 message");
		nlmsg_free(msg);
		return -ENOMEM;
	}

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, wlan_status->ifindex);

	ssid_attr = nlmsg_alloc();
	if (ssid_attr == NULL) {
		nl_perror(__func__);
		nlmsg_free(msg);
		return -ENOMEM;
	}
	
	if (ssid_len && ssid)
		NLA_PUT(ssid_attr, 1, ssid_len, ssid);
	else
		NLA_PUT(ssid_attr, 1, 0, "");
	
	if (nla_put_nested(msg, NL80211_ATTR_SCAN_SSIDS, ssid_attr)) {
		ret = -ENOMEM;
		goto nla_put_failure;
	}
	
	ret = nl_send_recv_msgs(msg, NULL, NULL);

nla_put_failure:
	nlmsg_free(ssid_attr);
	nlmsg_free(msg);
	return ret;
}

/**
    Function to send get scan results command.
    @param wlan_status Pointer to WLAN status structure.
    @param scan_results Pointer to scan result list.
    @return status.
*/
int nl80211_send_get_scan_results(wlan_status_t* const wlan_status, GSList** const scan_results)
{
	struct nl_msg *msg;

	void *hdr;
	gint ret = -ENOBUFS;

	msg = nlmsg_alloc();
	if (!msg) {
		nl_perror(__func__);
		return -ENOMEM;
	}

	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, NLM_F_DUMP, 
			  NL80211_CMD_GET_SCAN, 0);
	if (!hdr) {
		DLOG_ERR("nl80211: Failed to put nl80211 message");
		nlmsg_free(msg);
		return -ENOMEM;
	}

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, wlan_status->ifindex);

	scan_wlan_status = wlan_status;

	ret = nl_send_recv_msgs(msg, nl80211_scan_result_handler, scan_results);

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

/**
    Function to handle scan results.
    @param msg Netlink message.
    @param arg Pointer to scan result list.
    @return status.
*/
static int nl80211_scan_result_handler(struct nl_msg *msg, void* arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *bss[NL80211_BSS_MAX + 1];

	GSList **scan_results = (GSList**) arg;
	struct scan_results_t *scan_result;
	int ret = 0;

	static const struct nla_policy bss_policy[NL80211_BSS_MAX + 1] = {
		[NL80211_BSS_TSF] = { .type = NLA_U64 },
		[NL80211_BSS_FREQUENCY] = { .type = NLA_U32 },
		[NL80211_BSS_BSSID] = { },
		[NL80211_BSS_BEACON_INTERVAL] = { .type = NLA_U16 },
		[NL80211_BSS_CAPABILITY] = { .type = NLA_U16 },
		[NL80211_BSS_INFORMATION_ELEMENTS] = { },
		[NL80211_BSS_SIGNAL_MBM] = { .type = NLA_U32 },
		[NL80211_BSS_SIGNAL_UNSPEC] = { .type = NLA_U8 },
		[NL80211_BSS_STATUS] = { .type = NLA_U32 },
		[NL80211_BSS_SEEN_MS_AGO] = { .type = NLA_U32 },
		[NL80211_BSS_BEACON_IES] = { },
	};

	unsigned char *ie;
	size_t ie_len;
	guint16 capabilities;

	if (nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0), 
		      genlmsg_attrlen(gnlh, 0), NULL))
		return NL_SKIP;

	if (!tb[NL80211_ATTR_BSS])
		return NL_SKIP;

	if (nla_parse_nested(bss, NL80211_BSS_MAX, tb[NL80211_ATTR_BSS], 
			     (struct nla_policy *) bss_policy))
		return NL_SKIP;

	if (!bss[NL80211_BSS_BSSID]
	    || !bss[NL80211_BSS_INFORMATION_ELEMENTS]
	    || !bss[NL80211_BSS_CAPABILITY]
	    || !bss[NL80211_BSS_SIGNAL_MBM]
	    || !bss[NL80211_BSS_FREQUENCY])
		return NL_SKIP;

	int signal = nla_get_u32(bss[NL80211_BSS_SIGNAL_MBM]);
	if (signal > 0 || signal < WLANCOND_MINIMUM_SIGNAL * 100)
		return NL_SKIP;

	scan_result = g_slice_new0(struct scan_results_t);

	capabilities = nla_get_u16(bss[NL80211_BSS_CAPABILITY]);
	if (capabilities & 0x10)
		scan_result->cap_bits |= WLANCOND_WEP;

	scan_result->channel = freq_to_channel(
		nla_get_u32(bss[NL80211_BSS_FREQUENCY]));

	ie = nla_data(bss[NL80211_BSS_INFORMATION_ELEMENTS]);
	ie_len = nla_len(bss[NL80211_BSS_INFORMATION_ELEMENTS]);
	ret = nl80211_parse_ies(ie, ie_len, scan_result);
	
	if (ret == 0) {
		memcpy(scan_result->bssid, nla_data(bss[NL80211_BSS_BSSID]), 
		       ETH_ALEN);
		print_mac(WLANCOND_PRIO_LOW, "BSSID:", scan_result->bssid);
		
		scan_result->rssi = signal / 100;
		
		if (capabilities & 0x02)
			scan_result->cap_bits |= WLANCOND_ADHOC;
		else
			scan_result->cap_bits |= WLANCOND_INFRA;

		if (!(scan_result->cap_bits & WLANCOND_ENCRYPT_METHOD_MASK))
			scan_result->cap_bits |= WLANCOND_OPEN;

		*scan_results = g_slist_prepend(*scan_results, scan_result);
	} else {
		/* Don't print in the common case (wrong ESSID) */
		if (ret != -2)
			DLOG_DEBUG("Scan result discarded");
		clean_scan_results_item(scan_result, NULL);
	}

	return NL_SKIP;
}

/**
    Validates given data rate.
    @param data_rate Data rate to be validated.
    @return True if given data rate is supported by the device.
*/
static gboolean nl80211_is_supported_data_rate(guint8 data_rate) {

	switch (data_rate) {
	case 108: //  54 Mpbs
	case 96:  //  48 Mbps
	case 72:  //  36 Mbps
	case 48:  //  24 Mbps
	case 36:  //  18 Mbps
	case 24:  //  12 Mbps
	case 22:  //  11 Mbps
	case 18:  //   9 Mbps
	case 12:  //   6 Mbps
	case 11:  // 5.5 Mbps
	case 4:   //   2 Mbps
	case 2:   //   1 Mbps
		return TRUE;
	default:
		return FALSE;
	}
}

/**
   Function to parse WLAN information elements.
   @param ie Pointer to information element.
   @param ie_len Length of information element.
   @param scan_result Pointer to scan result structure.
   @return status.
*/
static int nl80211_parse_ies(unsigned char* const ie, size_t ie_len, scan_results_t* const scan_result)
{
	struct ap_info_t ap_info;
	unsigned char *ie_last_byte = ie + ie_len - 1;
	unsigned char *current_ie = ie;
	guchar* security_ie = NULL;
	guint security_ie_len = 0;
	unsigned int i;
	const guint32 WMM_QOS_OFFSET = 8;
	const guint8 WMM_OUI[] = { 0x00, 0x50, 0xF2, 0x02, 0x01 };
	const guint8 WMM_UAPSD_MASK = 0x80;
	const guint8 WPA1_OUI[] = { 0x00, 0x50, 0xf2, 1 };
	const guint8 WPS_OUI[] = { 0x00, 0x50, 0xf2, 0x04 };
	gboolean wapi_ie = FALSE;
	guint8 max_rate = 0;
	guint8 data_rate = 0;


	while (current_ie + 1 + current_ie[1] <= ie_last_byte) {

		switch (current_ie[0]) {

		case 0: /* SSID */
			if (scan_result->ssid_len)
				break; /* Bug 168117 workaround */

			if (current_ie[1] > WLANCOND_MAX_SSID_SIZE)
				return -1;
		
			scan_result->ssid_len = current_ie[1] + 1;

			if (scan_wlan_status->scan_ssid_len - 1 > 0 &&
			    ((gint) scan_result->ssid_len != scan_wlan_status->scan_ssid_len
			     || memcmp(current_ie + 2, scan_wlan_status->scan_ssid, scan_result->ssid_len - 1) != 0))
				return -2;

			strncpy(scan_result->ssid, (char*) current_ie + 2, current_ie[1]);
			scan_result->ssid[current_ie[1]] = '\0';
			DLOG_DEBUG("SSID: %s", scan_result->ssid);
			break;

		case 1: /* Supported data rates */
		case 50: /* Extended data rates */
			for (i = 0; i < current_ie[1]; i++ ) {
				data_rate = current_ie[2 + i];
				if (nl80211_is_supported_data_rate(data_rate & 0x7F)) {
					if ((data_rate & 0x7F) > max_rate)
						max_rate = data_rate & 0x7F;
				}
				else if (data_rate & 0x80) {
					/* Unsupported basic data rate */
					scan_result->cap_bits |= WLANCOND_UNSUPPORTED_NETWORK;
					DLOG_ERR("Basic data rate %u.%u Mbps not supported", (data_rate & 0x7F), data_rate % 2 ? 5 : 0);
				}
			}
			break;

		case 45: /* HT Capabilities */
			scan_result->cap_bits |= WLANCOND_RATE_HT_SUPPORT;
			DLOG_DEBUG("HT supported");
			break;

		case 48: /* RSN */
			if (current_ie[1] < 3)
				break;
			security_ie_len = current_ie[1] + 2;
			security_ie = current_ie;
			scan_result->cap_bits |= WLANCOND_WPA2;
			break;

		case 68: /* WAPI */
			if (current_ie[2] != WAPI_VERSION)
				break;
			security_ie_len = current_ie[1] + 2;
			security_ie = current_ie;
			wapi_ie = TRUE;
			break;
			
		case 221: /* WMM/WPA1/WPS */
			if (current_ie[1] < 6)
				break;

			if (!memcmp(&current_ie[2], WMM_OUI, sizeof(WMM_OUI))
			    && (current_ie[WMM_QOS_OFFSET] & WMM_UAPSD_MASK)) { 
				/* WMM U-APSD */
				scan_result->extra_cap_bits |= WLANCOND_UAPSD_SUPPORTED;
				DLOG_DEBUG("U-APSD supported");
			}
			else if (!(scan_result->cap_bits & WLANCOND_ENCRYPT_WPA2_MASK)
				 && memcmp(&current_ie[2], WPA1_OUI, 
					   sizeof(WPA1_OUI)) == 0) { /* WPA1 */
				security_ie_len = current_ie[1] + 2;
				security_ie = current_ie;
			}
			else if (memcmp(&current_ie[2], WPS_OUI, 
					sizeof(WPS_OUI)) == 0) { /* WPS */
				handle_wps_ie(&current_ie[6], scan_result, 
					      current_ie[1] + 2);
			}
			break;

		default:
			break;
		}

		current_ie += (current_ie[1] + 2);
	} /* while */

	if (scan_result->ssid_len == 0)
		return -1;

	/* Set maximum data rate */

	switch (max_rate) {
	case 108:
		scan_result->cap_bits |= WLANCOND_RATE_540;
		break;
	case 22:
		scan_result->cap_bits |= WLANCOND_RATE_110;
		break;
	case 96:
		scan_result->cap_bits |= WLANCOND_RATE_480;
		break;
	case 72:
		scan_result->cap_bits |= WLANCOND_RATE_360;
		break;
	case 48:
		scan_result->cap_bits |= WLANCOND_RATE_240;
		break;
	case 36:
		scan_result->cap_bits |= WLANCOND_RATE_180;
		break;
	case 24:
		scan_result->cap_bits |= WLANCOND_RATE_120;
		break;
	case 18:
		scan_result->cap_bits |= WLANCOND_RATE_90;
		break;
	case 12:
		scan_result->cap_bits |= WLANCOND_RATE_60;
		break;
	case 11:
		scan_result->cap_bits |= WLANCOND_RATE_55;
		break;
	case 4:
		scan_result->cap_bits |= WLANCOND_RATE_20;
		break;
	case 2:
		scan_result->cap_bits |= WLANCOND_RATE_10;
		break;
	default:
		DLOG_ERR("Invalid max rate: %d", max_rate);
		return -1;
	}

	if (security_ie) {
		memset(&ap_info, 0, sizeof(ap_info));

		if (wapi_ie == TRUE) {
			if (parse_wapi_ie(security_ie, security_ie_len, &ap_info) < 0) {
				return -1;
			}
		} else if (scan_result->cap_bits & WLANCOND_ENCRYPT_WPA2_MASK) {
			if (parse_rsn_ie(security_ie, security_ie_len, &ap_info) < 0) {
				return -1;
			}
		} else {
			if (parse_wpa_ie(security_ie, security_ie_len, &ap_info) < 0) {
				return -1;
			}
		}
	
		save_caps(scan_result, &ap_info, security_ie, security_ie_len);
	}

	return 0;
}

/**
   Set the connection quality monitor.
   @param threshold Threshold value for receiving events.
   @param hyst Hysteresis value.
   @return status.
*/
int nl80211_set_cqm(guint32 threshold, guint32 hyst)
{
	struct nl_msg *msg;
	struct nl_msg *cqm_msg;
	int ret = -1;
	void *hdr;

	/* CQM is not supported in Adhoc */
	if (get_mode() == WLANCOND_ADHOC)
		return 0;
	
	msg = nlmsg_alloc();
	if (!msg)
		return ret;

	cqm_msg = nlmsg_alloc();
	if (!cqm_msg) {
		nlmsg_free(msg);
		return ret;
	}

	hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_SET_CQM, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put set cqm message");
		goto nla_put_failure;
	}

	NLA_PUT_U32(cqm_msg, NL80211_ATTR_CQM_RSSI_THOLD, threshold);
	NLA_PUT_U32(cqm_msg, NL80211_ATTR_CQM_RSSI_HYST, hyst);

	if (nla_put_nested(msg, NL80211_ATTR_CQM, cqm_msg)) {
		ret = -ENOMEM;
		goto nla_put_failure;
	}

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, wlan_status.ifindex);

	ret = nl_send_recv_msgs(msg, NULL, NULL);
	if (ret) {
		DLOG_ERR("nl80211: set cqm failed: err=%d %s)", ret, 
			 strerror(-ret));
		goto nla_put_failure;
	}

	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
	nlmsg_free(cqm_msg);
	return ret;
}

/**
  Save WPA capabilities and print them to log.
  @param scan_results Pointer to scan results structure.
  @param ap_info Pointer to AP information structure.
  @param p Pointer to event buffer.
  @param length Event buffer length.
 */
static void save_caps(struct scan_results_t *scan_results,
		      struct ap_info_t *ap_info, unsigned char *p,
		      unsigned int length)
{
	scan_results->wpa_ie = g_memdup(p, length);
	scan_results->wpa_ie_len = length;

	/* Save method */
	scan_results->cap_bits |= 
		ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK;
	
	/* WPA */
	if ((ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) == 
	    WLANCOND_WPA_PSK || 
	    (ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) == 
	    WLANCOND_WPA_EAP) {
		if ((ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) == 
		    WLANCOND_WPA_EAP) {
			/* No WPS in EAP mode */
			scan_results->cap_bits &= ~WLANCOND_WPS_MASK;
		}
		/* Remove WEP bit to make UI show correct dialogs */
		scan_results->cap_bits &= ~WLANCOND_WEP;
		
		DLOG_DEBUG("%s %s supported",
			   (scan_results->cap_bits & WLANCOND_WPA2) ? 
			   "WPA2":"WPA",
			   (ap_info->key_mgmt & WLANCOND_WPA_PSK) ? 
			   "PSK":"EAP");
		
		/* Algorithms */
		/* Pairwise */
		if (ap_info->pairwise_cipher & CIPHER_SUITE_CCMP)
			scan_results->cap_bits |= WLANCOND_WPA_AES;
	       
		if (ap_info->pairwise_cipher & CIPHER_SUITE_TKIP)
			scan_results->cap_bits |= WLANCOND_WPA_TKIP;

		if (wlan_status.allow_all_ciphers == TRUE) {
			if (ap_info->pairwise_cipher & CIPHER_SUITE_WEP40) {
				scan_results->extra_cap_bits |= WLANCOND_WEP40;
			}
			if (ap_info->pairwise_cipher & CIPHER_SUITE_WEP104) {
				scan_results->extra_cap_bits |= WLANCOND_WEP104;
			}
			if (ap_info->group_cipher & CIPHER_SUITE_WEP40) {
				scan_results->extra_cap_bits |= WLANCOND_WEP40_GROUP;
			}
			if (ap_info->group_cipher & CIPHER_SUITE_WEP104) {
				scan_results->extra_cap_bits |= WLANCOND_WEP104_GROUP;
			}
		} else {
			if (ap_info->pairwise_cipher & CIPHER_SUITE_WEP40 ||
			    ap_info->pairwise_cipher & CIPHER_SUITE_WEP104 ||
			    ap_info->group_cipher & CIPHER_SUITE_WEP40 ||
			    ap_info->group_cipher & CIPHER_SUITE_WEP104) {
				DLOG_DEBUG("In WPA mode WEP is not allowed");
				scan_results->cap_bits |= 
					WLANCOND_UNSUPPORTED_NETWORK;
			}
		}
		DLOG_DEBUG("%s/%s/%s/%s for unicast",
			   (ap_info->pairwise_cipher & CIPHER_SUITE_CCMP)?"AES":"-",
			   (ap_info->pairwise_cipher & CIPHER_SUITE_TKIP)?"TKIP":"-",
			   (ap_info->pairwise_cipher & CIPHER_SUITE_WEP104)?"WEP104":"-",
			   (ap_info->pairwise_cipher & CIPHER_SUITE_WEP40)?"WEP40":"-");
		/* Group */
		if (ap_info->group_cipher & CIPHER_SUITE_CCMP)
			scan_results->cap_bits |= WLANCOND_WPA_AES_GROUP;
		
		if (ap_info->group_cipher & CIPHER_SUITE_TKIP)
			scan_results->cap_bits |= WLANCOND_WPA_TKIP_GROUP;
		
		DLOG_DEBUG("%s/%s/%s/%s for multicast",
			   (ap_info->group_cipher & CIPHER_SUITE_CCMP)?"AES":"-",
			   (ap_info->group_cipher & CIPHER_SUITE_TKIP)?"TKIP":"-",
			   (ap_info->group_cipher & CIPHER_SUITE_WEP104)?"WEP104":"-",
			   (ap_info->group_cipher & CIPHER_SUITE_WEP40)?"WEP40":"-");
		
	}

	/* WAPI */
	else if ((ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) == 
	    WLANCOND_WAPI_PSK || 
	    (ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) == 
	    WLANCOND_WAPI_CERT) {

		DLOG_DEBUG("WAPI %s supported",
			   ((ap_info->key_mgmt & WLANCOND_ENCRYPT_METHOD_MASK) 
			    == WLANCOND_WAPI_PSK) ? "PSK":"CERT");
		
		if (ap_info->pairwise_cipher & CIPHER_SUITE_SMS4)
			scan_results->cap_bits |= WLANCOND_WAPI_SMS4; 
		if (ap_info->group_cipher & CIPHER_SUITE_SMS4)
			scan_results->cap_bits |= WLANCOND_WAPI_SMS4_GROUP;
		
		DLOG_DEBUG("%s/%s ciphers",
			   (ap_info->pairwise_cipher & CIPHER_SUITE_SMS4)?
			   "SMS4":"-",
			   (ap_info->group_cipher & CIPHER_SUITE_SMS4)?
			   "SMS4":"-");
	}
}

/**
   Set the transmit power level.
   @param tx_power value.
   @param ifindex Interface index.
   @return status.
*/
int nl80211_set_tx_power(int tx_power, guint32 ifindex)
{
        struct nl_msg *msg;
	gint ret = -1;
	enum nl80211_tx_power_setting type = NL80211_TX_POWER_LIMITED;
	void *hdr;

        msg = nlmsg_alloc();
        if (!msg)
                return ret;

        hdr = genlmsg_put(msg, getpid(), 0, nl_family_id, 0, 0, 
			  NL80211_CMD_SET_WIPHY, 0);
	if (hdr == NULL) {
		DLOG_ERR("nl80211: Failed to put set tx power message");
		goto nla_put_failure;
	}

        NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, ifindex);
	NLA_PUT_U32(msg, NL80211_ATTR_WIPHY_TX_POWER_SETTING, type);
	NLA_PUT_U32(msg, NL80211_ATTR_WIPHY_TX_POWER_LEVEL, tx_power);

	ret = nl_send_recv_msgs(msg, NULL, NULL);
        if (ret) {
                DLOG_ERR("nl80211: set tx power command failed: ret=%d (%s)",
			 ret, strerror(-ret));
                goto nla_put_failure;
        }

nla_put_failure:
	nlmsg_free(msg);
	return ret;
}

