/**
   @file dbus-helper.c

   Copyright (C) 2004-2010 Nokia Corporation. All rights reserved.

   @author Janne Ylalehto <janne.ylalehto@nokia.com>
   @author Johan Hedberg <johan.hedberg@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


*/
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <osso-ic-dbus.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>
#include <eap-dbus.h>
#include <mce/mode-names.h>
#ifdef USE_MCE_MODE
#include <mce/dbus-names.h>
#endif
#include "log.h"
#include "dbus-helper.h"
#include "dbus.h"
#include "../common.h"
#include "../dbus-handler.h"

#define OUT_OF_MEMORY_STR "Out of memory"

static guint dbus_rule_state = 0;
/**
   Check DBUS rule sanity.
   @param rule Rule to check.
   @param add_or_remove Rule is added or removed.
   @return status.
*/
static gboolean rule_state_check(guint rule, guint add_or_remove) {
	if (add_or_remove == ADD) {
		if (rule & dbus_rule_state) {
			DLOG_ERR("Rule %02x exists!", rule);
			return FALSE;
		}
		dbus_rule_state |= rule;
		return TRUE;
	}
	if ((rule & dbus_rule_state) == 0) {
		DLOG_DEBUG("Rule %02x does not exist!", rule);
		return FALSE;
	}
	dbus_rule_state &=~rule;
	return TRUE;
}
		
DBusMessage *new_dbus_signal(const char *path,
		const char *interface,
		const char *name,
		const char *destination) {
	DBusMessage *signal;

	signal = dbus_message_new_signal(path, interface, name);
	if (signal == NULL) {
		die(OUT_OF_MEMORY_STR);
	}

	if (destination) {
		if (!dbus_message_set_destination(signal, destination)) {
			die(OUT_OF_MEMORY_STR);
		}
	}

	dbus_message_set_no_reply(signal, TRUE);

	return signal;
}

DBusMessage *new_dbus_method_return(DBusMessage *message) {
	DBusMessage *reply;

	reply = dbus_message_new_method_return(message);
	if (reply == NULL) {
		die(OUT_OF_MEMORY_STR);
	}

	return reply;
}

DBusMessage *new_dbus_error(DBusMessage *message, const char *name) {
	DBusMessage *error;

	error = dbus_message_new_error(message, name, NULL);
	if (error == NULL) {
		die(OUT_OF_MEMORY_STR);
	}

	return error;
}

void send_and_unref(DBusConnection *connection, DBusMessage *message) {
	if (!dbus_connection_send(connection, message, NULL)) {
		DLOG_ERR("Sending message failed!");
	}
	dbus_connection_flush(connection);
	dbus_message_unref(message);
}

void send_invalid_args(DBusConnection *connection, DBusMessage *message) {
	DBusMessage *reply;

	reply = new_dbus_error(message, DBUS_ERROR_INVALID_ARGS);

	return send_and_unref(connection, reply);
}

void append_dbus_args(DBusMessage *message, int first_arg_type, ...) {
	dbus_bool_t ret;
	va_list ap;

	va_start(ap, first_arg_type);
	ret = dbus_message_append_args_valist(message, first_arg_type, ap);
	va_end(ap);

	if (ret == FALSE) {
		die("dbus_message_append_args failed");
	}
}

gint get_device_mode(DBusConnection *connection, dbus_uint32_t *radio_mode) {
#ifdef USE_MCE_MODE
	DBusError derror;
	DBusMessage *message, *reply;

	message = dbus_message_new_method_call(MCE_SERVICE,
			MCE_REQUEST_PATH,
			MCE_REQUEST_IF,
			MCE_RADIO_STATES_GET);

	if (message == NULL)
		return -1;

	dbus_error_init(&derror);
	reply = dbus_connection_send_with_reply_and_block(connection,
			message,
			-1,
			&derror);
	dbus_message_unref(message);
	if (dbus_error_is_set(&derror)) {
		DLOG_ERR("Getting device mode from MCE failed: %s",
				derror.message);
		dbus_error_free(&derror);
		if (reply)
			dbus_message_unref(reply);
		return -1;
	}
	if (!dbus_message_get_args(reply, NULL,
				DBUS_TYPE_UINT32, radio_mode,
				DBUS_TYPE_INVALID)) {
		DLOG_ERR("Invalid arguments for MCE req_device_mode reply");
		dbus_message_unref(reply);
		return -1;
	}

	dbus_message_unref(reply);

	return 0;
#else /* USE_MCE_MODE */
	return -1;
#endif
}

#define MCE_DBUS_TIMEOUT 2000 //2s timeout for MCE
/**
   Check activity.
   @return activity.
*/
dbus_bool_t get_activity(void) {
	DBusMessage *msg, *reply;
	dbus_bool_t inactivity = FALSE;
	DBusError error;

	dbus_error_init(&error);

	msg = dbus_message_new_method_call(
		MCE_SERVICE,
		MCE_REQUEST_PATH,
		MCE_REQUEST_IF,
		MCE_INACTIVITY_STATUS_GET);
	
	if (msg == NULL) {
		return inactivity;
	}

	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, MCE_DBUS_TIMEOUT, &error);

	dbus_message_unref(msg);

	if (dbus_error_is_set(&error)) {
		DLOG_ERR("Failed to ask inactivity status: %s",
			 error.name);
		dbus_error_free(&error);
		if (reply)
			dbus_message_unref(reply);
		return inactivity;
	}
	
	if (!dbus_message_get_args(reply, NULL,
				   DBUS_TYPE_BOOLEAN, &inactivity,
                                   DBUS_TYPE_INVALID))
        {
		DLOG_ERR("Could not get inactivity");
		if (reply)
			dbus_message_unref(reply);
		return FALSE;
        }
	
	DLOG_DEBUG("Inactivity = %s", inactivity==FALSE?"FALSE":"TRUE");

	dbus_message_unref(reply);
	
	return inactivity;
}
/**
   Get call state.
   @return state.
*/
guint get_call_state(void) {
	DBusMessage *msg, *reply;
	char* active;
	char* call_type;
	DBusError error;
	guint ret = CALL_CELLULAR;

	dbus_error_init(&error);

	msg = dbus_message_new_method_call(
		MCE_SERVICE,
		MCE_REQUEST_PATH,
		MCE_REQUEST_IF,
		MCE_CALL_STATE_GET);
	
	if (msg == NULL) {
		return ret;
	}

	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, MCE_DBUS_TIMEOUT, &error);

	dbus_message_unref(msg);

	if (dbus_error_is_set(&error)) {
		DLOG_ERR("Failed to ask call state: %s",
			 error.name);
		dbus_error_free(&error);
		if (reply)
			dbus_message_unref(reply);
		return ret;
	}
	
	if (!dbus_message_get_args(reply, NULL,
				   DBUS_TYPE_STRING, &active,
				   DBUS_TYPE_STRING, &call_type,
				   DBUS_TYPE_INVALID))
        {
		DLOG_ERR("Could not get call state");
		if (reply)
			dbus_message_unref(reply);
		return ret;
        }
	
	if (active && !strncmp(active, MCE_CALL_STATE_NONE, 
			       strlen(MCE_CALL_STATE_NONE)))
		ret = 0;
	
	DLOG_DEBUG("Call state: %s", active);

	dbus_message_unref(reply);
	
	return ret;
}
#define WLANCOND_WAIT_COUNTRY 2000 //2s
/**
   Check country code.
   @return country code or an error.
*/
char* check_country_code(void) {
	DBusMessage *msg, *reply;
	DBusMessageIter iter, subiter;
	DBusError error;
	char* iface_name = (char*)CSNET_OPER_IFACE;
	char* prop_name = (char*)CSNET_OPER_MCC; 
	char* country = NULL;
	char* csnet_country = NULL;

	dbus_error_init(&error);

	msg = dbus_message_new_method_call(
		CSNET_INTERFACE,
		CSNET_OBJECT,
		DBUS_INTERFACE_PROPERTIES,
		CSNET_ACTION);

	if (msg == NULL) {
		return NULL;
	}

	dbus_message_append_args(msg, DBUS_TYPE_STRING, &iface_name,
				 DBUS_TYPE_STRING, &prop_name,
				 DBUS_TYPE_INVALID);

	reply = dbus_connection_send_with_reply_and_block(
		get_dbus_connection(), msg, WLANCOND_WAIT_COUNTRY, &error);

	dbus_message_unref(msg);

	if (dbus_error_is_set(&error)) {
		DLOG_ERR("Failed to ask registration status: %s",
			 error.name);
		dbus_error_free(&error);
		if (reply)
			dbus_message_unref(reply);
		return NULL;
	}

	dbus_message_iter_init(reply, &iter);

	dbus_message_iter_recurse(&iter, &subiter);

	if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_VARIANT &&
	    dbus_message_iter_get_arg_type(&subiter) == DBUS_TYPE_STRING) {
		
		dbus_message_iter_get_basic(&subiter, &csnet_country);

		country = g_strdup(csnet_country);

		if (country && strlen(country))
			DLOG_INFO("Device country code: %s", country);
		else
			DLOG_DEBUG("No country available");
	}
	
	dbus_message_unref(reply);
	
	return country;
}

#ifdef USE_MCE_MODE
gboolean add_mode_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_MCE_MODE, ADD) == FALSE)
		return TRUE;
	
	dbus_bus_add_match(connection,
			"interface=" MCE_SIGNAL_IF
			",member=" MCE_RADIO_STATES_SIG, NULL);

	return dbus_connection_add_filter(connection,
		(DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}
#ifdef ACTIVITY_CHECK
gboolean add_activity_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_MCE_ACTIVITY, ADD) == FALSE)
		return TRUE;

	dbus_bus_add_match(connection,
			"interface=" MCE_SIGNAL_IF
			",member=" MCE_INACTIVITY_SIG, NULL);

	return dbus_connection_add_filter(connection,
		(DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}
void remove_activity_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_MCE_ACTIVITY, REMOVE) == FALSE)
		return;

	dbus_bus_remove_match(connection,
			      "interface=" MCE_SIGNAL_IF
			      ",member=" MCE_INACTIVITY_SIG, NULL);

	dbus_connection_remove_filter(connection,
		(DBusHandleMessageFunction)wlancond_req_handler, NULL);
}
#endif
#endif

gboolean add_icd_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_ICD, ADD) == FALSE)
		return TRUE;

	dbus_bus_add_match(connection,
			"interface=" ICD_DBUS_INTERFACE
			",member=" ICD_STATUS_CHANGED_SIG, NULL);

	return dbus_connection_add_filter(connection,
		(DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}

void remove_icd_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_ICD, REMOVE) == FALSE)
		return;

	dbus_bus_remove_match(connection,
			      "interface=" ICD_DBUS_INTERFACE
			      ",member=" ICD_STATUS_CHANGED_SIG, NULL);

	dbus_connection_remove_filter(connection,
		(DBusHandleMessageFunction)wlancond_req_handler, NULL);
}
gboolean add_eap_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_EAP, ADD) == FALSE)
		return TRUE;

        dbus_bus_add_match(connection,
                        "interface=" EAP_SIG_INTERFACE
                        ",member=" EAP_AUTH_STATUS_SIG, NULL);

        return dbus_connection_add_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}
void remove_eap_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_EAP, REMOVE) == FALSE)
		return;

        dbus_bus_remove_match(connection,
			      "interface=" EAP_SIG_INTERFACE
			      ",member=" EAP_AUTH_STATUS_SIG, NULL);

        dbus_connection_remove_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL);
}

gboolean add_csd_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_CSD, ADD) == FALSE)
		return TRUE;

	dbus_bus_add_match(connection,
			"type='signal',interface=" CSNET_OPER_IFACE
			",member=" CSNET_OPERATOR_CHANGED_SIG, NULL);

	return dbus_connection_add_filter(
			connection,
			(DBusHandleMessageFunction)wlancond_req_handler,
			NULL, NULL);
}

gboolean add_context_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_CONTEXT, ADD) == FALSE)
		return TRUE;

	dbus_bus_add_match(connection,
			"type='signal',interface=" POLICY_SERVICE_NAME
			",member=" POLICY_ACTIONS_SIG, NULL);
        return dbus_connection_add_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}
void remove_context_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_CONTEXT, REMOVE) == FALSE)
		return;

        dbus_bus_remove_match(connection,
			"type='signal',interface=" POLICY_SERVICE_NAME
			",member=" POLICY_ACTIONS_SIG, NULL);

        dbus_connection_remove_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL);
}

gboolean add_gprs_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_GPRS, ADD) == FALSE)
		return TRUE;

	dbus_bus_add_match(connection,
			"type='signal',interface=" GPRS_DBUS_INTERFACE
			",member=" GPRS_DBUS_DATACOUNTERS_SIG, NULL);
        return dbus_connection_add_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL, NULL);
}
void remove_gprs_listener(DBusConnection *connection) {
	if (rule_state_check(RULE_GPRS, REMOVE) == FALSE)
		return;

        dbus_bus_remove_match(connection,
			"type='signal',interface=" GPRS_DBUS_INTERFACE
			",member=" GPRS_DBUS_DATACOUNTERS_SIG, NULL);

        dbus_connection_remove_filter(connection,
                (DBusHandleMessageFunction)wlancond_req_handler, NULL);
}
