/**
  @file wapi.c

  Copyright (C) 2009 Nokia Corporation. All rights reserved.

  @author Usman Rahim <usman.rahim@nokia.com>

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/ioctl.h>
#include <glib.h>
#include <glib-object.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/if_ether.h>

#define DBUS_API_SUBJECT_TO_CHANGE

#include <dbus/dbus.h>
#include <eap-dbus.h>
#include <wlancond.h>
#include "wpa.h"
#include "wapi.h"
#include "common.h"
#include "dbus-handler.h"
#include "log.h"

#define COUNT_LE(p,s) *p++ = (unsigned char)(s & 0x00ff); *p++ = (unsigned char)((s & 0xff00) >> 8); 

/**
  Parse wapi authentication and key management cipher suite. 
  @param suite Authentication and management suite.
  @return status.
 */
static guint parse_wapi_key_mgmt_suite(guint8 *suite)
{
	if (memcmp(suite, WAPI_KEY_MGMT_CERT, CIPHER_SUITE_LEN) == 0)
		return WLANCOND_WAPI_CERT;
	if (memcmp(suite, WAPI_KEY_MGMT_PSK, CIPHER_SUITE_LEN) == 0)
		return WLANCOND_WAPI_PSK;

	return 0;
}

/**
  Parse wapi cipher suite.
  @param suite Cipher suite.
  @return status.
 */

static guint parse_wapi_cipher_suite(guint8 *suite)
{
	if (memcmp(suite, CIPHER_SUITE_WAPI, CIPHER_SUITE_LEN) == 0)
		return CIPHER_SUITE_SMS4;

	return 0;
}

/**
  Generate WAPI Information Element.
  @param encryption Encryption settings.
  @param wlan_status Wlan status information.
  @return status.
 */
int generate_wapi_ie(guint32 encryption, struct wlan_status_t *wlan_status)
{
	guint8 *wapi_ie;
	struct wapi_ie_t *wapi_hdr;
	guint8* hdr_p;

	/* Update the length if you change something */
	wapi_ie = g_malloc(sizeof(struct wapi_ie_t)+8+CIPHER_SUITE_LEN*3+WLANCOND_PMKID_LEN);
	
	wapi_hdr = (struct wapi_ie_t*) wapi_ie;
	wapi_hdr->element_id = WAPI_ELEMENT;

	wapi_hdr->version = WAPI_VERSION;
	hdr_p = (guint8*)(wapi_hdr + 1);

	/* AKM count */
	COUNT_LE(hdr_p, 1);

	/* AKM suite */
	if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_PSK) {
		memcpy(hdr_p, WAPI_KEY_MGMT_PSK, CIPHER_SUITE_LEN);
	} else if ((encryption & WLANCOND_ENCRYPT_METHOD_MASK) == WLANCOND_WAPI_CERT){
		memcpy(hdr_p, WAPI_KEY_MGMT_CERT, CIPHER_SUITE_LEN);
	} else {
		DLOG_ERR("Unknown key management suite");
		g_free(wapi_ie);
		return -1;
	}

	hdr_p += CIPHER_SUITE_LEN;

	/* Unicast count */
	COUNT_LE(hdr_p, 1);

	/* Unicast suite */
	if (wlan_status->pairwise_cipher == CIPHER_SUITE_SMS4) {
		memcpy(hdr_p, CIPHER_SUITE_WAPI, CIPHER_SUITE_LEN);
	} else {
		DLOG_ERR("Unsupported pairwise cipher suite");
		g_free(wapi_ie);
		return -1;
	}
	
	hdr_p += CIPHER_SUITE_LEN;

	/* Multicast cipher suite */
	if (wlan_status->group_cipher == CIPHER_SUITE_SMS4) {
		memcpy(hdr_p, CIPHER_SUITE_WAPI, CIPHER_SUITE_LEN);
	} else {
		DLOG_ERR("Unsupported group cipher suite");
		g_free(wapi_ie);
		return -1;
	}

	hdr_p += CIPHER_SUITE_LEN;

	/* Capabilities are empty */
	
	COUNT_LE(hdr_p, 0);

	/* BKID count and BKID */

	 unsigned char *bkid;

	/* find_pmkid_from_pmk_cache can be used 
	   due to similar structure */

	if (find_pmkid_from_pmk_cache(wlan_status->conn.bssid, &bkid)) {
		DLOG_ERR("Error trying to retrieve the bkid.");
		return -1;
	}
	
	if (bkid != NULL) {
		COUNT_LE(hdr_p, 1);
		memcpy(hdr_p, bkid, WLANCOND_PMKID_LEN);
		hdr_p += WLANCOND_PMKID_LEN;
	}

	/* 2 is subtracted from length since 
	   element ID and length are excluded */	
	wapi_hdr->length = (hdr_p - wapi_ie) - 2;
	
#ifdef DEBUG_IE
	int i;
	printf("Own WPA IE:\n");
	for (i=0;i<wapi_hdr->length;i++) {
		printf("0x%02x ", i,wapi_ie[i]);
	}
	printf("\n");
#endif

	update_own_ie(wapi_ie, wapi_hdr->length + 2);

	return 0;
}

/**
  Parse WAPI IE.
  @param wapi_ie WPA IE.
  @param wapi_ie_len WPA IE length.
  @param ap_info AP Information pointer.
  @return status.
 */
int parse_wapi_ie(unsigned char* wapi_ie, unsigned int wapi_ie_len,
		struct ap_info_t* ap_info)
{
	struct wapi_ie_t *wapi_hdr;
	guint8 *hdr_p;
	guint ind;
	guint i;
	guint cipher_count = 0;

	/* Do data checking, we have to make sure all the time that we
	   don't go past the IE length , index variable counts the
	   remaining data, the spec says that all data after the version
	   field is optional */

	wapi_hdr = (struct wapi_ie_t*) wapi_ie;

	if (wapi_ie_len < sizeof(struct wapi_ie_t)) {
		DLOG_ERR("WAPI IE too short");
		return -1;
	}

	if (wapi_hdr->element_id != WAPI_ELEMENT) {
		DLOG_ERR("Unknown WAPI IE received");
		return -1;
	}

	ind = wapi_ie_len - sizeof(*wapi_hdr);
	hdr_p = (guint8*)(wapi_hdr + 1);

	/* AKM */
	if (ind >= 2) {
		ap_info->key_mgmt = 0;
		cipher_count = *(guint16*)hdr_p;
		hdr_p += 2;
		ind -= 2;

		if (cipher_count == 0 || ind < cipher_count *
				CIPHER_SUITE_LEN) {
			DLOG_ERR("Invalid key mgmt cipher count or length");
			return -1;
		}

		/* Currently supports only 1 cipher suite */
		for (i = 0; i < cipher_count; i++) {
			ap_info->key_mgmt |= parse_wapi_key_mgmt_suite(hdr_p);
			ind -= CIPHER_SUITE_LEN;
			hdr_p += CIPHER_SUITE_LEN;
		}

	} else if (ind == 1) {
		DLOG_ERR("Remaining data too short");
		return -1;
	}

	/* Unicast cipher suite */
	if (ind >= 2) {
		ap_info->pairwise_cipher = 0;
		cipher_count = *(guint16*)hdr_p;
		ind -= 2;

		if (cipher_count == 0) {
			DLOG_ERR("No pairwise ciphers");
			// Return 0 instead of an error
			return 0;
		}

		if (ind < cipher_count * CIPHER_SUITE_LEN) {
			DLOG_ERR("Invalid pairwise cipher length");
			return -1;
		}

		hdr_p += 2;

		for (i = 0; i < cipher_count; i++) {
			ap_info->pairwise_cipher |= parse_wapi_cipher_suite(hdr_p);
			ind -= CIPHER_SUITE_LEN;
			hdr_p += CIPHER_SUITE_LEN;
		}
	} else if (ind == 1) {
		DLOG_ERR("Remaining data too short");
		return -1;
	}

	/* Multicast cipher suite */

	if (ind >= CIPHER_SUITE_LEN) {
		ap_info->group_cipher = parse_wapi_cipher_suite(hdr_p);
		hdr_p += CIPHER_SUITE_LEN;
		ind -= CIPHER_SUITE_LEN;
	} else {
		DLOG_ERR("Strange length in WAPI IE");
		return -1;
	}

	/* Capabilities */	
	if (ind >= 2) {
		hdr_p += 2;
		ind -= 2;
	}

	if (ind > 0) {
		DLOG_DEBUG("IE includes BKID data");
	}

	return 0;
}
